/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <GL/glx.h>
#include <GL/gl.h>
#include <math.h>
#include "my_util.h"
#include "farray_util.h"
char glplot_font_small[128] =
  {"-adobe-times-medium-r-normal--10-100-75-75-p-54-iso8859-1"};
char glplot_font_medium[128] =
  {"-adobe-times-medium-r-normal--14-140-75-75-p-74-iso8859-1"};
char glplot_font_large[128] =
  {"-adobe-times-bold-r-normal--25-180-100-100-p-132-iso8859-1"};
char glplot_font_large1[128] =
  {"-adobe-times-bold-r-normal--34-240-100-100-p-177-iso8859-1"};
static float gxmin,gxmax,gymin,gymax,gyf,gxf = 0.0;
void glplot_get_win_loc_size(dpy,win,rx,ry,rw,rh)
     Display *dpy;
     Window win;
     int *rx,*ry,*rw,*rh;
{
  unsigned int valmask;
  XWindowAttributes xwa;
  XGetWindowAttributes(dpy,win,&xwa);
  *rx = xwa.x;
  *ry = xwa.y;
  *rw = xwa.width;
  *rh = xwa.height;
}
void glplot_set_window_loc(dpy,win,x,y)
     Display *dpy;
     Window win;
     int x,y;
{
  unsigned int valmask;
  XWindowChanges xwinchange;
  xwinchange.x = x;
  xwinchange.y = y;
  valmask = CWX | CWY;
  XConfigureWindow(dpy,win,valmask,&xwinchange);
}
void glplot_set_window_size(dpy,win,w,h)
     Display *dpy;
     Window win;
     int w,h;
{
  unsigned int valmask;
  XWindowChanges xwinchange;
  xwinchange.width = w;
  xwinchange.height = h;
  valmask = CWWidth | CWHeight;
  XConfigureWindow(dpy,win,valmask,&xwinchange);
}
void glplot_set_window_name(dpy,win,name)
     Display *dpy;
     Window win;
     char name[];
{
  int targc;
  char **targv;
  targv = (char **)myalloc(1*sizeof(char *));
  targv[0] = strdup("glplot0");
  targc = 0; 
  XSetStandardProperties(dpy,win,name,"glplot2",None,targv,targc,NULL);
  myfree(targv[0]);
  myfree(targv);
}
void glplot_set_window_focus(dpy,win)
     Display *dpy;
     Window win;
{
  XSetInputFocus(dpy,win,RevertToParent,CurrentTime);
}
unsigned char *convert_2d_to_gl_format(data,xn,yn,min,max)
     float **data;
     int xn,yn;
     float min,max;
{
  int i,j;
  unsigned char *d;
  float range,x,t;
  if (min == max){
    min = max = data[0][0];
    for(i=0;i<xn;i++){
      for(j=0;j<yn;j++){
	t = data[i][j];
	if (t > max)
	  max = t;
	else if (t < min)
	  min = t;
      }
    }
  }
  range = max - min;
  d = (unsigned char *)myalloc(xn*yn*sizeof(unsigned char));
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      t = data[i][j];
      x = (t - min)/range * 127.0;
      d[j*xn+i] = (unsigned char)x;
    }
  }
  return d;
}
unsigned char **convert_stim_to_gl_format(data,xn,yn,tn,min,max)
     float ***data;
     int xn,yn,tn;
     float min,max;
{
  int i,j,k;
  unsigned char **d;
  float range,x;
  range = max - min;
  d = (unsigned char **)myalloc(tn*sizeof(unsigned char *));
  for(k=0;k<tn;k++){
    d[k] = (unsigned char *)myalloc(xn*yn*sizeof(unsigned char));
    for(i=0;i<xn;i++){
      for(j=0;j<yn;j++){
	x = (data[i+1][j+1][k+1] - min)/range * 127.0;
	d[k][j*xn+i] = (unsigned char)x;
      }
    }
  }
  return d;
}
unsigned char **convert_data_to_gl_format(data,xn,yn,tn,nbytes,tcode,printflag,
					  min,max)
     void ***data;
     int xn,yn,tn,nbytes,tcode,printflag;
     float min,max;
{
  int i,j,k;
  int *t4;
  unsigned char **d,*t1;
  short *t2;
  float range,x,*t4f;
  if (min == max){
    if (nbytes == 1){
      t1 = data[0][0];         
      min = max = (float)t1[0];
      for(i=0;i<xn;i++){
	for(j=0;j<yn;j++){
	  t1 = data[i][j];
	  for(k=0;k<tn;k++){
	    if ((float)t1[k] > max)
	      max = (float)t1[k];
	    else if ((float)t1[k] < min)
	      min = (float)t1[k];
	  }
	}
      }
    }else if (nbytes == 2){
      t2 = data[0][0];
      min = max = (float)t2[0];
      for(i=0;i<xn;i++){
	for(j=0;j<yn;j++){
	  t2 = data[i][j];
	  for(k=0;k<tn;k++){
	    if ((float)t2[k] > max)
	      max = (float)t2[k];
	    else if ((float)t2[k] < min)
	      min = (float)t2[k];
	  }
	}
      }
    }else if (nbytes == 4){
      if (tcode == 1){ 
	t4 = data[0][0];
	min = max = (float)t4[0];
	for(i=0;i<xn;i++){
	  for(j=0;j<yn;j++){
	    t4 = data[i][j];
	    for(k=0;k<tn;k++){
	      if ((float)t4[k] > max)
		max = (float)t4[k];
	      else if ((float)t4[k] < min)
		min = (float)t4[k];
	    }
	  }
	}
      }else{ 
	t4f = data[0][0];
	min = max = t4f[0];
	for(i=0;i<xn;i++){
	  for(j=0;j<yn;j++){
	    t4f = data[i][j];
	    for(k=0;k<tn;k++){
	      if (t4f[k] > max)
		max = t4f[k];
	      else if (t4f[k] < min)
		min = t4f[k];
	    }
	  }
	}
      }
    }
  }
  if (printflag)
    printf("    min = %f  max = %f\n",min,max);
  range = max - min;
  if (nbytes == 1){
    d = (unsigned char **)myalloc(tn*sizeof(unsigned char *));
    for(k=0;k<tn;k++){
      d[k] = (unsigned char *)myalloc(xn*yn*sizeof(unsigned char));
      for(i=0;i<xn;i++){
	for(j=0;j<yn;j++){
  	  t1 = data[i][j];
  	  x = ((float)t1[k] - min)/range * 127.0;
	  d[k][j*xn+i] = (unsigned char)x;
	}
      }
    }
  }else if (nbytes == 2){
    d = (unsigned char **)myalloc(tn*sizeof(unsigned char *));
    for(k=0;k<tn;k++){
      d[k] = (unsigned char *)myalloc(xn*yn*sizeof(unsigned char));
      for(i=0;i<xn;i++){
	for(j=0;j<yn;j++){
  	  t2 = data[i][j];
	  x = ((float)t2[k] - min)/range * 127.0;
	  d[k][j*xn+i] = (unsigned char)x;
	}
      }
    }
  }else if (nbytes == 4){
    if (tcode == 1){ 
      d = (unsigned char **)myalloc(tn*sizeof(unsigned char *));
      for(k=0;k<tn;k++){
	d[k] = (unsigned char *)myalloc(xn*yn*sizeof(unsigned char));
	for(i=0;i<xn;i++){
	  for(j=0;j<yn;j++){
  	    t4 = data[i][j];
	    x = ((float)t4[k] - min)/range * 127.0;
	    d[k][j*xn+i] = (unsigned char)x;
	  }
	}
      }
    }else{ 
      d = (unsigned char **)myalloc(tn*sizeof(unsigned char *));
      for(k=0;k<tn;k++){
	d[k] = (unsigned char *)myalloc(xn*yn*sizeof(unsigned char));
	for(i=0;i<xn;i++){
	  for(j=0;j<yn;j++){
	    t4f = data[i][j];
	    x = (t4f[k] - min)/range * 127.0;
	    d[k][j*xn+i] = (unsigned char)x;
	  }
	}
      }
    }
  }else{
    d = NULL;
    exit_error("CONVERT_DATA_TO_GL_FORMAT","nbytes error");
  }
  return d;
}
GLubyte **convert_data_to_glubyte_format(data,xn,yn,tn,nbytes,tcode,normflag,
					       printflag)
     void ***data;
     int xn,yn,tn,nbytes,tcode;
     int normflag;
     int printflag;
{
  int i,j,k;
  int *t4;
  unsigned char *t1;
  GLubyte **d,td;
  short *t2;
  float min,max,range,x,*t4f;
  if (printflag){
    printf("    nbytes = %d  tcode = %d\n",nbytes,tcode);
    printf("    xn,yn,tn =  %d %d %d\n",xn,yn,tn);
  }
  if (nbytes == 1){
    exit_error("CONVERT_DATA_TO_GLUBYTE_FORMAT","Not fixed yet");
  }else if (nbytes == 2){
    exit_error("CONVERT_DATA_TO_GLUBYTE_FORMAT","Not fixed yet");
  }else if (nbytes == 4){
    if (tcode == 1){ 
      exit_error("CONVERT_DATA_TO_GLUBYTE_FORMAT","Not fixed yet");
    }else{ 
      t4f = data[0][0];
      min = max = t4f[0];
      for(i=0;i<xn;i++){
	for(j=0;j<yn;j++){
	  t4f = data[i][j];
	  for(k=0;k<tn;k++){
	    if (t4f[k] > max)
	      max = t4f[k];
	    else if (t4f[k] < min)
	      min = t4f[k];
	  }
	}
      }
    }
  }else{
    min = max = 0.0;
    exit_error("CONVERT_DATA_TO_GLUBYTE_FORMAT","nbytes error");
  }
  if (printflag)
    printf("    min = %f  max = %f\n",min,max);
  range = max - min;
  if (nbytes == 1){
    exit_error("CONVERT_DATA_TO_GLUBYTE_FORMAT","Not fixed yet");
  }else if (nbytes == 2){
    exit_error("CONVERT_DATA_TO_GLUBYTE_FORMAT","Not fixed yet");
  }else if (nbytes == 4){
    if (tcode == 1){ 
      exit_error("CONVERT_DATA_TO_GLUBYTE_FORMAT","Not fixed yet");
    }else{ 
      d = (GLubyte **)myalloc(tn*sizeof(GLubyte *));
      for(i=0;i<tn;i++){
	d[i] = (GLubyte *)myalloc(xn*yn*sizeof(GLubyte));
	for(j=0;j<xn;j++){
	  for(k=0;k<yn;k++){
	    t4f = data[k][j];  
	    if (normflag)
	      x = (t4f[i] - min)/range * 255.0;
	    else
	      x = t4f[i] * 255.0;  
	    td = (GLubyte)x;
	    d[i][j*yn+k] = td;
	  }
	}
      }
    }
  }else{
    d = NULL;
    exit_error("CONVERT_DATA_TO_GLUBYTE_FORMAT","nbytes error");
  }
  return d;
}
void glplot_find_delta(xrange,rd,rp)
     float xrange;
     int *rd,*rp;
{
  int d,p;
  float x;
  if (xrange <= 0.0)
    exit_error("GLPLOT_FIND_DELTA","xrange is <= 0");
  x = xrange/10.0;
  p = 0;
  while(x > 10.0){
    x /= 10.0;
    p += 1;
  }
  while(x <= 1.0){
    x *= 10.0;
    p -= 1;
  }
  if (x < 1.0)
    d = 1;
  else if (x < 2.0)
    d = 2;
  else if (x < 5.0)
    d = 5;
  else{
    d = 1;
    p += 1;
  }
  *rd = d; *rp = p;
}
float glplot_find_delta_start(delta,xmin,xmax)
     float delta,xmin,xmax;
{
  int k;
  float x0;
  k = (int)(xmin/delta);
  x0 = (float)k * delta;
  return x0;
}
void glplot_draw_line(dpy,win,doubleBuffer)
     Display *dpy;
     Window win;
     Bool doubleBuffer;
{
  static Bool displayListInited = False;
  if (displayListInited){
    glCallList(1);
  }else{
    glNewList(1,GL_COMPILE_AND_EXECUTE);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLineWidth(5.0);
    glBegin(GL_LINE_STRIP);
    glColor3f(0.5,0.2,0.3); 
    glVertex2f(1.0,1.0);
    glVertex2f(-1.0,1.0);
    glVertex2f(-1.0,-1.0);
    glVertex2f(1.0,-1.0);
    glVertex2f(1.0,1.0);
    glEnd();
    glEndList();
  }
  if (doubleBuffer) 
    glXSwapBuffers(dpy,win);
  else
    glFlush();
}
void glplot_draw_box(r,g,b,x,y,w,h)
     float r,g,b;      
     float x,y,w,h;    
{
  glBegin(GL_POLYGON);
  glColor3f(r,g,b);
  glVertex2f(x,y);
  glVertex2f(x+w,y);
  glVertex2f(x+w,y+h);
  glVertex2f(x,y+h);
  glEnd();
}
void glplot_draw_disk(r,g,b,x,y,dx,dy)
     float r,g,b;      
     float x,y,dx,dy;  
{
  int i;
  int ns;
  float tx,ty;
  ns = 12; 
  glBegin(GL_POLYGON);
  glColor3f(r,g,b);
  for(i=0;i<ns;i++){
    tx = x+dx/2.0 + dx/2.0 * cos((float)i/(float)ns * 2.0 * M_PI);
    ty = y+dy/2.0 + dy/2.0 * sin((float)i/(float)ns * 2.0 * M_PI);
    glVertex2f(tx,ty);
  }
  glEnd();
}
void glplot_set_xfont_generic(dpy,win,cx,fsize)
     Display *dpy;
     Window win;
     GLXContext cx;
     char fsize[];
{
  XFontStruct *font;
  Font fid;
  glXMakeCurrent(dpy,win,cx);
  if (strcmp(fsize,"small")==0)
    font = XLoadQueryFont(dpy,glplot_font_small);
  else if (strcmp(fsize,"medium")==0)
    font = XLoadQueryFont(dpy,glplot_font_medium);
  else if (strcmp(fsize,"large")==0)
    font = XLoadQueryFont(dpy,glplot_font_large);
  else if (strcmp(fsize,"large1")==0)
    font = XLoadQueryFont(dpy,glplot_font_large1);
  else{
    font = NULL;
    exit_error("GLPLOT_SET_XFONT_GENERIC","Unknown font size");
  }
  fid = font->fid;
  glXUseXFont(fid,32,96,2000+32);
}
XFontStruct *glplot_set_xfont_generic_return(dpy,win,cx,fsize)
     Display *dpy;
     Window win;
     GLXContext cx;
     char fsize[];
{
  XFontStruct *font;
  Font fid;
  glXMakeCurrent(dpy,win,cx);
  if (strcmp(fsize,"small")==0)
    font = XLoadQueryFont(dpy,glplot_font_small);
  else if (strcmp(fsize,"medium")==0)
    font = XLoadQueryFont(dpy,glplot_font_medium);
  else if (strcmp(fsize,"large")==0)
    font = XLoadQueryFont(dpy,glplot_font_large);
  else if (strcmp(fsize,"large1")==0)
    font = XLoadQueryFont(dpy,glplot_font_large1);
  else{
    font = NULL;
    exit_error("GLPLOT_SET_XFONT_GENERIC","Unknown font size");
  }
  fid = font->fid;
  glXUseXFont(fid,32,96,2000+32);
  return font;
}
void glplot_set_xfont(dpy,win,cx,fontname)
     Display *dpy;
     Window win;
     GLXContext cx;
     char *fontname;
{
  XFontStruct *font;
  Font fid;
  font = XLoadQueryFont(dpy,fontname);
  fid = font->fid;
  glXMakeCurrent(dpy,win,cx);
  glXUseXFont(fid,32,96,2000+32);
}
void glplot_use_font(font)
     XFontStruct *font;
{
  Font fid;
  fid = font->fid;
  glXUseXFont(fid,32,96,2000+32);
}
float glplot_text_len_pix(font,s)
     XFontStruct *font;
     char *s;
{
  float pix;
  pix = XTextWidth(font,s,strlen(s)); 
  return pix;
}
void glplot_draw_txt(tstr,x,y,r,g,b)
     char tstr[];
     float x,y,r,g,b;
{
  glColor3f(r,g,b);
  glListBase(2000);
  glRasterPos2f(x,y);
  glCallLists(strlen(tstr),GL_UNSIGNED_BYTE,tstr);
}
void glplot_draw_text(dpy,win,cx,tstr,x,y,r,g,b)
     Display *dpy;
     Window win;
     GLXContext cx;
     char tstr[];
     float x,y,r,g,b;
{
  glColor3f(r,g,b);
  glListBase(2000);
  glRasterPos2f(x,y);
  glCallLists(strlen(tstr),GL_UNSIGNED_BYTE,tstr);
}
void glplot_draw_text_boxes(dpy,win,cx,slist,ns,r,g,b,x0,y0,w,h,frac,aflag)
     Display *dpy;
     Window win;
     GLXContext cx;
     char **slist;     
     int ns;
     float *r,*g,*b;   
     float x0,y0,w,h;  
     float frac;       
     int aflag;        
{
  int i;
  float x,y,dx,dy;
  x = x0;
  y = y0;
  dx = w*frac;
  dy = h*frac;
  for(i=0;i<ns;i++){
    glplot_draw_box(r[i],g[i],b[i],x,y,dx,dy);
    glClear(GL_DEPTH_BUFFER_BIT);
    if ((r[i] + g[i] + b[i]) < 1.5)
      glplot_draw_txt(slist[i],x+0.1*w,y+0.2*h,1.0,1.0,1.0);
    else
      glplot_draw_txt(slist[i],x+0.1*w,y+0.2*h,0.0,0.0,0.0);
    x += w;
  }
}
void glplot_draw_simple_axes(dpy,win,cx,dbflag,flflag,x0,xn,y0,yn,x0p,xnp,
			     y0p,ynp,r,g,b,linewidth,xorigin,yorigin)
     Display *dpy;
     Window win;
     GLXContext cx;
     int dbflag,flflag; 
     float x0,xn,y0,yn,x0p,xnp,y0p,ynp,r,g,b,linewidth,xorigin,yorigin;
{
  float x,y;
  glNewList(1,GL_COMPILE_AND_EXECUTE);
  glLineWidth(linewidth);
  glBegin(GL_LINE_STRIP);
  glColor3f(r,g,b);
  x = x0p;
  y = y0p + ynp * (yorigin - y0)/yn;
  glVertex2f(x,y);
  x = x0p + xnp;
  glVertex2f(x,y);
  glEnd();
  glBegin(GL_LINE_STRIP);
  glColor3f(r,g,b);
  x = x0p + xnp * (xorigin - x0)/xn;
  y = y0p;
  glVertex2f(x,y);
  y = y0p + ynp;
  glVertex2f(x,y);
  glEnd();
  glEndList();
  if (flflag == 1){
    if (dbflag == 1)
      glXSwapBuffers(dpy,win);
    else
      glFlush();
  }
}
void glplot_get_win_coord_plot_auto(winw,winh,x,y,rx,ry)
     int winw,winh;      
     float x,y;          
     float *rx,*ry;      
{
  float fx,fy;  
  float px,py;  
  fx = (float)x/(float)winw;
  px = gxmin + fx * (gxmax-gxmin);
  fy = 1.0 - (float)y/(float)winh;
  py = gymin + fy * (gymax-gymin);
  *rx = px; *ry = py;
}
void glplot_draw_plot_line_auto(dpy,win,cx,aflag,dbflag,clrflag,xdata,ydata,n,
				linewidth,r,g,b)
     Display *dpy;
     Window win;
     GLXContext cx;
     int aflag;   
     int dbflag;  
     int clrflag; 
     float *xdata,*ydata;
     int n;
     float linewidth,r,g,b;
{
  int i;
  int dp,dd;
  float x,y,z,d0,delta;
  char tstr[SLEN];
  if (aflag || (gxf == 0)){ 
    get_min_max_farray(xdata,n,&gxmin,&gxmax);
    get_min_max_farray(ydata,n,&gymin,&gymax);
    gxf = (gxmax-gxmin);
    gyf = (gymax-gymin);
  }
  if (clrflag)
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glplot_find_delta(gxmax-gxmin,&dd,&dp);
  delta = dd * pow(10.0,(float)dp);
  d0 = glplot_find_delta_start(delta,gxmin,gxmax);
  glLineWidth(1.0); 
  glColor3f(0.5,0.5,0.5);
  glBegin(GL_LINES);
  z = d0;
  while(z <= (gxmax + gxf*0.005)){
    x = -1.0 + 2*(z-gxmin)/gxf;
    glVertex2f(x,-1.0);
    glVertex2f(x,1.0);
    z += delta;
  }
  glEnd();
  z = d0;
  while(z <= (gxmax + gxf*0.005)){
    x = -1.0 + 2*(z-gxmin)/gxf;
    sprintf(tstr,"%d",(int)z);
    glplot_draw_text(dpy,win,cx,tstr,x+0.005,-0.96,1.0,1.0,1.0);
    z += delta;
  }
  sprintf(tstr,"hscale %.2f",delta);
  glplot_draw_text(dpy,win,cx,tstr,-0.95,0.9,1.0,1.0,1.0);
  glLineWidth(linewidth);
  glColor3f(r,g,b);
  glBegin(GL_LINE_STRIP);
  for(i=0;i<n;i++){
    x = (xdata[i] - gxmin)/gxf;
    y = (ydata[i] - gymin)/gyf;
    x = 2.0*x - 1.0;
    y = 2.0*y - 1.0;
    glVertex2f(x,y);
  }
  glEnd();
  if (dbflag == 1) 
    glXSwapBuffers(dpy,win);
  else
    glFlush();
}
void glplot_draw_hline(dpy,win,cx,dbflag,xa,xb,yy,
		       x0,xn,y0,yn,x0p,xnp,y0p,ynp,r,g,b,linewidth)
     Display *dpy;
     Window win;
     GLXContext cx;
     int dbflag; 
     float xa,xb,yy;
     float x0,xn,y0,yn,x0p,xnp,y0p,ynp,r,g,b,linewidth;
{
  float x,y;
  glNewList(1,GL_COMPILE_AND_EXECUTE);
  glLineWidth(linewidth);
  glBegin(GL_LINE_STRIP);
  glColor3f(r,g,b); 
  x = x0p + xnp * (xa - x0)/xn;
  y = y0p + ynp * (yy - y0)/yn;
  glVertex2f(x,y);
  x = x0p + xnp * (xb - x0)/xn;
  glVertex2f(x,y);
  glEnd();
  glEndList();
  if (dbflag == 1) 
    glXSwapBuffers(dpy,win);
  else
    glFlush();
}
void glplot_draw_plot_line(dpy,win,cx,dbflag,xdata,ydata,n,
			   x0,xn,y0,yn,x0p,xnp,y0p,ynp,r,g,b,linewidth)
     Display *dpy;
     Window win;
     GLXContext cx;
     int dbflag;   
     float *xdata,*ydata;
     int n;
     float x0,xn,y0,yn,x0p,xnp,y0p,ynp,r,g,b,linewidth;
{
  int i;
  float x,y;
  glNewList(1,GL_COMPILE_AND_EXECUTE);
  glLineWidth(linewidth);
  glBegin(GL_LINE_STRIP);
  glColor3f(r,g,b); 
  for(i=0;i<n;i++){
    x = x0p + xnp * (xdata[i] - x0)/xn;
    y = y0p + ynp * (ydata[i] - y0)/yn;
    glVertex2f(x,y);
  }
  glEnd();
  glEndList();
  if (dbflag == 1) 
    glXSwapBuffers(dpy,win);
  else
    glFlush();
}
void glplot_xevent(dpy,win,cx,dbflag)
     Display *dpy;
     Window win;
     GLXContext cx;
     int dbflag; 
{
  XEvent event;
  while (XPending(dpy)){
    XNextEvent(dpy,&event); 
    switch(event.type){
    case ButtonPress:
      break;
    case ConfigureNotify:
      printf("CONFIG\n");
      glViewport(0,0,event.xconfigure.width,event.xconfigure.height);
    case Expose:  
      glCallList(1);
      if (dbflag == 1) 
	glXSwapBuffers(dpy,win);
      else
	glFlush();
      break;
    }
  }
}
void glplot_init_win(w,h,zoom,wintitle,dbflag,rdpy,rwin,rctx,rdbflag,motflag,
		     mylogf)
     int w,h;
     float zoom;
     char wintitle[];
     int dbflag;           
     Display **rdpy;
     Window *rwin;
     GLXContext *rctx;
     int *rdbflag;
     int motflag;    
     char *mylogf;   
{
  static int snglBuf[] = {GLX_RGBA,GLX_RED_SIZE,1,GLX_GREEN_SIZE,1,
			  GLX_BLUE_SIZE,1,GLX_DEPTH_SIZE,12,None};
  static int dblBuf[]  = {GLX_RGBA,GLX_RED_SIZE,1,GLX_GREEN_SIZE,1,
			  GLX_BLUE_SIZE,1,GLX_DEPTH_SIZE,12,GLX_DOUBLEBUFFER,
			  None};
  Bool doubleBuffer = True;
  Display *dpy;
  Window win;
  XVisualInfo *vi;
  GLXContext ctx;
  Colormap cmap;
  XSetWindowAttributes swa;
  XEvent event;
  int dummy,targc;
  char **targv;
  int pflag;
  mylog(mylogf,"  GLPLOT_INIT_WIN\n");
  pflag = 0;
  dpy = XOpenDisplay(NULL); 
  if (dpy == NULL)
    mylog_exit(mylogf,"GLPLOT_INIT_WIN  Could not open display");
  if (!glXQueryExtension(dpy,&dummy,&dummy)) 
    mylog_exit(mylogf,"GLPLOT_INIT_WIN  X server has no OpenGL GLX entension");
  if (pflag) printf("HERE WYETH ....glplot_init_win.... 01\n");
  if (dbflag == 1)
    vi = glXChooseVisual(dpy,DefaultScreen(dpy),dblBuf);
  if ((vi == NULL) || (dbflag == 0)){
    vi = glXChooseVisual(dpy,DefaultScreen(dpy),snglBuf);
    if (vi == NULL)
      mylog_exit(mylogf,"GLPLOT_INIT_WIN  No RGB visual with depth buffer");
    doubleBuffer = False;
    mylog(mylogf,"    doubleBuffer is FALSE\n");
  }
  if (pflag) printf("HERE WYETH ....glplot_init_win.... 02\n");
  if (vi->class != TrueColor)
    mylog(mylogf,"TrueColor visual required for this program\n");
  ctx = glXCreateContext(dpy,vi,None,True);
  if (ctx == NULL)
    mylog(mylogf,"could not create rendering context\n");
  cmap = XCreateColormap(dpy,RootWindow(dpy,vi->screen),vi->visual,AllocNone);
  swa.colormap = cmap;
  swa.border_pixel = 0;
  if (motflag == 1){
    swa.event_mask = ExposureMask | ButtonPressMask | ButtonReleaseMask | 
      StructureNotifyMask | PointerMotionMask | KeyPressMask |
      Button1MotionMask;
  }else{
    swa.event_mask = ExposureMask | ButtonPressMask | ButtonReleaseMask | 
      StructureNotifyMask | KeyPressMask;
  }
  if (pflag) printf("HERE WYETH ....glplot_init_win.... 03\n");
  win = XCreateWindow(dpy,RootWindow(dpy,vi->screen),0,0,w,h,0,vi->depth,
		      InputOutput,vi->visual,
		      CWBorderPixel | CWColormap | CWEventMask,&swa);
  if (pflag) printf("*** win = %d\n",(int)win);
  targv = (char **)myalloc(1*sizeof(char *));
  targv[0] = strdup("glplot0");
  targc = 0; 
  XSetStandardProperties(dpy,win,wintitle,"glplot2",None,targv,targc,NULL);
  myfree(targv[0]); myfree(targv);
  if (pflag) printf("HERE WYETH ....glplot_init_win.... 04\n");
  glXMakeCurrent(dpy,win,ctx);  
  if (pflag) printf("HERE WYETH ....glplot_init_win.... 05\n");
  XMapWindow(dpy,win);          
  if (pflag) printf("HERE WYETH ....glplot_init_win.... 06\n");
  *rdpy = dpy; *rwin = win; *rctx = ctx; *rdbflag = (int)doubleBuffer;
}
void glplot_init(w,h,zoom,wintitle,rdpy,rwin,rctx,rdbflag,motflag)
     int w,h;
     float zoom;
     char wintitle[];
     Display **rdpy;
     Window *rwin;
     GLXContext *rctx;
     int *rdbflag;
     int motflag;    
{
  Display *dpy;
  Window win;
  GLXContext ctx;
  XEvent event;
  int dbflag;
  printf("  GLPLOT_INIT\n");
  printf("    w,h,z  %d %d %.2f  title: %s   motflag: %d\n",w,h,zoom,wintitle,
	 motflag);
  glplot_init_win(w,h,zoom,wintitle,1,&dpy,&win,&ctx,&dbflag,motflag,NULL);
  glPixelZoom(zoom,zoom);
  glEnable(GL_DEPTH_TEST); 
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho((GLdouble)-1.0,(GLdouble)1.0,(GLdouble)-1.0,(GLdouble)1.0,-1.0,1.0);
  *rdpy = dpy; *rwin = win; *rctx = ctx; *rdbflag = dbflag;
}
void glplot_init_sing(w,h,zoom,wintitle,rdpy,rwin,rctx,motflag)
     int w,h;
     float zoom;
     char wintitle[];
     Display **rdpy;
     Window *rwin;
     GLXContext *rctx;
     int motflag;    
{
  Display *dpy;
  Window win;
  GLXContext ctx;
  XEvent event;
  int dbflag;
  printf("  GLPLOT_INIT_SING\n");
  glplot_init_win(w,h,zoom,wintitle,0,&dpy,&win,&ctx,&dbflag,motflag,NULL);
  glPixelZoom(zoom,zoom);
  glEnable(GL_DEPTH_TEST); 
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho((GLdouble)-1.0,(GLdouble)1.0,(GLdouble)-1.0,(GLdouble)1.0,-1.0,1.0);
  *rdpy = dpy; *rwin = win; *rctx = ctx;
}
void glplot_init1(w,h,zoom,wintitle,rdpy,rwin,rctx,rdbflag,motflag,mylogf)
     int w,h;
     float zoom;
     char wintitle[];
     Display **rdpy;
     Window *rwin;
     GLXContext *rctx;
     int *rdbflag;     
     int motflag;      
     char *mylogf;     
{
  Display *dpy;
  Window win;
  GLXContext ctx;
  XEvent event;
  int dbflag;
  mylog(mylogf,"  GLPLOT_INIT1\n");
  glplot_init_win(w,h,zoom,wintitle,*rdbflag,&dpy,&win,&ctx,&dbflag,motflag,
		  mylogf);
  glPixelZoom(zoom,zoom);
  glEnable(GL_DEPTH_TEST); 
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0,w,0,h,-1.0,1.0);
  *rdpy = dpy;
  *rwin = win;
  *rctx = ctx;
  *rdbflag = dbflag;
}
void glplot_util_make_current(dpy,win,cx)
     Display *dpy;
     Window win;
     GLXContext cx;
{
  glXMakeCurrent(dpy,win,cx);
}
void glplot_init_pos_pix(w,h,rdpy,rwin,rcx,rdbflag)
     int w,h;
     Display **rdpy;
     Window *rwin;
     GLXContext *rcx;
     int *rdbflag;
{
  static int snglBuf[] = {GLX_RGBA,GLX_RED_SIZE,1,GLX_GREEN_SIZE,1,
			  GLX_BLUE_SIZE,1,GLX_DEPTH_SIZE,12,None};
  static int dblBuf[]  = {GLX_RGBA,GLX_RED_SIZE,1,GLX_GREEN_SIZE,1,
			  GLX_BLUE_SIZE,1,GLX_DEPTH_SIZE,12,GLX_DOUBLEBUFFER,
			  None};
  Bool doubleBuffer = True;
  Bool recalcModelView = True;
  Display *dpy;
  Window win;
  XVisualInfo *vi;
  GLXContext cx;
  Colormap cmap;
  XSetWindowAttributes swa;
  XEvent event;
  int dummy,targc;
  char **targv;
  dpy = XOpenDisplay(NULL); 
  if (dpy == NULL)
    exit_error("GLPLOT_INIT","Could not open display");
  if (!glXQueryExtension(dpy,&dummy,&dummy)) 
    exit_error("GLPLOT_INIT","X server has no OpenGL GLX entension");
  vi = glXChooseVisual(dpy,DefaultScreen(dpy),dblBuf);
  if (vi == NULL){
    vi = glXChooseVisual(dpy,DefaultScreen(dpy),snglBuf);
    if (vi == NULL)
      exit_error("GLPLOT_INIT","no RGB visual with depth buffer");
    doubleBuffer = False;
  }
  if (vi->class != TrueColor)
    printf("TrueColor visual required for this program\n");
  cx = glXCreateContext(dpy,vi,None,True);
  if (cx == NULL)
    printf("could not create rendering context\n");
  cmap = XCreateColormap(dpy,RootWindow(dpy,vi->screen),vi->visual,AllocNone);
  swa.colormap = cmap;
  swa.border_pixel = 0;
  swa.event_mask = ExposureMask | ButtonPressMask | StructureNotifyMask;
  win = XCreateWindow(dpy,RootWindow(dpy,vi->screen),0,0,w,h,0,vi->depth,
		      InputOutput,vi->visual,
		      CWBorderPixel | CWColormap | CWEventMask,&swa);
  targv = (char **)myalloc(1*sizeof(char *));
  targv[0] = strdup("glplot");
  targc = 0; 
  XSetStandardProperties(dpy,win,"glplot","glplot",None,targv,targc,NULL);
  glXMakeCurrent(dpy,win,cx);   
  XMapWindow(dpy,win);   
  glEnable(GL_DEPTH_TEST); 
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0,(float)w,0.0,(float)h,1.0,10.0);
  do{
    XNextEvent(dpy,&event);
    switch(event.type){
    case ButtonPress:
      break;
    case ConfigureNotify:
      glViewport(0,0,event.xconfigure.width,event.xconfigure.height);
    case Expose:  
      break;
    }
  } while (XPending(dpy)); 
  if (recalcModelView){
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef(0.0,0.0,-1.1); 
    recalcModelView = False;
  }
  glplot_draw_line(dpy,win,doubleBuffer);
  *rdpy = dpy; *rwin = win; *rcx = cx; *rdbflag = (int)doubleBuffer;
}
void glplot_shutdown(dpy,win,ctx,dpyflag)
     Display *dpy;
     Window win;
     GLXContext ctx;
     int dpyflag;
{
  static int ndest_cx = 0;
  XDestroyWindow(dpy,win);
  if ((dpyflag) && (ndest_cx < 1)) {
    glXDestroyContext(dpy,ctx);
    ndest_cx += 1;
  }
  XCloseDisplay(dpy);
}
void glplot_clear(dpy,win)
     Display *dpy;
     Window win;
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}
void glplot_clear_double(dpy,win)
     Display *dpy;
     Window win;
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glXSwapBuffers(dpy,win);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glXSwapBuffers(dpy,win);
}
void glplot_show_sequence_pick_value(data,m,n,x0,x1,y0,y1,ndelay,rval)
     float **data;
     int m,n;
     float x0,x1,y0,y1;
     float *rval;
{
  int i,j;
  Display *dpy;
  Window win;
  GLXContext cx;
  int dbflag,ns;
  float v0,val,*xdata,zoom;
  if (ndelay < 0)
    ndelay = 0;
  zoom = 1.0;
  glplot_init(1200,400,zoom,"glplot_show_sequence_pick_values",
	      &dpy,&win,&cx,&dbflag,1);
  glplot_clear_double(dpy,win);
  xdata = get_sequence_farray(0.0,1.0,n);
  val = *rval;
  i=0;
  glplot_clear(dpy,win);
  glplot_draw_hline(dpy,win,cx,0,x0,x1,val,
		    x0,x1-x0,y0,y1-y0,-1.0,2.0,-1.0,2.0,
		    1.0,0.0,0.0,1.0);
  glplot_draw_plot_line(dpy,win,cx,dbflag,xdata,data[i],n,
			x0,x1-x0,y0,y1-y0,-1.0,2.0,-1.0,2.0,
			0.0,1.0,0.0,1.0);
  v0 = 1.0;
  while (v0 != 0.0){
    printf("Enter value (%.4f, 0-exit):  ",val); fflush(stdout);
    ns = scanf("%f",&v0);
    if (v0 != 0.0){
      val = v0;
      for(i=0;i<m;i++){
	glplot_clear(dpy,win);
	glplot_draw_hline(dpy,win,cx,0,x0,x1,val,
			  x0,x1-x0,y0,y1-y0,-1.0,2.0,-1.0,2.0,
			  1.0,0.0,0.0,1.0);
	glplot_draw_plot_line(dpy,win,cx,dbflag,xdata,data[i],n,
			      x0,x1-x0,y0,y1-y0,-1.0,2.0,-1.0,2.0,
			      0.0,1.0,0.0,1.0);
	glplot_clear(dpy,win);
	glplot_draw_hline(dpy,win,cx,0,x0,x1,val,
			  x0,x1-x0,y0,y1-y0,-1.0,2.0,-1.0,2.0,
			  1.0,0.0,0.0,1.0);
	glplot_draw_plot_line(dpy,win,cx,dbflag,xdata,data[i],n,
			      x0,x1-x0,y0,y1-y0,-1.0,2.0,-1.0,2.0,
			      0.0,1.0,0.0,1.0);
	for(j=0;j<ndelay;j++)
	  glXSwapBuffers(dpy,win); 
      }
    }
  }
  myfree(xdata);
  glplot_shutdown(dpy,win,cx,1);
  *rval = val;
}
void glplot_open_show_2d(data,m,n,zoom,flag,min,max)
     float **data;
     int m,n;
     float zoom;
     int flag;
     float min,max; 
{
  int zm,zn;
  unsigned char *d;
  static Display *dpy = (Display *)NULL;
  static Window win;
  static GLXContext cx;
  static int dbflag;
  if (flag == 2){
    glplot_shutdown(dpy,win,cx,1);
    dpy = (Display *)NULL;
  }else if (flag == 0){
    d = convert_2d_to_gl_format(data,m,n,min,max);
    glClear(GL_DEPTH_BUFFER_BIT); 
    glDrawPixels(m,n,GL_LUMINANCE,GL_BYTE,d);
    glFlush();
    glXSwapBuffers(dpy,win);
    myfree(d);
  }else if (flag == 1){
    if (dpy != NULL){
      printf("HERE WYETH glplot_util shutdown 01\n");
      glplot_shutdown(dpy,win,cx,1);
      printf("HERE WYETH glplot_util shutdown 02\n");
      dpy = (Display *)NULL;
    }
    zm = (int)(0.5 + (float)m * zoom);
    zn = (int)(0.5 + (float)n * zoom);
    printf("HERE WYETH glplot_util shutdown 03\n");
    glplot_init(zm,zn,zoom,"glplot_open_show_2d",&dpy,&win,&cx,&dbflag,1);
    printf("HERE WYETH glplot_util shutdown 04\n");
    glplot_clear_double(dpy,win);
  }
}
