#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "my_util.h"
#include "carray_util.h"
#include "iarray_util.h"
#include "farray_util.h"
#include "myrand_util.h"
#include "noise_util.h"
#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)
/*C  (C) Copr. 1986-92 Numerical Recipes Software #.3.                       */
float grid_util_ran2(idum)
     int *idum; 
{
  int j;
  int k; 
  static int idum2=123456789; 
  static int iy=0; 
  static int iv[NTAB]; 
  float temp;
  if (*idum <= 0) {
    if (-(*idum) < 1) *idum=1;
    else *idum = -(*idum);
    idum2=(*idum);
    for (j=NTAB+7;j>=0;j--) {
      k=(*idum)/IQ1;
      *idum=IA1*(*idum-k*IQ1)-k*IR1;
      if (*idum < 0) *idum += IM1;
      if (j < NTAB) iv[j] = *idum;
    }
    iy=iv[0];
  }
  k=(*idum)/IQ1;
  *idum=IA1*(*idum-k*IQ1)-k*IR1;
  if (*idum < 0) *idum += IM1;
  k=idum2/IQ2;
  idum2=IA2*(idum2-k*IQ2)-k*IR2;
  if (idum2 < 0) idum2 += IM2;
  j=iy/NDIV;
  iy=iv[j]-idum2;
  iv[j] = *idum;
  if (iy < 1) iy += IMM1;
  if ((temp=AM*iy) > RNMX) return RNMX;
  else return temp;
}
#undef IM1
#undef IM2
#undef AM
#undef IMM1
#undef IA1
#undef IA2
#undef IQ1
#undef IQ2
#undef IR1
#undef IR2
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX
void grid_get_dgrid_coords(rat,nw,nh,w,h,transflag,rx,ry,rxn,ryn,rn)
     float rat;
     int nw,nh;                   
     int w,h;                     
     int transflag;               
     float **rx,**ry,**rxn,**ryn; 
     int *rn;
{
  int i,j,k,ii,jj;
  int n;
  float *x,*y,*xn,*yn;
  n = nw*nh;
  x = (float *)myalloc(n*sizeof(float));
  y = (float *)myalloc(n*sizeof(float));
  xn = (float *)myalloc(n*sizeof(float));
  yn = (float *)myalloc(n*sizeof(float));
  k = 0;
  for(i=0;i<nh;i++)
    for(j=0;j<nw;j++){
      if (transflag){ 
	jj = i;
	ii = j;
      }else{
	ii = i;
	jj = j;
      }
      x[k] = rat * (float)(jj*w);
      y[k] = rat *(float)(ii*h + h); 
      xn[k] = rat * (float)w;
      yn[k] = rat * (float)h;
      k += 1;
    }
  *rx = x; *ry = y; *rxn = xn; *ryn = yn; *rn = n;
}
void grid_get_dgrid_stim_for_params(seed,dist,dt,duration,framerate,ngrid,
				    dpbox,deriv_flag,rfdt,rxdata,rydata,
				    rnframes)
     int seed,dist,dt,duration,framerate,ngrid,dpbox,deriv_flag;
     float *rfdt,**rxdata,***rydata;
     int *rnframes;
{
  int i,j,k;
  int count,nframes;
  float *xdata,**ydata,fdt,*t;
  nframes = duration * framerate;
  printf(" ************ WYETH - DOES THIS GET OCTANE OR O2 dt????\n");
  fdt = get_sgi_actual_dt(framerate,0);
  if (seed > 0)
    seed = -seed;
  ydata = get_2d_farray(nframes,ngrid); 
  xdata = (float *)myalloc(nframes*sizeof(float));
  for(i=0;i<nframes;i++){
    xdata[i] = (double)i*fdt;
  }
  for(i=0;i<ngrid;i++){ 
    count = dt;
    for (j=0;j<nframes;j++){ 
      if (j==0){ 
	for(k=0;k<dpbox;k++){ 
	  (void)grid_util_ran2(&seed);
	  (void)grid_util_ran2(&seed);
	}
	ydata[j][i] = 0.0;
      }else{ 
	if (count == dt){
	  if (dist == 2){
	    if (grid_util_ran2(&seed) < 0.5){ 
	      ydata[j][i] = -1.0;
	    }else{ 
	      ydata[j][i] = 1.0;
	    }
	  }else
	    exit_error("GRID_GET_DGRID_STIM_FOR_PARAMS","Unkown distrib type");
	  count = 0;
	}else
	  ydata[j][i] = ydata[j-1][i];
	count += 1;
      }
    }
  }
  if (deriv_flag == 1){
    printf("  *** CHANGE = 1,  NO_CHANGE = -1\n");
    t = (float *)myalloc(nframes*sizeof(float));
    for(i=0;i<ngrid;i++){ 
      t[0] = 1.0; 
      for (j=1;j<nframes;j++){ 
	if (ydata[j][i] != ydata[j-1][i])
	  t[j] = 1.0; 
	else
	  t[j] = -1.0; 
      }
      for (j=0;j<nframes;j++) 
	ydata[j][i] = t[j];
    }
    myfree(t);
  }
  *rfdt = fdt; *rxdata = xdata; *rydata = ydata; *rnframes = nframes;
}	
void grid_get_00_index(rndx,rn)
     int **rndx,*rn;
{
  int i;
  int n,*ndx;
  n = 255;
  ndx = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++)
    ndx[i] = i;
  shuffle_iarray_nyu(ndx,n,2,1777);
  *rndx = ndx; *rn = n;
}
void grid_get_01_index(rndx,rn)
     int **rndx,*rn;
{
  int i;
  int n,*ndx;
  n = 228;
  ndx = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++)
    ndx[i] = i;  
  shuffle_iarray_nyu(ndx,n,2,1777);
  *rndx = ndx; *rn = n;
}
void grid_get_02_index(rndx,rn)
     int **rndx,*rn;
{
  int i;
  int n,*ndx;
  printf("WYETH - FIX THIS, must have 255 boxes.\n");
  n = 256;
  ndx = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++)
    ndx[i] = i;   
  shuffle_iarray_nyu(ndx+1,n-1,2,1777); 
  ndx[0] = ndx[208]; 
  ndx[208] = 0; 
  *rndx = ndx; *rn = n;
}
void grid_get_03_index(rndx,rn)
     int **rndx,*rn;
{
  int i;
  int n,*ndx;
  n = 244;
  ndx = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++)
    ndx[i] = i;  
  shuffle_iarray_nyu(ndx,n,2,1777);
  *rndx = ndx; *rn = n;
}
void grid_get_04_index(rndx,rn)
     int **rndx,*rn;
{
  int i;
  int n,*ndx;
  n = 249;
  ndx = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++)
    ndx[i] = i; 
  shuffle_iarray_nyu(ndx,n,2,1777);
  *rndx = ndx; *rn = n;
}
void grid_get_05_index(rndx,n)
     int **rndx,n;
{
  int i;
  int *ndx;
  if (n > 256)
    exit_error("GRID_GET_05_INDEX","n too high");
  else if (n==256)
    n = 255; 
  ndx = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++)
    ndx[i] = i; 
  shuffle_iarray_nyu(ndx,n,2,1777);
  *rndx = ndx;
}
void grid_get_00_coords(size,rx,ry,rxn,ryn,rn)
     float size;
     float **rx,**ry,**rxn,**ryn;
     int *rn;
{
  int i,j,k;
  int n;
  float *x,*y,*xn,*yn;
  n = 255;
  x = (float *)myalloc(n*sizeof(float));
  y = (float *)myalloc(n*sizeof(float));
  xn = (float *)myalloc(n*sizeof(float));
  yn = (float *)myalloc(n*sizeof(float));
  k = 0;
  for(i=0;i<16;i++){
    for(j=0;j<16;j++){
      if (!((i==0) && (j==0))){ 
	x[k] = (float)j * size/16.0;
	y[k] = (float)i * size/16.0;
	xn[k] = size/16.0;
	yn[k] = size/16.0;
	k += 1;
      }
    }
  }
  for(i=0;i<n;i++) 
    y[i] = size - y[i];
  *rx = x; *ry = y; *rxn = xn; *ryn = yn; *rn = n;
}
void grid_get_01_coords(size,rx,ry,rxn,ryn,rn)
     float size;
     float **rx,**ry,**rxn,**ryn;
     int *rn;
{
  int i,j,k;
  int n;
  float *x,*y,*xn,*yn;
  double s,tx0,ty0,tx,ty,txn,tyn;
  n = 228;
  x = (float *)myalloc(n*sizeof(float));
  y = (float *)myalloc(n*sizeof(float));
  xn = (float *)myalloc(n*sizeof(float));
  yn = (float *)myalloc(n*sizeof(float));
  s = size/96.0; 
  tx = tx0 = size/2.0 - 4.0*s;
  ty = ty0 = size/2.0 - 4.0*s;
  txn = tyn = s;
  k = 0;
  for(i=0;i<8;i++){
    for(j=0;j<8;j++){
      x[k] = tx; y[k] = ty; xn[k] = txn; yn[k] = tyn; k += 1;
      tx += txn;
    }
    ty += tyn;
    tx = tx0;
  }
  txn = tyn = 2.0*s;
  tx = tx0 - 2*txn;
  ty = ty0 - 2*tyn;
  for(j=0;j<8;j++){ 
    x[k] = tx; y[k] = ty;         xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx; y[k] = ty+tyn;     xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx; y[k] = ty+6.0*tyn; xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx; y[k] = ty+7.0*tyn; xn[k] = txn; yn[k] = tyn; k += 1;
    tx += txn;
  }
  tx = tx0 - 2*txn;
  ty = ty0;
  for(i=0;i<4;i++){ 
    x[k] = tx;         y[k] = ty; xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx+txn;     y[k] = ty; xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx+6.0*txn; y[k] = ty; xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx+7.0*txn; y[k] = ty; xn[k] = txn; yn[k] = tyn; k += 1;
    ty += tyn;
  }
  txn = tyn = 4.0*s;
  tx = tx0 - 3*txn;
  ty = ty0 - 3*tyn;
  for(j=0;j<8;j++){ 
    x[k] = tx; y[k] = ty;         xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx; y[k] = ty+tyn;     xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx; y[k] = ty+6.0*tyn; xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx; y[k] = ty+7.0*tyn; xn[k] = txn; yn[k] = tyn; k += 1;
    tx += txn;
  }
  tx = tx0 - 3.0*txn;
  ty = ty0 - tyn;
  for(i=0;i<4;i++){ 
    x[k] = tx;         y[k] = ty; xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx+txn;     y[k] = ty; xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx+6.0*txn; y[k] = ty; xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx+7.0*txn; y[k] = ty; xn[k] = txn; yn[k] = tyn; k += 1;
    ty += tyn;
  }
  txn = tyn = 8.0*s;
  tx = tx0 - 3.0*txn - txn/2;
  ty = ty0 - 3.0*tyn - tyn/2;
  for(j=0;j<8;j++){ 
    x[k] = tx; y[k] = ty;         xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx; y[k] = ty+tyn;     xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx; y[k] = ty+6.0*tyn; xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx; y[k] = ty+7.0*tyn; xn[k] = txn; yn[k] = tyn; k += 1;
    tx += txn;
  }
  tx = tx0 - 3.0*txn - txn/2;
  ty = ty0 - tyn - tyn/2;
  for(i=0;i<4;i++){ 
    x[k] = tx;         y[k] = ty; xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx+txn;     y[k] = ty; xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx+6.0*txn; y[k] = ty; xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx+7.0*txn; y[k] = ty; xn[k] = txn; yn[k] = tyn; k += 1;
    ty += tyn;
  }
  txn = tyn = 16.0*s;
  tx = tx0 - 2.0*txn - txn/2 - txn/4;
  ty = ty0 - 2.0*tyn - tyn/2 - txn/4;
  for(j=0;j<6;j++){ 
    x[k] = tx; y[k] = ty;         xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx; y[k] = ty+5.0*tyn; xn[k] = txn; yn[k] = tyn; k += 1;
    tx += txn;
  }
  tx = tx0 - 2.0*txn - txn/2 - txn/4;
  ty = ty0 - tyn - tyn/2 - txn/4;
  for(i=0;i<4;i++){ 
    x[k] = tx;         y[k] = ty; xn[k] = txn; yn[k] = tyn; k += 1;
    x[k] = tx+5.0*txn; y[k] = ty; xn[k] = txn; yn[k] = tyn; k += 1;
    ty += tyn;
  }
  for(i=0;i<n;i++) 
    y[i] = size - y[i];
  *rx = x; *ry = y; *rxn = xn; *ryn = yn; *rn = n;
}
void grid_get_02_coords(size,rx,ry,rxn,ryn,rn)
     float size;
     float **rx,**ry,**rxn,**ryn;
     int *rn;
{
  int i,j,k;
  int n;
  float *x,*y,*xn,*yn;
  double s,tx0,ty0,tx,ty,txn,tyn;
  n = 228;
  x = (float *)myalloc(n*sizeof(float));
  y = (float *)myalloc(n*sizeof(float));
  xn = (float *)myalloc(n*sizeof(float));
  yn = (float *)myalloc(n*sizeof(float));
  s = size/96.0; 
  i = j = k = tx = ty = tx0 = ty0 = txn = tyn = s;
  n = tx + ty + tx0 + ty0 + txn + tyn + i + j + k;
  exit_error("GRID_GET_02_COORDS","Not yet implemented");
  *rx = x; *ry = y; *rxn = xn; *ryn = yn; *rn = n;
}
void grid_get_03_coords(size,rx,ry,rxn,ryn,rn)
     float size;
     float **rx,**ry,**rxn,**ryn;
     int *rn;
{
  int i,j,k;
  int n;
  float *x,*y,*xn,*yn;
  double s,tx0,ty0,tx,ty,txn,tyn;
  n = 228;
  x = (float *)myalloc(n*sizeof(float));
  y = (float *)myalloc(n*sizeof(float));
  xn = (float *)myalloc(n*sizeof(float));
  yn = (float *)myalloc(n*sizeof(float));
  s = size/96.0; 
  i = j = k = tx = ty = tx0 = ty0 = txn = tyn = s;
  n = tx + ty + tx0 + ty0 + txn + tyn + i + j + k;
  exit_error("GRID_GET_03_COORDS","Not yet implemented");
  *rx = x; *ry = y; *rxn = xn; *ryn = yn; *rn = n;
}
void grid_get_04_coords(size,rx,ry,rxn,ryn,rn)
     float size;
     float **rx,**ry,**rxn,**ryn;
     int *rn;
{
  int i,j,k;
  int n;
  float *x,*y,*xn,*yn;
  double s,tx0,ty0,tx,ty,txn,tyn;
  n = 228;
  x = (float *)myalloc(n*sizeof(float));
  y = (float *)myalloc(n*sizeof(float));
  xn = (float *)myalloc(n*sizeof(float));
  yn = (float *)myalloc(n*sizeof(float));
  s = size/96.0; 
  i = j = k = tx = ty = tx0 = ty0 = txn = tyn = s;
  n = tx + ty + tx0 + ty0 + txn + tyn + i + j + k;
  exit_error("GRID_GET_04_COORDS","Not yet implemented");
  *rx = x; *ry = y; *rxn = xn; *ryn = yn; *rn = n;
}
void grid_get_05_coords(xsize,ysize,theta,nw,nh,barw,barh,rx,ry,rxn,ryn)
     float xsize,ysize,theta;
     int nw,nh,barw,barh;
     float **rx,**ry,**rxn,**ryn;
{
  int i,j,k;
  int n;
  float *x,*y,*xn,*yn,sf;
  float r,wo2,ho2,phi1,phi2,hscale,vscale,hdist,vdist,xoff,yoff;
  wo2 = (float)(nw*barw)/2.0;
  ho2 = (float)(nh*barh)/2.0;
  phi1 = atan2(ho2,wo2); 
  phi1 = 180.0 * phi1/M_PI;
  phi1 += theta;
  phi2 = atan2(ho2,-wo2); 
  phi2 = 180.0 * phi2/M_PI;
  phi2 += theta;
  r = sqrt((float)(nw*barw)*(nw*barw) + (nh*barh)*(nh*barh));
  if (fabs(sin(M_PI*phi1/180.0)) > fabs(sin(M_PI*phi2/180.0)))
    vdist = r*fabs(sin(M_PI*phi1/180.0));
  else
    vdist = r*fabs(sin(M_PI*phi2/180.0));
  if (fabs(cos(M_PI*phi1/180.0)) > fabs(cos(M_PI*phi2/180.0)))
    hdist = r*fabs(cos(M_PI*phi1/180.0));
  else
    hdist = r*fabs(cos(M_PI*phi2/180.0));
  hscale = xsize/hdist;
  vscale = ysize/vdist;
  if (hscale < vscale)
    sf = hscale;
  else
    sf = vscale;
  xoff = xsize/2.0 - (float)(nw*barw*sf)/2.0;
  yoff = ysize/2.0 - (float)(nh*barh*sf)/2.0;
  n = nw*nh;
  x = (float *)myalloc(n*sizeof(float));
  y = (float *)myalloc(n*sizeof(float));
  xn = (float *)myalloc(n*sizeof(float));
  yn = (float *)myalloc(n*sizeof(float));
  k = 0;
  for(i=0;i<nh;i++){
    for(j=0;j<nw;j++){
      if (!((i==0) && (j==0) && (n==256))){ 
	x[k] = xoff + (float)(j * barw)*sf;
	y[k] = yoff + (float)((i+1) * barh)*sf;
	xn[k] = barw*sf;
	yn[k] = barh*sf;
	k += 1;
      }
    }
  }
  *rx = x; *ry = y; *rxn = xn; *ryn = yn;
}
int grid_get_size(gtype)
     int gtype;
{
  if (gtype == 0)
    return 16;
  else if (gtype == 1)
    return 96;
  else if (gtype == 2)
    return 138; 
  else if (gtype == 3)
    return 1; 
  else if (gtype == 4)
    return 1; 
  else
    exit_error("GRID_GET_SIZE","Unknown grid type");
  return -1;
}
void grid_get_coords(gtype,size,rx,ry,rxn,ryn,rn)
     int gtype;
     float size,**rx,**ry,**rxn,**ryn;
     int *rn;
{
  if (gtype == 0)
    grid_get_00_coords(size,rx,ry,rxn,ryn,rn);
  else if (gtype == 1)
    grid_get_01_coords(size,rx,ry,rxn,ryn,rn);
  else if (gtype == 2)
    grid_get_02_coords(size,rx,ry,rxn,ryn,rn);
  else if (gtype == 3)
    grid_get_03_coords(size,rx,ry,rxn,ryn,rn);
  else if (gtype == 4)
    grid_get_04_coords(size,rx,ry,rxn,ryn,rn);
  else
    exit_error("GRID_GET_COORDS","Unknown grid type");
}
void grid_get_index(gtype,rndx,rn)
     int gtype;
     int **rndx,*rn;
{
  if (gtype == 0)
    grid_get_00_index(rndx,rn);
  else if (gtype == 1)
    grid_get_01_index(rndx,rn);
  else if (gtype == 2)
    grid_get_02_index(rndx,rn);
  else if (gtype == 3)
    grid_get_03_index(rndx,rn);
  else if (gtype == 4)
    grid_get_04_index(rndx,rn);
  else if (gtype == 5)
    grid_get_05_index(rndx,*rn); 
  else
    exit_error("GRID_GET_INDEX","Unknown grid type");
}
void grid_get_stim_for_params(seed,dist,dwell,duration,framerate,ngrid,
			      deriv_flag,rfdt,rxdata,rydata,rnframes)
     int seed,dist,dwell,duration,framerate,ngrid,deriv_flag;
     float *rfdt,**rxdata,***rydata;
     int *rnframes;
{
  int i,j,k;
  int count,nframes;
  float *xdata,**ydata,*t,fdt;
  printf("    ngrid = %d\n",ngrid);
  nframes = duration * framerate;
  fdt = get_wn_actual_dt(framerate,0);
  xdata = (float *)myalloc(nframes*sizeof(float));
  ydata = get_2d_farray(nframes,ngrid); 
  yarlsrand(seed); 
  count = dwell; 
  for(i=0;i<nframes;i++){
    xdata[i] = (double)i*fdt;
    if (count == dwell){
      if (dist == 2){
	for(j=0;j<ngrid;j++){
	  if (yarlrand() > 16383)
	    ydata[i][j] = -1.0;
	  else
	    ydata[i][j] = 1.0;
	}
      }else if (dist == 3){
	for(j=0;j<ngrid;j++){
	  k = yarlrand();
	  if (k < 10922)
	    ydata[i][j] = -1.0;
	  else if (k < 21846)
	    ydata[i][j] = 0.0;
	  else
	    ydata[i][j] = 1.0;
	}
      }else
	exit_error("GRID_GET_STIM_FOR_PARAMS","Unkown distrib type");
      count = 0;
    }else{
      for(j=0;j<ngrid;j++)
	ydata[i][j] = ydata[i-1][j];
    }
    count += 1;
  }
  if (deriv_flag == 1){
    printf("  *** CHANGE = 1,  NO_CHANGE = -1\n");
    t = (float *)myalloc(nframes*sizeof(float));
    for(i=0;i<ngrid;i++){ 
      t[0] = 1.0; 
      for (j=1;j<nframes;j++){ 
	if (ydata[j][i] != ydata[j-1][i])
	  t[j] = 1.0; 
	else
	  t[j] = -1.0; 
      }
      for (j=0;j<nframes;j++) 
	ydata[j][i] = t[j];
    }
    myfree(t);
  }
  *rfdt = fdt; *rxdata = xdata; *rydata = ydata; *rnframes = nframes;
}	
