/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "my_util.h"
#include "nr_util.h"
char *get_path_without_name(name)
     char *name;
{
  int i;
  int nch;
  char *ptr,*p;
  ptr = strrchr(name,'/');
  if (ptr!=NULL){
    nch = (int)(ptr - name) + 1;
    p = (char *)myalloc((nch+1)*sizeof(char));  
    for(i=0;i<nch;i++)
      p[i] = name[i];
    p[nch] = '\0';
  }else
    p = NULL;
  return p;
}
void get_name_without_path(name,short_name)
     char name[],short_name[];
{
  char *ptr;
  ptr = strrchr(name,'/');
  if (ptr!=NULL)
    strcpy(short_name,(ptr+1));
  else
    strcpy(short_name,name);
}
void get_filename_extension(name,extension)
     char name[],extension[];
{
  char *ptr;
  ptr = strrchr(name,'.');
  if (ptr!=NULL)
    strcpy(extension,(ptr+1));
  else
    extension[0] = '\0';
}
void get_filename_without_extension(name,shortname)
     char name[],shortname[];
{
  char *ptr;
  strcpy(shortname,name);
  ptr = strrchr(shortname,'.');
  if (ptr!=NULL)
    *ptr = '\0'; 
}
float cumulative_binomial(k,n,p)
     int k,n;
     float p;
{
  float a,b,x;
  float t;
  a = (float)k;   
  b = (float)(n-k+1);
  x = p;
  t = mybetai(a,b,x);
  if (t==-1.0){
    printf("  *** k=%d n=%d p=%f\n",k,n,p);
    exit_error("CUMULATIVE_BINOMIAL","Error from mybetai");
  }
  return t;
}
float binomial_signif(k,n,p)
     int k,n;
     float p;
{
  int nn;
  if (n>5000){
    nn = 5000 + my_rint(pow((float)(n-5000),0.8));
    k = my_rint((float)k * (float)nn/(float)n);
    n = nn;
  }
  if ((float)k >= (float)n*p)
    return cumulative_binomial(k,n,p); 
  else
    return cumulative_binomial(n-k,n,1.0-p); 
}
void projection_2d(u1,u2,v1,v2,r1,r2)
     float u1,u2,v1,v2,*r1,*r2;
{
  float m,d;
  d = (v1*v1 + v2*v2);
  if (d==0.0){
    m = 0.0;
    exit_error("PROJECTION_2D","Vector has length 0");
  }else
    m = (u1*v1 + u2*v2)/d;
  *r1 = m*v1;
  *r2 = m*v2;
}
double difference_log(a,b,base)
     double a,b;
     int base;
{
  double d;
  if (base == 0)
    d = log(a) - log(b);
  else if (base == 2)
    d = my_log2(a) - my_log2(b);
  else if (base == 10)
    d = log10(a) - log10(b);
  else{
    d = 0.0;
    exit_error("LOG_DIFF","Unknown base");
  }
  return d;
}
float get_circular_diff(a,b,d)
     float a,b,d;
{
  float x;
  x = (a-b);
  if (x < 0.0)
    x *= -1;
  if (x > d/2.0)
    x = d-x;
  return x;
}
float get_signed_circular_diff(a,b,d)
     float a,b,d;
{
  float x;
  x = (a-b);
  if (x > d/2.0)
    x = -(d-x);
  else if (x < -d/2)
    x = d+x;
  return x;
}
float get_circular_interval_length_containing_point(a,b,c,d)
     float a,b,c,d;
{
  float dab,dac,dbc,t;
  dab = get_circular_diff(a,b,d); 
  dac = get_circular_diff(a,c,d); 
  dbc = get_circular_diff(b,c,d); 
  t = fabs((dac + dbc) - dab);
  if (t < 0.01*d)
    return dab;
  else
    return (d - dab);
}
void swap_int(a,b)
     int *a,*b;
{
  int t;
  t = *a;
  *a = *b;
  *b = t;
}  
void swap_float(a,b)
     float *a,*b;
{
  float t;
  t = *a;
  *a = *b;
  *b = t;
}  
void wrap_float(f,size)
     float *f,size;
{
  while(*f < 0.0)
    *f += size;
  while(*f >= size)
    *f -= size;
}  
void wrap_double(f,size)
     double *f,size;
{
  while(*f < 0.0)
    *f += size;
  while(*f >= size)
    *f -= size;
}  
float diff_sdev(a,b,sdev)
     float a,b,sdev;
{
  if (a > b)
    return (a-b)/sdev;
  else
    return (b-a)/sdev;
}  
float interp_two(x0,x1,x2,y1,y2)
     float x0,x1,x2,y1,y2;
{
  float f,y0;
  f = (x0-x1)/(x2-x1);
  y0 = y1 + f*(y2-y1);
  return y0;
}  
int misc_util_get_round_number(x)
     float x;
{
  int ten,y;
  float xt;
  xt = x;
  ten = 1;
  if (xt < 0.0){
    xt = -xt;
    ten = -1;
  }
  while(xt > 10.0){
    xt /= 10.0;
    ten *= 10;
  }
  if (xt < 1.5)
    y = 1;
  else if (xt < 3.5)
    y = 2;
  else if (xt < 7.5)
    y = 5;
  else
    y = 10;
  return y * ten;
}  
