/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "my_util.h"
#include "farray_util.h"
#include "carray_util.h"
#include "func_util.h"
#include "myrand_util.h"
#include "spike_util.h"
#include "paramfile_util.h"
#include "kernel_util.h"
#include "mod_util.h"
#include "ifc_util.h"
#include "mod.h" 
#include "paramfile.h"
static int myid = -1;         
static char *mylogf = NULL;   
char ggstr[LONG_SLEN];        
float *mod_lin_f = NULL;    
int mod_lin_fn;             
float mod_lin_tscale;
int   mod_lin_tn;
struct onode *mod_lin_sgo;   
char         *mod_lin_sgen;  
struct mod_lin_unit{
  char *name;
  float rate;         
  float prob;         
  int delay;          
  int spread;         
  int burst_dur;      
  float burst_prob;   
  float *ps;          
  int    pn;          
  float *cs;          
  int    cn;          
  float *ss;          
  int    sn;          
};
int   mod_lin_unit_n;             
struct mod_lin_unit **mod_lin_u;  
struct mod_lin_unit *mod_lin_get_unit(uo)
     struct onode *uo;
{
  float d,s,bdur;
  struct mod_lin_unit *lu;
  lu = (struct mod_lin_unit *)myalloc(sizeof(struct mod_lin_unit));
  lu->name   = onode_getpar_chr_exit(uo,"name");
  lu->rate   = onode_getpar_flt_exit(uo,"rate");
  lu->prob   = onode_getpar_flt_exit(uo,"prob_c");
  d          = onode_getpar_flt_dflt(uo,"delay_c",0.0);
  s          = onode_getpar_flt_dflt(uo,"spread_c",0.0);
  bdur       = onode_getpar_flt_dflt(uo,"burst_dur",0.0);
  lu->burst_prob = onode_getpar_flt_dflt(uo,"burst_prob",0.0);
  lu->burst_dur = my_rint(bdur / mod_lin_tscale);
  lu->delay     = my_rint(   d / mod_lin_tscale);
  lu->spread    = my_rint(   s / mod_lin_tscale);
  printf("  Unit %s\n",lu->name);
  printf("    rate  %f\n",lu->rate);
  printf("    prob  %f\n",lu->prob);
  printf("    delay  %d\n",lu->delay);
  printf("    spread  %d\n",lu->spread);
  printf("    burst_dur  %d\n",lu->burst_dur);
  printf("    burst_prob  %f\n",lu->burst_prob);
  lu->ps = NULL;
  lu->pn = 0;
  lu->ss = NULL;
  lu->sn = 0;
  lu->cs = NULL;
  lu->cn = 0;
  return lu;
}
int mod_lin_unit_i(name)
     char *name;
{
  int i,k;
  int n;
  struct mod_lin_unit **u;  
  n = mod_lin_unit_n;
  u = mod_lin_u;
  k = -1;
  i = 0;
  while((i < n) && (k == -1)){
    if (compare_prefix_string_order(u[i]->name,name) == 1){
      k = i;
    }else
      i += 1;
  }
  return k;
}
void mod_lin_01_prep(m,r)
     struct model_struct *m;    
     struct response_struct *r; 
{
  int i;
  char *sgen;
  struct onode *fo,*sgo,*t;
  mylog(mylogf,"  MOD_LIN_01_PREP\n");
  mod_lin_tn     = onode_getpar_int_exit(m->o,"tn");
  mod_lin_tscale = onode_getpar_flt_exit(m->o,"tscale");
  fo = onode_child_get_unique(m->o,"filter");
  mod_lin_fn = -1;  
  mod_lin_f = kernu_o_filter(mylogf,fo,mod_lin_tscale,&mod_lin_fn);
  mod_lin_sgo  = onode_child_get_unique(m->o,"spike_gen");
  mod_lin_sgen = onode_getpar_chr_exit(mod_lin_sgo,"type");
  mod_lin_unit_n = onode_count_otype(m->o,"unit");
  if (mod_lin_unit_n > 0){
    mod_lin_u = (struct mod_lin_unit **)myalloc(mod_lin_unit_n*
						sizeof(struct mod_lin_unit *));
    i = 0;
    t = onode_get_next_type(m->o->o,"unit");
    while(t != NULL){
      mod_lin_u[i] = mod_lin_get_unit(t);
      i += 1;
      t = onode_get_next_type(t->next,"unit");
    }
  }
}
void mod_lin_01_done(m)
     struct model_struct *m;        
{
  mylog(mylogf,"  MOD_LIN_01_DONE\n");
  myfree(mod_lin_f);
}
void mod_lin_01_get_response(m,s,r,k)
     struct model_struct *m;        
     struct stim_struct *s;         
     struct response_struct *r;     
     int k;                         
{
  int i,j;
  int si,ti,tn,seed,*spk,ns,vn,n,*pseed,tseed,sdt,tsn,bdur;
  float tscale,samp,*conv,min,max,*fs,*v;
  float pr,fdt,*cs,*ts;
  struct mod_lin_unit *tu;
  tn     = mod_lin_tn;
  tscale = mod_lin_tscale;
  n      = mod_lin_unit_n;
  samp = 1.0/tscale;
  conv = convolve_with_mask_causal(s->d1,tn,mod_lin_f,mod_lin_fn);
  get_min_max_farray(conv,tn,&min,&max);
  sprintf(ggstr,"    min, max of raw conv:  %f %f\n",min,max);
  mylog(mylogf,ggstr);
  if (strcmp(mod_lin_sgen,"poisson")==0){
    seed = m->mseed[r->tsi];
    ifc_util_poisson(mylogf,m,mod_lin_sgo,conv,tn,samp,seed,0,1,&fs,&ns,&v,&vn);
    sprintf(ggstr,"    Made %d Poisson spikes\n",ns);
    mylog(mylogf,ggstr);
  }else
    mylogx(mylogf,"MOD_LIN_01_GET_RESPONSE","Spike generation type not imp'd");
  if (n > 0){
    pseed = get_seeds(m->mseed[r->tsi],100000,2*n);
    for(i=0;i<n;i++){
      tu = mod_lin_u[i];
      if (tu->ps != NULL)
	myfree(tu->ps);
      if (tu->ss != NULL)
	myfree(tu->ss);
      if (tu->cs != NULL)
	myfree(tu->cs);
      tseed = pseed[i];
      make_poisson_float_spikes(&(tu->ps),&(tu->pn),0,tn,samp,tu->rate,0.0,
				0.0,1,&tseed);
      tseed = pseed[2*i];
      if (tseed > 0)
	tseed *= -1;
      pr  = tu->prob;
      fdt = tu->delay;
      cs = (float *)myalloc(ns*sizeof(float));
      si = 0;
      for(j=0;j<ns;j++){
	if (myrand_util_ran2(&tseed) < pr){
	  sdt = 0;  
	  if (tu->spread > 0) 
	    sdt = my_rint((float)tu->spread * myrand_util_ran2(&tseed));
	  cs[si] = fs[j] + fdt + sdt;
	  si += 1;
	}
      }
      tu->cs = copy_farray(cs,si);
      tu->cn = si;
      myfree(cs);
      ts = merge_spike_arrays_float(tu->ps,tu->pn,tu->cs,tu->cn);
      tsn = tu->cn + tu->pn;
      bdur = tu->burst_dur;
      if ((bdur > 1) && (tsn > 0)){  
	pr = tu->burst_prob;
	cs = (float *)myalloc(tsn*bdur*sizeof(float));
	si = 0;
	for(j=0;j<tsn;j++){  
	  for(ti=0;ti<bdur;ti++){  
	    if (myrand_util_ran2(&tseed) < pr){
	      cs[si] = ts[j] + ti;
	      si += 1;
	    }
	  }
	}
	sort_farray(cs,si);
	myfree(ts);  
	ts = cs;
	tsn = si;
      }
      tu->ss = copy_farray(ts,tsn);
      tu->sn = tsn;
      if (ts != NULL)  
	myfree(ts);
    }
    myfree(pseed);
  }
  for(i=0;i<r->n;i++){ 
    if (strcmp(r->datid[i],"spikes")==0){
      mod_util_resp_store_s(r,i,fs,ns,1,mylogf);  
    }else if (strcmp(r->datid[i],"prob")==0){
      mod_util_resp_store_f(r,i,v,vn,1,mylogf);  
    }else if (strcmp(r->datid[i],"f")==0){
      mod_util_resp_store_f(r,i,conv,tn,1,mylogf);  
    }else if (mod_lin_unit_i(r->datid[i]) >=0){
      tu = mod_lin_u[mod_lin_unit_i(r->datid[i])];
      if (strcmp(r->datid[i],tu->name)==0){
	mod_util_resp_store_s(r,i,tu->ss,tu->sn,1,mylogf); 
      }else{
	printf(" response:  %s\n",r->datid[i]);
	exit(0);
      }
    }
  }
  myfree(conv);
  if (fs != NULL)
    myfree(fs);
  myfree(v);
}
void mod_lin_run_lin_01(m,s,r,k,action)
     struct model_struct *m;     
     struct stim_struct *s;      
     struct response_struct *r;  
     int k;                      
     int action;                 
{
  myid = mylog_set_id_log(m->process_id,&mylogf);  
  if (action == 0){
    mod_lin_01_prep(m,r);
  }else if (action == 1){
    mod_lin_01_get_response(m,s,r,k);
  }else if (action == -1){
    mod_lin_01_done(m);
  }
}
