/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "my_util.h"
#include "carray_util.h"
#include "iarray_util.h"
#include "farray_util.h"
#include "func_util.h"
#include "myrand_util.h"
#include "paramfile_util.h"
#include "mod_util.h"
#include "ifc_util.h"
#include "pop_cell_util.h"
#include "pop_util.h"
#include "pop_low_util.h"
#include "mod_conn_util.h"
#include "mod_dog_util.h"
#include "mod_mesh_util.h"
#include "mod_gui_util.h"
#include "mod_mon_util.h"
#include "mod.h"
#include "ifc.h"
#include "paramfile.h"
static int myid = -1;         
static char *mylogf = NULL;   
char ggstr[LONG_SLEN];        
struct pop_top *mpt;          
void mod_pop_init_lgn_cells_resp(laylist,nlay,r)
     struct pop_layer **laylist;
     int nlay;
     struct response_struct *r;    
{
  int i,j;
  int xi,yi,zi;
  char tname[SLEN],*datid;
  struct pop_cell *cl,*tc;
  struct pop_layer *lay;
  for(i=0;i<r->n;i++){
    if ((strcmp(r->plname[i],"lgn_off")==0)||
	(strcmp(r->plname[i],"lgn_on")==0)){
      mylog_exit(mylogf,
	"MOD_POP_INIT_LGN_CELLS_RESP  Old LGN pop names in .rsp file");
    }
    lay = pop_cell_get_layer_pointer(laylist,nlay,r->plname[i]);
    if (lay == NULL){
      sprintf(ggstr,"*** Layer name for requested response:  %s\n",
	      r->plname[i]);
      mylog(mylogf,ggstr);
      mylog_exit(mylogf,"MOD_POP_INIT_LGN_CELLS_RESP  Layer name not found.");
    }
    if (pop_util_lclass_is_lgn0(lay)){  
      xi = r->xi[i];
      yi = r->yi[i];
      zi = r->zi[i];
      if ((xi < 0) || (xi >= lay->xn))
	exit_error("MOD_POP_INIT_LGN_CELLS_RESP","Response xi out of bounds");
      if ((yi < 0) || (yi >= lay->yn))
	exit_error("MOD_POP_INIT_LGN_CELLS_RESP","Response yi out of bounds");
      if ((zi < 0) || (zi >= lay->zn))
	exit_error("MOD_POP_INIT_LGN_CELLS_RESP","Response zi out of bounds");
      sprintf(tname,"%s_%d_%d_%d",r->plname[i],xi,yi,zi);
      if (lay->c[xi][yi] == NULL){  
	cl = (struct pop_cell *)myalloc(lay->zn*sizeof(struct pop_cell));
	for(j=0;j<lay->zn;j++){
	  pop_cell_init_cell(&(cl[j])); 
	}
	lay->c[xi][yi] = cl; 
      }else
	cl = lay->c[xi][yi];  
      tc = &(cl[zi]); 
      if (tc->name == NULL){
	tc->layx = xi; 
	tc->layy = yi;
	tc->layz = zi;
	tc->name = strdup(tname);
	tc->subclass = NULL;
	tc->pl = lay;
      }
      datid = r->datid[i];
      if (strcmp(datid,"spikes")==0){
	tc->savs = 1;  
      }else if (strcmp(datid,"vm")==0){
	tc->savv = 1;  
      }else if (strcmp(datid,"rate")==0){
	tc->savv = 1;  
      }else if (strcmp(datid,"gx")==0){
	tc->savg = 1;  
      }else if (strcmp(datid,"gi")==0){
	tc->savg = 2;  
      }else if (strcmp(datid,"raw")==0){
	tc->savg = 3;  
      }else{
	printf("***  %s\n",datid);
	exit_error("MOD_POP_INIT_LGN_CELLS_RESP","Unknown Data ID");
      }
    }
  }
}
void mod_pop_store_lgn(laylist,nlay,r)
     struct pop_layer **laylist;
     int nlay;
     struct response_struct *r;    
{
  int i;
  int xi,yi,zi,tcnt,n;
  float *ts,samp;
  char *datid;
  struct pop_cell *c;
  struct pop_layer *lay;
  for(i=0;i<r->n;i++){
    lay = pop_low_get_layer_for_name(laylist,nlay,r->plname[i]);
    if (lay != NULL){
      if (pop_util_lclass_is_lgn0(lay) ||  
	  pop_util_lclass_is_rgci(lay)){   
	xi = r->xi[i];
	yi = r->yi[i];
	zi = r->zi[i];
	c = pop_cell_laylist_get_cell_pointer(laylist,nlay,r->plname[i],
					      xi,yi,zi);
	datid = r->datid[i];
	if (strcmp(datid,"spikes")==0){
	  tcnt = lay->cnt[zi][xi][yi];
	  ts   = lay->s[zi][xi][yi];
	  c->ns = tcnt;
	  if (c->ns < 0){
	    sprintf(ggstr,"*** LGN %d %d %d  NOT CONNECTED, NO SPIKES MADE\n",
		    xi,yi,zi);
	    mylog(mylogf,ggstr);
	    c->ns = 0;
	    c->s = (float *)NULL;
	  }else{
	    c->s = copy_farray(ts,tcnt);
	  }
	}else if (strcmp(datid,"vm")==0){
	  ;
	}else if (strcmp(datid,"rate")==0){
	  ;
	}else if (strcmp(datid,"gx")==0){
	  ;
	}else if (strcmp(datid,"gi")==0){
	  ;
	}else{
	  exit_error("MOD_POP_STORE_LGN","Unknown data ID.");
	}
      }else if (strcmp(lay->laytype,"virtual")==0){
	xi = r->xi[i];  
	yi = r->yi[i];
	zi = r->zi[i];
	c = pop_cell_laylist_get_cell_pointer(laylist,nlay,r->plname[i],
					      xi,yi,zi);
	datid = r->datid[i];
	if (strcmp(datid,"photo")==0){
	  ts = mod_mesh_get_mm_resp(xi,yi,zi,"photo",&n,&samp);
	  popc_csav_set_f(c,"photo",ts,n,samp);
	}else if (strcmp(datid,"horiz")==0){
	  ts = mod_mesh_get_mm_resp(xi,yi,zi,"horiz",&n,&samp);
	  popc_csav_set_f(c,"horiz",ts,n,samp);
	}else if (strcmp(datid,"h2")==0){
	  ts = mod_mesh_get_mm_resp(xi,yi,zi,"h2",&n,&samp);
	  popc_csav_set_f(c,"h2",ts,n,samp);
	}else if (strcmp(datid,"diff")==0){
	  ts = mod_mesh_get_mm_resp(xi,yi,zi,"diff",&n,&samp);
	  popc_csav_set_f(c,"diff",ts,n,samp);
	}else{
	  exit_error("MOD_POP_STORE_LGN","Unknown data ID for virtual.");
	}
      }
    }
  }
}
void mod_pop_store_lgn_sav(mpt,r,lgn_lay_name,lzi,gx,gi,vm,vn,tn)
     struct pop_top *mpt;
     struct response_struct *r;    
     char *lgn_lay_name;           
     int lzi;                      
     float ***gx,***gi;            
     float ***vm;                  
     int vn;                       
     int tn;                       
{
  int i;
  int xi,yi,zi;
  char *datid;
  struct pop_cell *c;
  for(i=0;i<r->n;i++){                         
    if (strcmp(lgn_lay_name,r->plname[i])==0){ 
      if (strcmp(r->datid[i],"spikes")!=0){    
	zi = r->zi[i];
	if (r->zi[i] == lzi){  
	  xi = r->xi[i];
	  yi = r->yi[i];
	  c = pop_cell_laylist_get_cell_pointer(mpt->lay,mpt->nlay,
						r->plname[i],xi,yi,zi);
	  datid = r->datid[i];
	  if ((strcmp(datid,"vm")==0)||(strcmp(datid,"rate")==0)){
	    c->vm = copy_farray(vm[xi][yi],vn);  
	    c->vn = vn;
	    c->dxsav = 1000.0;  
	    c->vmt = NULL;      
	  }else if (strcmp(datid,"gx")==0){
	    c->gtx0 = copy_farray(gx[xi][yi],tn);  
	    c->n0 = tn;
	  }else if (strcmp(datid,"gi")==0){
	    c->gti0 = copy_farray(gi[xi][yi],tn);  
	    c->n0 = tn;
	  }
	}
      }
    }
  }
}
int mod_pop_lgn_set_sav_flag(laylist,nlay,r,flag,xn,yn,lgn_lay_name,lzi)
     struct pop_layer **laylist;
     int nlay;
     struct response_struct *r;    
     int **flag;      
     int xn,yn;
     char *lgn_lay_name;
     int lzi;            
{
  int i;
  int xi,yi,zi,sav_cnt;
  sav_cnt = 0;  
  for(i=0;i<r->n;i++){                         
    if (strcmp(lgn_lay_name,r->plname[i])==0){ 
      if (strcmp(r->datid[i],"spikes")!=0){    
	zi = r->zi[i];      
	if (r->zi[i] == lzi){  
	  xi = r->xi[i];  
	  yi = r->yi[i];
	  flag[xi][yi] = 2;  
	  sav_cnt += 1;
	}
      }
    }
  }
  return sav_cnt;
}
struct pop_mech *mod_pop_get_mech(pl,name)
     struct pop_layer *pl;
     char *name;
{
  int k;
  int done;
  struct pop_mech *pm;
  pm = NULL;
  k = 0;
  done = 0;
  while(!done){
    if (k < pl->nmech){
      if (strcmp(pl->mech[k]->name,name)==0){
	pm = pl->mech[k];
	done = 1;
      }else
	k += 1;
    }else
      done = 1;
  }
  if (pm == NULL){
    printf("  *** mechanism name:  %s\n",name);
    exit_error("MOD_POP_GET_MECH","Mechanism not found");
  }
  return pm;
}
void mod_pop_config_cell(o,po,pl,c,r,xi,yi,zi)
     struct onode *o;              
     struct onode *po;             
     struct pop_layer *pl;         
     struct pop_cell *c;           
     struct response_struct *r;    
     int xi,yi,zi;
{
  int nmdaflag,ri,nbx,ai_ocdom;
  char tstr[SLEN],xyzstr[SLEN];
  float tau_r_ad,tau_f_ad;
  struct onode *bgo;
  struct pop_mech *mech;
  nmdaflag = 0;
  if (pl->ifcp != NULL)
    if (pl->ifcp->g_tran > 0.0)
      nmdaflag = 1;
  sprintf(xyzstr,"%d_%d_%d",xi,yi,zi);
  sprintf(tstr,"%s_%s",pl->name,xyzstr);
  c->name = strdup(tstr);
  c->layx = xi; 
  c->layy = yi;
  c->layz = zi;
  c->pl   = pl; 
  if (pl->attrib_fn > 0){
    c->attrib_f = (float *)myalloc(pl->attrib_fn*sizeof(float)); 
    pop_cell_attrib_default_set(c); 
  }else
    c->attrib_f = NULL;
  c->nout = c->nin = 0;
  c->out = (struct pop_syn *)NULL;
  c->in = (struct pop_syn *)NULL;
  if (pl->csav_n > 0){
    c->csav_cnt = get_zero_iarray(pl->csav_n);
    c->csav_f = get_2d_farray(pl->csav_n,0);  
  }
  if ((strcmp(c->pl->laytype,"virtual")==0) ||
      (strcmp(c->pl->laytype,"retina0_gc0")==0)){
    return;
  }
  nbx = onode_getpar_int_dflt(po,"circbuf_ex_n",100);
  c->ex1s = pop_cell_get_init_circbuf(nbx,0.0,0.0005,1); 
  c->in1s = pop_cell_get_init_circbuf(nbx,0.0,0.0005,1); 
  bgo = onode_get_node_type_item_val(po,"input","type","bg");
  if (bgo != NULL){
    c->bg_x_rate = onode_getpar_flt_dflt(bgo,"ex_rate",0.0);
    c->bg_x_amp  = onode_getpar_flt_dflt(bgo,"ex_amp",1.0);
    c->bg_i_rate = onode_getpar_flt_dflt(bgo,"in_rate",0.0);
    c->bg_i_amp  = onode_getpar_flt_dflt(bgo,"in_amp",1.0);
    c->bg_x_corr_f = onode_getpar_flt_dflt(bgo,"ex_corr_frac",0.0);
    c->bg_i_corr_f = onode_getpar_flt_dflt(bgo,"in_corr_frac",0.0);
  }else{
    c->bg_x_rate = 0.0;
    c->bg_x_amp  = 0.0;
    c->bg_i_rate = 0.0;
    c->bg_i_amp  = 0.0;
    c->bg_x_corr_f = 0.0; 
    c->bg_i_corr_f = 0.0;
  }
  c->bg_x_s = c->bg_i_s = NULL;
  c->bg_x_corr_c = c->bg_i_corr_c = NULL;
  c->bg_x_n = c->bg_i_n = -1; 
  c->ex1g = pop_cell_get_init_circbuf(100,0.0,0.00005,1); 
  if (nmdaflag)
    c->ex1g_nmda = pop_cell_get_init_circbuf(100,0.0,0.00005,1);
  c->in1g = pop_cell_get_init_circbuf(100,0.0,0.00005,1); 
  c->gex1t = 0.0; 
  c->gex1 = 0.0;  
  c->gdex1 = 0.0;
  mech = mod_pop_get_mech(pl,"ex");
  c->ex1tau = mech->tau;
  c->ex1amp = mech->amp;
  c->input_ex_sdf   = onode_getpar_flt_dflt(po,"ctx2ctx_ex_sdf",1.0);
  c->input_ex_sdtau = onode_getpar_flt_dflt(po,"ctx2ctx_ex_sdtau",0.0);
  c->input_ex_sdtau /= 1000.0;  
  c->gin1 = 0.0;  
  c->gdin1 = 0.0;
  mech = mod_pop_get_mech(pl,"in");
  c->in1tau = mech->tau;
  c->in1amp = mech->amp;
  c->gex1_nmda1 = c->gex1_nmda2 = c->gex1_nmda3 = 0.0;
  c->gdex1_nmda1 = c->gdex1_nmda2 = c->gdex1_nmda3 = 0.0;
  c->savg = 0;
  c->n0 = -1;
  c->samp0 = 1000.0;  
  c->gtx0 = c->gtn0 = c->gti0 = c->gta0 = (float *)NULL;
  c->ggain = NULL;
  if (pl->ifcp != NULL){
    c->cifcp = pl->ifcp;
    c->cpoissp = pl->poissp;
  }else
    exit_error("MOD_POP_CONFIG_CELL","No ifc param pointer given");
  if (c->cifcp->gbar_ad > 0.0){
    tau_r_ad = c->cifcp->tau_r_ad * c->samp0/1000.0;
    tau_f_ad = c->cifcp->tau_f_ad * c->samp0/1000.0;
    c->nad = (int)(c->cifcp->tau_f_ad*4.0 * c->samp0/1000.0);
    c->gta_pulse = diff_exp_farray(0.0,c->cifcp->gbar_ad,tau_f_ad,tau_r_ad,
				   c->nad);
    c->ad1g = pop_cell_get_init_circbuf(c->nad+5,0.0,1.0/c->samp0,1);
    c->ad1g->maskn = c->nad;  
  }else{
    c->nad = 0;
    c->gta_pulse = (float *)NULL;
    c->ad1g = NULL;
  }
  c->vm = NULL;
  c->vmd = NULL; 
  c->vmt = NULL;
  c->vnmax = 0;  
  c->vn = 0;
  c->savv = 0;  
  c->h1 = 0.0005;      
  c->eps = 0.0001;     
  c->reft = -1.0;      
  c->hmin = 1.0e-16;   
  c->nok = c->nbad = 0;
  c->ystart = c->cifcp->v_leak_x; 
}
void mod_pop_init_pop_null(pl)
     struct pop_layer *pl;
{
  int i,j,k;
  struct pop_cell ***c;
  pl->runflag = 0;
  pl->x0 = pl->y0 = 0.0;
  pl->xf = pl->yf = 1.0;
  pl->oddxoff = 0.0;
  pl->umx = pl->umy = -1.0;  
  pl->nmda_p = NULL;
  pl->ifcp = NULL;
  pl->poissp = NULL;
  c = (struct pop_cell ***)myalloc(pl->xn*sizeof(struct pop_cell **));
  for(i=0;i<pl->xn;i++){
    c[i] = (struct pop_cell **)myalloc(pl->yn*sizeof(struct pop_cell *));
    for(j=0;j<pl->yn;j++){
      c[i][j] = NULL;
    }
  }
  pl->c = c;
}
void mod_pop_init_pop_lgn(o,po,pl)
     struct onode *o;       
     struct onode *po;      
     struct pop_layer *pl;  
{
  int i,j,k;
  int ci,cj,xn,yn,zn;
  int mblock,mblock_n,mblock_seed;
  char *mblock_alg;
  struct pop_cell ***c,*cp;
  static char ***celltype = NULL;
  xn = pl->xn;
  yn = pl->yn;
  zn = pl->zn;
  mod_pop_init_pop_null(pl);
  pl->cnt = (int ***)myalloc(zn*sizeof(int **));
  pl->s   = (float ****)myalloc(zn*sizeof(float ***));
  for(i=0;i<zn;i++){
    pl->cnt[i] = NULL;
    pl->s[i]   = NULL;
  }
  pl->rl = NULL;
  pl->rr = NULL;
  mblock = onode_getpar_int_dflt(po,"mblock",-1);
  if (mblock > 0){
    exit_error("MOD_POP_INIT_POP_LGN","Is this needed w/ newlgn?");
    mblock_alg = onode_getpar_chr_exit(po,"mblock_alg");
    mblock_seed = onode_getpar_int_exit(po,"mblock_seed");
    if (mblock_seed > 0)
      mblock_seed = -mblock_seed;
    if (celltype == NULL){
      celltype = get_2d_pointer_carray(xn,yn);
      for(i=0;i<xn;i++)
	for(j=0;j<yn;j++)
	  celltype[i][j] = strdup("p");  
      for(i=0;i<xn;i+=mblock)
	for(j=0;j<yn;j+=mblock){
	  if (strcmp(mblock_alg,"center")==0){
	    ci = i + mblock/2;
	    cj = j + mblock/2;
	  }else{ 
	    ci = i + (int)((float)mblock * myrand_util_ran2(&mblock_seed));
	    cj = j + (int)((float)mblock * myrand_util_ran2(&mblock_seed));
	  }
	  if ((ci < xn)&&(cj < yn)){
	    myfree(celltype[ci][cj]);
	    celltype[ci][cj] = strdup("m");
	  }
	}
    }
    c = pl->c;
    for(i=0;i<xn;i++){
      for(j=0;j<yn;j++){
	c[i][j] = (struct pop_cell *)myalloc(sizeof(struct pop_cell));
	cp = &(c[i][j][0]);      
	pop_cell_init_cell(cp);  
	cp->subclass = strdup(celltype[i][j]);
      }
    }
  }
}
struct pop_mech *mod_pop_make_mech_psg_alpha(o)
     struct onode *o;     
{
  struct pop_mech *pm;
  pm = (struct pop_mech *)myalloc(sizeof(struct pop_mech));
  pm->o  = o;
  pm->name = onode_getpar_chr_exit(o,"name");
  pm->type = onode_getpar_chr_exit(o,"type");
  pm->tau  = onode_getpar_flt_exit(o,"tau");
  pm->amp  = onode_getpar_flt_exit(o,"amp");
  return pm;
}
struct pop_mech *mod_pop_make_mech_psg_doe(o)
     struct onode *o;     
{
  struct pop_mech *pm;
  pm = (struct pop_mech *)myalloc(sizeof(struct pop_mech));
  pm->o  = o;
  pm->name   = onode_getpar_chr_exit(o,"name");
  pm->type   = onode_getpar_chr_exit(o,"type");
  pm->tau_r  = onode_getpar_flt_exit(o,"tau_r");
  pm->tau_f  = onode_getpar_flt_exit(o,"tau_f");
  pm->amp    = onode_getpar_flt_exit(o,"amp");
  pm->sdf    = onode_getpar_flt_exit(o,"sdf");
  pm->sdtau  = onode_getpar_flt_exit(o,"sdtau");
  return pm;
}
struct pop_mech *mod_pop_make_mech_si_ds02(o)
     struct onode *o;     
{
  struct pop_mech *pm;
  pm = (struct pop_mech *)myalloc(sizeof(struct pop_mech));
  pm->name   = onode_getpar_chr_exit(o,"name");
  pm->type   = onode_getpar_chr_exit(o,"type");
  pm->o = o;
  pop_util_mech_config_mask(mylogf,pm);
  return pm;
}
struct pop_mech *mod_pop_make_mech_si_pair_rfdist(o)
     struct onode *o;     
{
  char *nonlin;
  struct pop_mech *pm;
  pm = (struct pop_mech *)myalloc(sizeof(struct pop_mech));
  pm->name   = onode_getpar_chr_exit(o,"name");
  pm->type   = onode_getpar_chr_exit(o,"type");
  pm->o = o;
  nonlin    = onode_getpar_chr_exit(o,"nonlin");  
  if (strcmp(nonlin,"mask")==0){
    pop_util_mech_config_mask(mylogf,pm);
  }else if (strcmp(nonlin,"symmask")==0){
    pop_util_mech_config_mask(mylogf,pm);
  }else if ((strcmp(nonlin,"mult")==0)||(strcmp(nonlin,"divide")==0)){
    pop_util_mech_config_mask(mylogf,pm);
  }else
    exit_error("MOD_POP_MAKE_MECH_SI_PAIR_RFDIST","Unknown mask type");
  myfree(nonlin);
  return pm;
}
struct pop_mech *mod_pop_make_mech_multiply(o)
     struct onode *o;  
{
  struct pop_mech *pm;
  pm = (struct pop_mech *)myalloc(sizeof(struct pop_mech));
  pm->o = o;
  pm->normv = 1.0;  
  pm->a = 1.0;      
  pm->b = 1.0;      
  pm->type = strdup("multiply");
  return pm;
}
struct pop_mech *mod_pop_make_mech_divide(o)
     struct onode *o;  
{
  struct pop_mech *pm;
  pm = (struct pop_mech *)myalloc(sizeof(struct pop_mech));
  pm->o = o;
  pm->normv = 1.0;  
  pm->a = onode_getpar_flt_exit(o,"div_a",1.0);
  pm->b = onode_getpar_flt_exit(o,"div_b",1.0);
  pm->type = strdup("divide");
  return pm;
}
void mod_pop_make_mech(po,pl)
     struct onode *po;     
     struct pop_layer *pl; 
{
  int k;
  struct onode *t;
  char *mtype;
  pl->nmech = onode_count_otype(po,"mech");
  pl->mech = (struct pop_mech **)myalloc(pl->nmech*sizeof(struct pop_mech *));
  k = 0;
  t = po->o;
  while(t != NULL){
    if (strcmp(t->otype,"mech")==0){
      mtype = onode_getpar_chr_exit(t,"type");
      if (strcmp(mtype,"psg_alpha")==0)
	pl->mech[k] = mod_pop_make_mech_psg_alpha(t);
      else if (strcmp(mtype,"psg_doe")==0)
	pl->mech[k] = mod_pop_make_mech_psg_doe(t);
      else if (strcmp(mtype,"si_ds02")==0)
	pl->mech[k] = mod_pop_make_mech_si_ds02(t);
      else if (strcmp(mtype,"si_pair_rfdist")==0)
	pl->mech[k] = mod_pop_make_mech_si_pair_rfdist(t);
      else if (strcmp(mtype,"si_nonlin")==0)
	pl->mech[k] = mod_pop_make_mech_si_pair_rfdist(t);
      else if (strcmp(mtype,"multiply")==0)
	pl->mech[k] = mod_pop_make_mech_multiply(t);
      else if (strcmp(mtype,"divide")==0)
	pl->mech[k] = mod_pop_make_mech_divide(t);
      else{
	printf("  *** mech %s\n",mtype);
	exit_error("MOD_POP_MAKE_MECH","Unknown mechanism");
      }
      myfree(mtype);
      k += 1;
    }
    t = t->next;
  }
}
void mod_pop_init_geom(o,po,pl)
     struct onode *o;    
     struct onode *po;   
     struct pop_layer *pl;
{
  char *gsource;
  struct onode *geo;
  mylog(mylogf,"  MOD_POP_INIT_GEOM\n");
  popu_init_geom(mylogf,o,po,pl,mpt->retflag);  
  geo = onode_child_get_unique(po,"geometry");
  if (strcmp(pl->geomt,"irregular")==0){
    gsource = onode_getpar_chr_dflt(geo,"source",NULL);
    if (strcmp(gsource,"retina0")==0){
      pl->irr_n = mod_mesh_get_mm_mosaic_cn();
      pl->irr_x  = copy_farray(mod_mesh_ptr_mm_mosaic_cx(),pl->irr_n);
      pl->irr_y  = copy_farray(mod_mesh_ptr_mm_mosaic_cy(),pl->irr_n);
      pl->irr_id = copy_iarray(mod_mesh_ptr_mm_mosaic_cid(),pl->irr_n);
      pl->irr_dens = mod_mesh_get_mm_mosaic_dens_deg();  
      if (pl->area != NULL){
	pl->area->umx = 1000.0 / mod_mesh_get_mm_mosaic_degpmm() * mpt->sscale;
	pl->area->umy = pl->area->umx;
      }
      pl->xn = pl->irr_n;
      pl->yn = 1;
      if (mpt->retflag == 100)
	mpt->retflag = 101;
    }else{
      mylogx(mylogf,"MOD_POP_INIT_GEOM","Unknown 'source' for irreg geom");
    }
  }
}
void mod_pop_init_pop(o,po,r,pl)
     struct onode *o;    
     struct onode *po;   
     struct response_struct *r;
     struct pop_layer *pl;
{
  int i,j,k;
  int nattrib,cidflag;
  float tx,ty;
  char tstr[SLEN],*name_ptr,**tattrib,*maptype;
  struct pop_cell ***c,*cp;
  struct onode *spko,*t;
  struct pop_map *orimap,*odmap,*rfxmap,*sfmap;
  mylog(mylogf,"  MOD_POP_INIT_POP\n");
  pl->nmda_p = NULL;
  pl->ifcp = NULL;
  pl->poissp = NULL;
  spko = onode_child_get_unique(po,"spike_gen");
  if (spko != NULL){
    if (onode_test_chr(spko,"type","ifc") == 1){
      pl->ifcp = ifc_util_get_param_o(spko);  
    }else if (onode_test_chr(spko,"type","bernoulli") == 1){
      pl->ifcp = ifc_util_get_param_poiss_o(spko);  
      pl->poissp = ifc_util_get_param_poisson(spko); 
      dll_string_insert(pl->save_name_list,"rate");  
    }else if (onode_test_chr(spko,"type","poisson") == 1){
      pl->ifcp = NULL;
      pl->poissp = NULL;
    }else{
      sprintf(ggstr,"*** <pop> = %s, <spike_gen> type = %s\n",
	      onode_getpar_chr_ptr(po,"name"),
	      onode_getpar_chr_ptr(spko,"type"));
      mylog(mylogf,ggstr);
      mylog_exit(mylogf,"MOD_POP_INIT_POP  Unknown type in <spike_gen>\n");
    }
  }
  nattrib = 7;
  tattrib = (char **)myalloc(nattrib*sizeof(char *));
  tattrib[0] = strdup("ori");
  tattrib[1] = strdup("ori_cv");
  tattrib[2] = strdup("phase");
  tattrib[3] = strdup("ocdom");
  tattrib[4] = strdup("sz1");
  tattrib[5] = strdup("conetype");
  tattrib[6] = strdup("sf");
  pop_cell_attrib_init_f(pl,nattrib,tattrib,0); 
  free_2d_carray(tattrib,nattrib);
  if (onode_test_chr(po,"attrib_conetype","012SML") == 1)
    cidflag = 1;
  else
    cidflag = 0;
  mod_pop_make_mech(po,pl);
  c = (struct pop_cell ***)myalloc(pl->xn*sizeof(struct pop_cell **));
  for(i=0;i<pl->xn;i++){
    c[i] = (struct pop_cell **)myalloc(pl->yn*sizeof(struct pop_cell *));
    for(j=0;j<pl->yn;j++){
      if (j%2 == 1)
	tx = (pl->x0 + (float)i * pl->xf + pl->oddxoff) * pl->area->umx;
      else
	tx = (pl->x0 + (float)i * pl->xf) * pl->area->umx;
      ty = (pl->y0 + (float)j * pl->yf) * pl->area->umy;
      c[i][j] = (struct pop_cell *)myalloc(pl->zn*sizeof(struct pop_cell));
      for(k=0;k<pl->zn;k++){    
	cp = &(c[i][j][k]);     
	pop_cell_init_cell(cp); 
	mod_pop_config_cell(o,po,pl,cp,r,i,j,k);  
	if (strcmp(pl->geomt,"irregular")==0){
	  cp->cx = pl->irr_x[i];
	  cp->cy = pl->irr_y[i];
	  if (cidflag == 1){
	    pop_cell_attrib_set(cp,"conetype",(float)pl->irr_id[i],
				"MOD_POP_INIT_POP");
	  }
	}else{
	  cp->cx = tx;
	  cp->cy = ty;
	}
	cp->rfx = pl->area->x0 + (pl->x0 + (float)i*pl->xf) * pl->area->xf;
	cp->rfy = pl->area->y0 + (pl->y0 + (float)j*pl->yf) * pl->area->yf;
      }
    }
  }
  pl->c = c;
  name_ptr = onode_getpar_chr_ptr(po,"map_ori");
  if (name_ptr != NULL){
    orimap = pop_util_get_map_by_name(mpt,name_ptr);
    if (orimap == NULL)
      mylog_exit(mylogf,"MOD_POP_INIT_POP  orimap not found\n");
    if (orimap->map2 == NULL){ 
      mod_conn_onode_make_map_ori(mylogf,orimap,pl->area->xn,pl->area->yn);
    }
    maptype = onode_getpar_chr_ptr(orimap->o,"type");
    if (strcmp(maptype,"dir_1")==0)
      mod_conn_set_layer_ori_map(mylogf,pl,orimap,0);  
    else
      mod_conn_set_layer_ori_map(mylogf,pl,orimap,1);
  }
  name_ptr = onode_getpar_chr_ptr(po,"map_od");
  if (name_ptr != NULL){
    odmap = pop_util_get_map_by_name(mpt,name_ptr);
    if (odmap == NULL)
      mylog_exit(mylogf,"MOD_POP_INIT_POP  odmap not found\n");
    if (odmap->map2 == NULL){ 
      mod_conn_onode_make_map_od(mylogf,odmap,pl->area->xn,pl->area->yn);
    }
    mod_conn_set_layer_od_map(mylogf,pl,odmap);
  }
  name_ptr = onode_getpar_chr_ptr(po,"map_rfx");
  if (name_ptr != NULL){
    rfxmap = pop_util_get_map_by_name(mpt,name_ptr);
    if (rfxmap == NULL)
      mylog_exit(mylogf,"MOD_POP_INIT_POP  rfxmap not found\n");
    if (rfxmap->map2 == NULL){ 
      mod_conn_onode_make_map_rfx(mylogf,rfxmap,pl->area->xn,pl->area->yn,
				  pl->area->xf);
    }
    mod_conn_set_layer_rfx_map(mylogf,pl,rfxmap);
  }
  name_ptr = onode_getpar_chr_ptr(po,"map_dir_flag");
  if (name_ptr != NULL){
    mod_conn_set_layer_dir_map_flag(mylogf,pl,name_ptr);
  }
  name_ptr = onode_getpar_chr_ptr(po,"map_sf");
  if (name_ptr != NULL){
    sfmap = pop_util_get_map_by_name(mpt,name_ptr);
    if (sfmap == NULL)
      mylog_exit(mylogf,"MOD_POP_INIT_POP  sfmap not found\n");
    if (sfmap->map2 == NULL){ 
      mod_conn_onode_make_map_sf(mylogf,sfmap,pl->area->xn,pl->area->yn);
    }
    mod_conn_set_layer_sf_map(mylogf,pl,sfmap);  
  }
}
void mod_pop_init_nmda(pl,o,po,m)
     struct pop_layer *pl;       
     struct onode *o;            
     struct onode *po;           
     struct model_struct *m;
{
  float vt;
  struct pop_layer *lay;
  struct nmda_param *tp;
  tp = (struct nmda_param *)myalloc(sizeof(struct nmda_param));
  pl->nmda_p = tp;  
  tp->nmda_ppmv = 10;
  tp->nmda_v0 = -100;
  tp->nmda_v0f = (float)tp->nmda_v0;
  pop_util_compute_nmda_table(tp->nmda_ppmv,tp->nmda_v0,100,
			      &(tp->nmda_gv),&(tp->nmda_n));
  tp->nmda_frac      = onode_getpar_flt_exit(po,"nmda_frac");
  tp->nmda_alpha_1   = onode_getpar_flt_exit(po,"nmda_alpha_1");
  tp->nmda_amp_1     = 1.0; 
  tp->nmda_alpha_2   = onode_getpar_flt_exit(po,"nmda_alpha_2");
  tp->nmda_amp_2     = onode_getpar_flt_exit(po,"nmda_amp_2");
  tp->nmda_alpha_3   = onode_getpar_flt_exit(po,"nmda_alpha_3");
  tp->nmda_amp_3     = onode_getpar_flt_exit(po,"nmda_amp_3");
  tp->lgn_nmda_frac  = onode_getpar_flt_exit(po,"lgn_nmda_frac");
  tp->lgn_nmda_amp_1 = tp->nmda_amp_1;
  tp->lgn_nmda_amp_2 = tp->nmda_amp_2;
  tp->lgn_nmda_amp_3 = tp->nmda_amp_3;
  vt = pl->c[0][0][0].cifcp->v_th_x;
  pop_util_compute_intcurr_ampa_nmda_layer(myid,mylogf,pl,vt,mpt->out_prefix);
  pop_util_nmda_prep_constants_layer(pl);
  pop_util_compute_intcurr_ampa_nmda_lgn(myid,mylogf,pl,m,vt,mpt->out_prefix);
}
struct dll_slist *mod_pop_make_save_name_list()
{
  struct dll_slist *slist;
  slist = dll_string_get_init();
  dll_string_insert(slist,"vm");
  dll_string_insert(slist,"spikes");
  return slist;
}
void mod_pop_lay_resp_csav(pl,po)
     struct pop_layer *pl;     
     struct onode *po;         
{
  int k;
  int nresp;
  struct onode *t;
  mylog(mylogf,"  MOD_POP_LAY_RESP_CSAV\n");
  nresp = onode_count_otype(po,"response");        
  if (nresp > 0){
    if (pl->csav_n != 0)
      exit_error("MOD_POP_LAY_RESP_CSAV","csav_n is not zero");
    sprintf(ggstr,"    %d response definitions\n",nresp);
    mylog(mylogf,ggstr);
    pl->csav_n = nresp;
    pl->csav_datid = (char **)myalloc(pl->csav_n * sizeof(char *));
    pl->csav_source = (char **)myalloc(pl->csav_n * sizeof(char *));
    pl->csav_samp = (float *)myalloc(pl->csav_n * sizeof(float));
    k = 0;
    t = onode_get_next_type(po->o,"response");
    while(t != NULL){
      pl->csav_datid[k]  = onode_getpar_chr_exit(t,"data_id");
      pl->csav_source[k] = onode_getpar_chr_exit(t,"source");
      pl->csav_samp[k]   = onode_getpar_flt_exit(t,"sampling");
      dll_string_insert(pl->save_name_list,pl->csav_datid);
      k += 1;
      t = onode_get_next_type(t->next,"response");
    }
  }
}
struct pop_layer *mod_pop_make_pop_layer(o,po,m,r)
     struct onode *o;          
     struct onode *po;         
     struct model_struct *m;
     struct response_struct *r;
{
  int biflag;
  struct pop_layer *pl;
  struct onode *t,*ico;
  char *aname,tstr[SLEN];
  mylog(mylogf,"  MOD_POP_MAKE_POP_LAYER\n");
  t = onode_get_item_child(po,"name");
  sprintf(ggstr,"    Making layer %s\n",t->val);
  mylog(mylogf,ggstr);
  pl = popu_make_pop_layer(NULL);  
  pl->name = onode_getpar_chr_exit(po,"name");
  aname = onode_getpar_chr_ptr_exit(po,"area");
  pl->area = pop_util_get_area_by_name(mpt,aname); 
  pl->laytype = onode_getpar_chr_dflt(po,"type","default");
  pl->icon = (struct pop_icon *)myalloc(sizeof(struct pop_icon));
  pl->icon->r1 = pl->icon->g1 = pl->icon->b1 = -1.0;
  ico = onode_child_get_unique(po,"icon");
  if (ico == NULL){
    pl->icon->shape = strdup("circle");
    pl->icon->r = pl->icon->g = pl->icon->b = 0.5;
    pl->icon->nside = 0;
    pl->icon->o = NULL;
  }else{
    pl->icon->shape = onode_getpar_chr_dflt(ico,"shape","circle");
    pl->icon->r     = onode_getpar_flt_dflt(ico,"r",0.5);
    pl->icon->g     = onode_getpar_flt_dflt(ico,"g",0.5);
    pl->icon->b     = onode_getpar_flt_dflt(ico,"b",0.5);
    if (strcmp(pl->icon->shape,"cs")==0){
      pl->icon->r1     = onode_getpar_flt_dflt(ico,"r1",0.5);
      pl->icon->g1     = onode_getpar_flt_dflt(ico,"g1",0.5);
      pl->icon->b1     = onode_getpar_flt_dflt(ico,"b1",0.5);
    }else if (strcmp(pl->icon->shape,"polygon")==0){
      pl->icon->nside = onode_getpar_flt_exit(ico,"nside");
    }else if (strcmp(pl->icon->shape,"stellate")==0){
      pl->icon->nside = onode_getpar_flt_exit(ico,"npoint");
    }
    pl->icon->o     = ico;
  }
  mod_pop_init_geom(o,po,pl);  
  pl->save_name_list = mod_pop_make_save_name_list(); 
  dll_string_insert(pl->save_name_list,"gad");
  dll_string_insert(pl->save_name_list,"in_gi");
  dll_string_insert(pl->save_name_list,"ex_gx");
  mod_pop_lay_resp_csav(pl,po);
  if (strcmp(pl->laytype,"default") == 0){
    sprintf(tstr,"    Default pop initialization for %s\n",pl->name);
    mylog(mylogf,tstr);
    pl->runflag = onode_getpar_int_dflt(po,"run_flag",1);
    mod_pop_init_pop(o,po,r,pl);  
    if (pl->ifcp->g_tran > 0.0){
      mod_pop_init_nmda(pl,o,po,m);
    }
  }else if ((strcmp(pl->laytype,"virtual")==0) || pop_util_lclass_is_rgci(pl)){
    sprintf(tstr,"    Virtual pop initialization for %s\n",pl->name);
    mylog(mylogf,tstr);
    pl->runflag = 0;
    mod_pop_init_pop(o,po,r,pl);
  }else if (pop_util_lclass_is_lgn0(pl) ||   
	    ((strcmp(pl->laytype,"retina0_gc0") == 0) && (mpt->retflag==100))){
    sprintf(tstr,"    LGN DOG pop initialization for %s\n",pl->name);
    mylog(mylogf,tstr);
    biflag = onode_getpar_int_exit(po,"binocular",0);
    if (biflag == 1){
      mpt->binoc = 1; 
      if (pl->zn != 4)
	mylog_exit(mylogf,"MOD_POP_MAKE_POP_LAYER  Use zn 4 for binoc lgn");
    }
    pl->runflag = 0;  
    mod_pop_init_pop_lgn(o,po,pl);
  }else{
    mylog_exit(mylogf,"MOD_POP_MAKE_POP_LAYER  No initialization");
  }
  return pl;
}
void mod_pop_si_config_ds02(o,po,io,msi,inindex)
     struct onode *o;       
     struct onode *po;      
     struct onode *io;      
     struct pop_mech *msi;  
     short inindex;         
{
  struct pop_layer *lay1,*lay2;
  mylog(mylogf,"    MOD_POP_SI_CONFIG_DS02\n");
  lay2 = popu_get_lay_named(mpt,po,"name","MOD_POP_SI_CONFIG_DS02 lay2");
  lay1 = popu_get_lay_named(mpt,io,"pop_origin","MOD_POP_SI_CONFIG_DS02 lay1");
  mod_conn_syn_convert_layer(mylogf,lay1,lay2,msi,inindex);
}
void mod_pop_si_config_pair_rfdist(o,po,io,msi,inindex)
     struct onode *o;       
     struct onode *po;      
     struct onode *io;      
     struct pop_mech *msi;  
     short inindex;         
{
  struct pop_layer *lay1,*lay2;
  mylog(mylogf,"    MOD_POP_SI_CONFIG_PAIR_RFDIST\n");
  lay2 = popu_get_lay_named(mpt,po,"name",
			    "MOD_POP_SI_CONFIG_PAIR_RFDIST lay2");
  lay1 = popu_get_lay_named(mpt,io,"pop_origin",
			    "MOD_POP_SI_CONFIG_PAIR_RFDIST lay1");
  mod_conn_syn_convert_layer(mylogf,lay1,lay2,msi,inindex);
}
void mod_pop_si_input_main(o,po,io,msi,inindex)
     struct onode *o;       
     struct onode *po;      
     struct onode *io;      
     struct pop_mech *msi;  
     short inindex;         
{
  mylog(mylogf,"  MOD_POP_SI_INPUT_MAIN\n");
  if (strcmp(msi->type,"si_ds02")==0){
    mod_pop_si_config_ds02(o,po,io,msi,inindex);
  }else if (strcmp(msi->type,"si_pair_rfdist")==0){
    mod_pop_si_config_pair_rfdist(o,po,io,msi,inindex);
  }else{
    sprintf(ggstr,"*** %s\n",msi->type);
    mylog(mylogf,ggstr);
    mylog_exit(mylogf,"MOD_POP_SI_INPUT_MAIN  Unknown SI type\n");
  }
}
void mod_pop_input_config_bg(io,pl)
     struct onode *io;       
     struct pop_layer *pl;   
{
  int i,j,k;
  float xrate,irate,xamp,iamp,xcfr,icfr;
  struct pop_cell *c;
  xrate = onode_getpar_flt_dflt(io,"ex_rate",0.0);
  irate = onode_getpar_flt_dflt(io,"in_rate",0.0);
  xamp  = onode_getpar_flt_dflt(io,"ex_amp",1.0);
  iamp  = onode_getpar_flt_dflt(io,"in_amp",1.0);
  xcfr  = onode_getpar_flt_dflt(io,"ex_corr_frac",0.0);
  icfr  = onode_getpar_flt_dflt(io,"in_corr_frac",0.0);
  for(i=0;i<pl->xn;i++){
    for(j=0;j<pl->yn;j++){
      for(k=0;k<pl->zn;k++){    
	c = &(pl->c[i][j][k]);  
	c->bg_x_rate = xrate;
	c->bg_i_rate = irate;
	c->bg_x_amp  = xamp;
	c->bg_i_amp  = iamp;
	c->bg_x_s = c->bg_i_s = NULL;
	c->bg_x_corr_c = c->bg_i_corr_c = NULL;
	c->bg_x_n = c->bg_i_n = -1; 
	c->bg_x_corr_f = xcfr;
	c->bg_i_corr_f = icfr;
      }
    }
  }
}
char *mod_pop_get_conn_fname_read(mpt,io)
     struct pop_top *mpt;    
     struct onode *io;       
{
  char *fname,*rname;
  char tname[LONG_SLEN];
  rname = NULL;
  if (strcmp(mpt->conn_rw,"read")==0){  
    fname = onode_getpar_chr_ptr(io,"file_read"); 
    if (fname != NULL){ 
      if (mpt->conn_read_dir != NULL){ 
	sprintf(tname,"%s/%s",mpt->conn_read_dir,fname);
      }else{
	sprintf(tname,"%s",fname);
      }
      rname = strdup(tname);  
    }
  }
  return rname;
}
char *mod_pop_get_conn_fname_write(mpt,io)
     struct pop_top *mpt;    
     struct onode *io;       
{
  char *fname,*rname;
  char tname[LONG_SLEN];
  rname = NULL;
  if (strcmp(mpt->conn_rw,"write")==0){  
    fname = onode_getpar_chr_ptr(io,"file_write"); 
    if (fname != NULL){ 
      if (mpt->conn_write_dir != NULL){ 
	sprintf(tname,"%s/%s",mpt->conn_write_dir,fname);
      }else{
	sprintf(tname,"%s",fname);
      }
      rname = strdup(tname);  
    }
  }
  return rname;
}
void mod_pop_input(o,po,io,inindex)
     struct onode *o;    
     struct onode *po;   
     struct onode *io;   
     short inindex;      
{
  int i,j,k;
  int syntype,lii;
  float sig,sscale,axonv,axondt,normw;
  char *intype,*tname,*dtype,*wfile,*si_name,*pre_pop_name,*fname;
  char *ori_pop,tstr[SLEN];
  struct pop_layer *lay1,*lay2,*lay,*layo1,*layo2;
  struct pop_mech *mechr,*msi,*mechp;
  struct pop_cell *c;
  struct onode *disto,*t;
  mylog(mylogf,"  MOD_POP_INPUT\n");
  lay2 = popu_get_lay_named(mpt,po,"name","MOD_POP_INPUT post-syn");
  intype = onode_getpar_chr_dflt(io,"type","regular");
  if (strcmp(intype,"regular")==0){
    lay1 = popu_get_lay_named(mpt,io,"pop_origin","MOD_POP_INPUT regular");
    mechr = popu_get_mech_named(mpt,lay2,io,"receptor","MOD_POP_INPUT");
    syntype = onode_getpar_int_exit(mechr->o,"syntype");
    disto = onode_child_get_unique(io,"distrib");
    if (disto == NULL){
      mylog_exit(mylogf,"  MOD_POP_INPUT  No distrib for input.\n");
    }
    fname = mod_pop_get_conn_fname_read(mpt,io);
    if (fname != NULL){
      axondt = onode_getpar_flt_dflt(io,"axon_dt",0.0); 
      axonv = onode_getpar_flt_dflt(io,"axon_vel",0.0); 
      mod_conn_read_conn_t1(mylogf,lay1,lay2,fname,axonv,axondt,inindex);
      myfree(fname);
    }else{
      dtype = onode_getpar_chr_ptr_exit(disto,"type");
      if (strcmp(dtype,"ori_dist")==0){
	mod_conn_onode_ori_dist_01(mylogf,io,lay1,lay2,syntype,inindex);
      }else if (strcmp(dtype,"corr_lgn")==0){
	tname = onode_getpar_chr_dflt(disto,"lgn_pname","lgn");
	t = onode_get_node_type_item_val(o,"pop","name",tname); 
	myfree(tname);
	if (t == NULL)
	  mylog_exit(mylogf,"MOD_POP_INPUT  Cannot find 'lgn_pname'.");
	sig = onode_getpar_flt_exit(t,"sig1"); 
	sscale = mpt->sscale;
	mod_conn_onode_corr_on_off(mylogf,io,lay2,lay1,syntype,sig,sscale,
				   mpt->xn,mpt->yn,inindex); 
      }else if (strcmp(dtype,"binoc_mask")==0){
	mylog(mylogf,"********* WARNING - ASSUMING 'lgn' name for layer.\n");
	t = onode_get_node_type_item_val(o,"pop","name","lgn");
	if (t == NULL)
	  mylog_exit(mylogf,"MOD_POP_INPUT  Cannot find pop 'lgn'.");
	sig = onode_getpar_flt_exit(t,"sig1");
	sscale = mpt->sscale;
	mod_conn_onode_binoc_mask(mylogf,io,lay2,lay1,syntype,sig,sscale,
				  mpt->xn,mpt->yn,1,inindex); 
      }else if (strcmp(dtype,"gabor_mask")==0){
	pre_pop_name = onode_getpar_chr_ptr_exit(disto,"mask_pop_pre");
	t = onode_get_node_type_item_val(o,"pop","name",pre_pop_name);
	if (t == NULL)
	  mylog_exit(mylogf,"MOD_POP_INPUT  Cannot find 'mask_pop_pre'.");
	sig = onode_getpar_flt_exit(t,"sig1");
	sscale = mpt->sscale;
	mod_conn_onode_binoc_mask(mylogf,io,lay2,lay1,syntype,sig,sscale,
				  mpt->xn,mpt->yn,0,inindex); 
      }else{
	sprintf(ggstr,"*** %s\n",dtype);
	mylog(mylogf,ggstr);
	mylog_exit(mylogf,"MOD_POP_INPUT  Unknown distrib type\n");
      }
    }
    wfile = mod_pop_get_conn_fname_write(mpt,io);
    if (wfile != NULL){
      mod_conn_write_conn_t1(mylogf,lay1,lay2,syntype,wfile);
      myfree(wfile);
    }
    if (onode_get_next_type(io->o,"input_pair") != NULL){
      struct input_pair *ip;
      ip = pop_util_input_pair(mpt,lay2,io,inindex);
      mod_conn_input_pair_main(mpt,lay2,io,ip);
      if (ip->z1 != NULL)	myfree(ip->z1);
      if (ip->distrib != NULL)	myfree(ip->distrib);
      myfree(ip);
    }
    si_name = onode_getpar_chr_ptr(io,"syn_int");
    if (si_name != NULL){
      msi = pop_cell_get_mech_by_name(lay2,si_name); 
      mylog(mylogf," *** Using old method for SYN_INT PAIRING\n");
      if (msi == NULL){
	sprintf(ggstr,"*** Mechanism %s\n",si_name);
	mylog(mylogf,ggstr);
	sprintf(ggstr,"*** Layer %s\n",lay2->name);
	mylog(mylogf,ggstr);
	mylog_exit(mylogf,"MOD_POP_INPUT  Mechanism not found in layer.");
      }
      mod_pop_si_input_main(o,po,io,msi,inindex);
    }
    sprintf(tstr,"all_spk_%s",lay1->name);
    dll_string_insert(lay2->save_name_list,tstr);
  }else if (strcmp(intype,"lgn_on_off")==0){
    lay2->lgn_in_flag = 1; 
    lay1 = popu_get_lay_named(mpt,io,"pop_origin","MOD_POP_INPUT lgn_on_off");
    mechr = popu_get_mech_named(mpt,lay2,io,"receptor","MOD_POP_INPUT");
    lay2->lgn_gain     = onode_getpar_int_dflt(io,"gainflag",0);
    lay2->lgn_gain_tau = onode_getpar_flt_dflt(io,"gain_tau",0.010);
    lay2->lgn_gain_ca  = onode_getpar_flt_dflt(io,"gain_ca",1.0);
    lay2->lgn_gain_cb  = onode_getpar_flt_dflt(io,"gain_cb",1.0);
    if (lay2->lgn_gain == 1)
      lay1->lgn_gain = 1;  
    fname = mod_pop_get_conn_fname_read(mpt,io);
    if (fname != NULL){
      if (strcmp(lay1->geomt,"irregular")==0){
	axondt = onode_getpar_flt_dflt(io,"axon_dt",0.0); 
	axonv = onode_getpar_flt_dflt(io,"axon_vel",0.0); 
	mod_conn_read_conn_t1(mylogf,lay1,lay2,fname,axonv,axondt,inindex);
      }else{
	normw = onode_getpar_flt_dflt(io,"normw",-1.0);
	if (normw > 0.0)
	  mylog(mylogf,"    NOTE: applying 'normw' to LGN inputs\n");
	mod_conn_read_conn_t0(mylogf,lay1,lay2,mechr,fname,normw);
      }
      mod_conn_onode_layer_gabor_map_set_phase(mylogf,io,lay2,mpt->sscale);
    }else{
      mod_conn_onode_layer_gabor_map(mylogf,io,lay1,lay2,mechr,
				     mpt->xn,mpt->yn,mpt->sscale,inindex);
    }
    wfile = mod_pop_get_conn_fname_write(mpt,io);
    if (wfile != NULL){
      if (strcmp(lay1->geomt,"irregular")==0){
	mod_conn_write_conn_t1(mylogf,lay1,lay2,11,wfile);  
      }else{
	lii = lay2->lgn_in_n;  
	mod_conn_write_conn_t0(mylogf,lay2,lii,1,wfile); 
      }
      myfree(wfile);
    }
    dll_string_insert(lay2->save_name_list,"lgn_ga");
    if (lay2->nmda_p != NULL){
      dll_string_insert(lay2->save_name_list,"lgn_gn");
    }
    if (lay2->lgn_gain == 1){
      dll_string_insert(lay2->save_name_list,"lgn_gain");
    }
    if (strcmp(lay1->geomt,"irregular")==0){
      lay2->lgn_in_n = 0;  
    }else
      lay2->lgn_in_n += 1; 
  }else if (strcmp(intype,"lgn_pair")==0){
    lay2->lgn_in_flag = 1; 
    lay2->lgn_gain = onode_getpar_int_dflt(io,"gainflag",0);
    layo1 = popu_get_lay_named(mpt,io,"pop_origin_1","MOD_POP_INPUT lgn_pair");
    layo2 = popu_get_lay_named(mpt,io,"pop_origin_2","MOD_POP_INPUT lgn_pair");
    mechr = popu_get_mech_named(mpt,lay2,io,"receptor","MOD_POP_INPUT");
    mechp = popu_get_mech_named(mpt,lay2,io,"pair_mech","MOD_POP_INPUT");
    fname = mod_pop_get_conn_fname_read(mpt,io);
    if (fname != NULL){
      myfree(fname);
      exit_error("MOD_POP_INPUT","FILE not imp'd yet for 'lgn_pair'");
    }else
      mod_conn_lgn_pair(mpt,io,lay2,layo1,layo2,mechr,mechp);
    wfile = mod_pop_get_conn_fname_write(mpt,io);
    if (wfile != NULL){
      exit_error("MOD_POP_INPUT","FILE_WRITE not imp'd yet for 'lgn_pair'");
      myfree(wfile);
    }
    dll_string_insert(lay2->save_name_list,"lgn_ga");
    if (lay2->nmda_p != NULL){
      dll_string_insert(lay2->save_name_list,"lgn_gn");
    }
    if (lay2->lgn_gain == 1)
      dll_string_insert(lay2->save_name_list,"lgn_gain");
    lay2->lgn_in_n += 1; 
  }else if (strcmp(intype,"mesh_bp_to_rgc")==0){
    lay1 = popu_get_lay_named(mpt,io,"pop_origin",
			      "MOD_POP_INPUT mesh_bp_to_rgc");
    {
      int i,j,k,l;
      int xn,yn,zn,xi,zi,cn,*ci,nsyn;
      float *cw;
      struct pop_cell *cpre,*cpost;
      xn = lay2->xn;
      yn = lay2->yn;
      zn = lay2->zn;
      nsyn = 0;
      for(i=0;i<xn;i++){  
	for(j=0;j<yn;j++){  
	  for(k=0;k<zn;k++){  
	    cpost = &(lay2->c[i][j][k]);  
	    if (cpost == NULL){
	      printf("******  cpost is null\n");
	      printf("******  %d %d %d\n",i,j,k);
	      exit(0);
	    }
	    zi = k;
	    mod_mesh_get_rgc_conn(i,zi,&cn,&ci,&cw);  
	    for(l=0;l<cn;l++){ 
	      xi = ci[l];
	      cpre = &(lay1->c[xi][0][zi]);  
	      (void)pop_cell_add_synapse(cpre,cpost,1,cw[l],0.0,inindex);
	      nsyn += 1;
	    }
	    if (cn > 0){
	      myfree(ci);
	      myfree(cw);
	    }
	  }
	}
      }
      sprintf(ggstr,"    %d synapses for 'mesh_bp_to_rgc'\n",nsyn);
      mylog(mylogf,ggstr);
    }
  }else if (strcmp(intype,"bg")==0){
    mod_pop_input_config_bg(io,lay2);
  }else{
    sprintf(ggstr,"  *** %s\n",intype);
    mylog(mylogf,ggstr);
    mylog_exit(mylogf,"MOD_POP_INPUT  Unknown input type\n");
  }
  ori_pop = onode_getpar_chr_dflt(io,"attr_ori_derive_from_pop",NULL);
  if (ori_pop != NULL){
    lay = pop_util_get_layer_by_name(mpt,ori_pop);  
    for(i=0;i<lay2->xn;i++){
      for(j=0;j<lay2->yn;j++){
	for(k=0;k<lay2->zn;k++){ 
	  c = &(lay2->c[i][j][k]);  
	  pop_cell_derive_ori_from_layer(mylogf,lay,c,0);
	}
      }
    }
    myfree(ori_pop);
  }
  ori_pop = onode_getpar_chr_dflt(io,"attr_dir_derive_from_pop",NULL);
  if (ori_pop != NULL){
    lay = pop_util_get_layer_by_name(mpt,ori_pop);  
    for(i=0;i<lay2->xn;i++){
      for(j=0;j<lay2->yn;j++){
	for(k=0;k<lay2->zn;k++){ 
	  c = &(lay2->c[i][j][k]);  
	  pop_cell_derive_ori_from_layer(mylogf,lay,c,1);
	}
      }
    }
    myfree(ori_pop);
  }
  myfree(intype);
}
void mod_pop_set_conn_rw(mpt,o)
     struct pop_top *mpt;    
     struct onode *o;        
{
  char *tdir;
  char ggstr[LONG_SLEN];
  mylog(mylogf,"    MOD_POP_SET_CONN_RW\n");
  mpt->conn_rw = onode_getpar_chr_dflt(o,"mod_conn_rw","none");
  if (strcmp(mpt->conn_rw,"none")==0){
    mpt->conn_read_dir = NULL;  
    mpt->conn_write_dir = NULL;  
    return;
  }else if (strcmp(mpt->conn_rw,"read")==0){
    tdir = onode_getpar_chr_ptr(o,"mod_conn_dir_r");
    if (tdir != NULL){
      if (is_dir(tdir) == 0){
	tdir = onode_getpar_chr_ptr(o,"mod_conn_dir_r2");
	if (tdir != NULL){
	  if (is_dir(tdir) == 0){
	    sprintf(ggstr,"  mod_conn_dir_r2 = %s\n",tdir);
	    mylog(mylogf,ggstr);
	    mylogx(mylogf,"MOD_POP_SET_CONN_RW","mod_conn_dir_r2 not found");
	  }
	}else{
	  tdir = onode_getpar_chr_ptr(o,"mod_conn_dir_r");
	  sprintf(ggstr,"  mod_conn_dir_r = %s\n",tdir);
	  mylog(mylogf,ggstr);
	  mylogx(mylogf,"MOD_POP_SET_CONN_RW","mod_conn_dir_r not found");
	}
      }
    }
    if (tdir == NULL)
      mylog(mylogf,"      No default connections directory for reading.\n");
    else{
      sprintf(ggstr,"      Connections directory for reading:  %s\n",tdir);
      mylog(mylogf,ggstr);
    }
    mpt->conn_read_dir = tdir;
    mpt->conn_write_dir = NULL;
  }else if (strcmp(mpt->conn_rw,"write")==0){
    tdir = onode_getpar_chr_ptr(o,"mod_conn_dir_w");
    if (tdir != NULL){
      if (is_dir(tdir) == 0){
	sprintf(ggstr,"  mod_conn_dir_w = %s\n",tdir);
	mylog(mylogf,ggstr);
	mylogx(mylogf,"MOD_POP_SET_CONN_RW","mod_conn_dir_w not found");
      }
    }
    if (tdir == NULL)
      mylog(mylogf,"      No default connections directory for writing.\n");
    else{
      sprintf(ggstr,"      Connections directory for writing:  %s\n",tdir);
      mylog(mylogf,ggstr);
    }
    mpt->conn_read_dir = NULL;
    mpt->conn_write_dir = tdir;
  }else{
    mylogx(mylogf,"MOD_POP_SET_CONN_RW","Unknown 'mod_conn_rw' value");
  }
}
void mod_pop_prep(m,r,s)
     struct model_struct *m;     
     struct response_struct *r;  
     struct stim_struct *s;      
{
  int i,k;
  int gui_flag,n,xflag,lgn_flag;
  char *tstr;
  struct onode *t,*tin,*wo;
  struct pop_layer *lay;
  myid = mylog_set_id_log(m->process_id,&mylogf);  
  mylog(mylogf,"  MOD_POP_PREP\n");
  mpt = (struct pop_top *)myalloc(sizeof(struct pop_top));
 if  (mylogf == NULL){
   mpt->logf = NULL;
 }else
   mpt->logf = strdup(mylogf);
  mpt->tscale = onode_getpar_flt_exit(m->o,"tscale");
  mpt->sscale = onode_getpar_flt_exit(m->o,"sscale");
  mpt->xn = onode_getpar_int_exit(m->o,"xn");
  mpt->yn = onode_getpar_int_exit(m->o,"yn");
  mpt->tn = onode_getpar_int_exit(m->o,"tn");
  mpt->x1 = 0.0;
  mpt->x2 = mpt->tscale * (float)mpt->tn;
  mpt->out_prefix = NULL;
  if (r != NULL){
    if (r->ppl != NULL)
      mpt->out_prefix = param_getc_dflt(r->ppl,"outfile","zz");
  }
  mpt->binoc = 0;   
  if (onode_child_get_unique(m->o,"retina0") != NULL)
    mpt->retflag = 100; 
  else
    mpt->retflag = 0;   
  mpt->conn_rw = NULL;
  mpt->conn_read_dir = NULL;
  mod_pop_set_conn_rw(mpt,m->o);
  n = onode_count_otype(m->o,"area");
  mpt->area = (struct pop_area **)myalloc(n*sizeof(struct pop_area *));
  mpt->narea = n;
  sprintf(ggstr,"    %d areas\n",mpt->narea);
  mylog(mylogf,ggstr);
  k = 0;
  t = onode_get_next_type(m->o->o,"area");
  while(t != NULL){
    mpt->area[k] = popu_make_area(mylogf,t,mpt->sscale); 
    k += 1;
    t = onode_get_next_type(t->next,"area");
  }
  mpt->nmap = onode_count_otype(m->o,"map");
  mpt->map = (struct pop_map **)myalloc(mpt->nmap*sizeof(struct pop_map *));
  sprintf(ggstr,"    %d maps\n",mpt->nmap);
  mylog(mylogf,ggstr);
  k = 0;
  t = onode_get_next_type(m->o->o,"map");
  while(t != NULL){
    mpt->map[k] = (struct pop_map *)myalloc(sizeof(struct pop_map));
    mpt->map[k]->o  = t; 
    mpt->map[k]->xn = 0;
    mpt->map[k]->yn = 0;
    mpt->map[k]->zn = 0;
    mpt->map[k]->map2 = NULL;
    pop_util_map_make(mpt,k);  
    k += 1;
    t = onode_get_next_type(t->next,"map");
  }
  wo = onode_child_get_unique(m->o,"warp");
  if (wo != NULL){
    popu_warp_map(mpt,wo);
  }
  mpt->nlay = onode_count_otype(m->o,"pop");
  mpt->lay = (struct pop_layer **)myalloc(mpt->nlay*
					  sizeof(struct pop_layer *));
  sprintf(ggstr,"    %d layers\n",mpt->nlay);
  mylog(mylogf,ggstr);
  k = 0;
  t = onode_get_next_type(m->o->o,"pop");
  while(t != NULL){
    mpt->lay[k] = mod_pop_make_pop_layer(m->o,t,m,r); 
    k += 1;
    t = onode_get_next_type(t->next,"pop");
  }
  for(i=0;i<mpt->nlay;i++){
    lay = mpt->lay[i];
    t = onode_get_node_type_item_val(m->o,"pop","name",lay->name);
    if (t == NULL){
      printf("*** i=%d  %s\n",i,lay->name);
      exit_error("MOD_POP_PREP","Layer name not found");
    }
    lay->ninlist = onode_count_otype(t,"input");
    lay->inlist = (struct onode **)myalloc(lay->ninlist*
					   sizeof(struct onode *));
    k = 0;
    tin = onode_get_next_type(t->o,"input");
    while(tin != NULL){
      mod_pop_input(m->o,t,tin,(short)k);  
      lay->inlist[k] = tin;         
      k += 1;
      tin = onode_get_next_type(tin->next,"input");
    }
  }
  pop_util_customize(m,mpt);
  if ((m->action == 10) && (m->marfile != NULL)){
    popu_mar_write(mpt,m->marfile,4);  
    return;
  }else if ((m->action == 20) && (m->marfile != NULL)){
    popu_rsp_gen(mpt,m,r);
    return;
  }
  lgn_flag = 0;
  for(i=0;i<mpt->nlay;i++){
    lay = mpt->lay[i];
    if (pop_util_lclass_is_lgn0(lay)){
      lgn_flag = 1;
    }
  }
  if ((mpt->retflag != 100) && (mpt->retflag != 101) && (lgn_flag == 0))
    mpt->retflag = -1;  
  if (mpt->retflag == -1){
    pop_util_init_bg_rate(mpt->lay,mpt->nlay);
  }
  if (myid == -1){
    if (paramfile_test_param(r->ppl,"astat_lgn_proj_layer")){
      mod_conn_astat_lgn_postsyn(mpt,r);
      xflag = paramfile_get_int_param_default(r->ppl,"astat_exit",1);
      printf("    Exiting after computing statistics.\n\n");
      exit(0);
    }else if (paramfile_test_param(r->ppl,"astat_lms_input_post")){
      mod_conn_astat_lms_input(mpt,r);
      xflag = paramfile_get_int_param_default(r->ppl,"astat_exit",1);
      printf("    Exiting after computing statistics.\n\n");
      exit(0);
    }else if (paramfile_test_param(r->ppl,"astat_unit_attrib_layer")){
      char *popname,*a_name,*outfile;
      struct param_pair *pp;
      pp = paramfile_get_first_prefix_pointer(r->ppl,"astat_unit_attrib_layer");
      popname = paramfile_get_nth_char_param_pp_or_exit(pp,0);  
      a_name  = paramfile_get_nth_char_param_pp_or_exit(pp,1);  
      outfile = paramfile_get_nth_char_param_pp_or_exit(pp,2);  
      lay = pop_cell_get_layer_pointer(mpt->lay,mpt->nlay,popname);
      popc_attrib_dump_layer(lay,outfile,a_name);
      myfree(popname);
      myfree(a_name);
      myfree(outfile);
    }
  }
  mod_pop_init_lgn_cells_resp(mpt->lay,mpt->nlay,r);
  pop_util_response_link(r,mpt->lay,mpt->nlay,mylogf);
  if (myid == -1){
    gui_flag = paramfile_get_int_param_default(r->ppl,"gui_flag",0);
    if (gui_flag)
      mod_gui_pop_main(mpt->lay,mpt->nlay,mpt,m,s,r);
  }
  if (m->run_mode == 0) 
    mpt->monflag = paramfile_get_int_param_default(r->ppl,"mon_flag",0);
  else
    mpt->monflag = 0;
  if ((myid == -1) && (mpt->monflag)){
    mod_mon_init(m,r,s,mpt->lay,mpt->nlay,"monitor");
  }
}
int **mod_pop_get_lgn_flag(m,r,pname,onflag)
     struct model_struct *m;        
     struct response_struct *r;     
     char *pname;                   
     int onflag;                    
{
  int i;
  int xn,yn,nlay,**flag;
  float corr_d;
  struct pop_layer *lay;
  mylog(mylogf,"  MOD_POP_GET_LGN_FLAG\n");
  nlay = mpt->nlay;
  corr_d = onode_getpar_int_dflt(m->o,"lgn_corr_distance",1);
  xn = mpt->xn;
  yn = mpt->yn;
  if (corr_d <= 1){
    flag = get_zero_2d_iarray(xn,yn);
    for(i=0;i<nlay;i++){ 
      lay = mpt->lay[i];
      if ((lay->runflag == 1) && (lay->lgn_in_flag == 1)){
	printf("    LayerName %s,  lgn_in_n %d \n",lay->name,lay->lgn_in_n);
	pop_util_set_flag_grid_input_layer(mylogf,flag,xn,yn,0,lay,pname);
      }
    }
  }else{ 
    flag = get_const_2d_iarray(xn,yn,1);
  }
  return flag;
}
void mod_pop_get_lgn_spikes(m,r,rdog,lgn_name,seed0,seed1,seed2,
			    rs0,rs1,rcnt0,rcnt1)
     struct model_struct *m;        
     struct response_struct *r;     
     float ***rdog;
     char lgn_name[];               
     int seed0,seed1,seed2;
     float ****rs0,****rs1;
     int ***rcnt0,***rcnt1;
{
  int i;
  int xn,yn,tn,nlay,sv_vn,sav0,sav1;
  int **flag0,**flag1,**cnt0,**cnt1,corr_d;
  float samp,***s0,***s1,corr_p,corr_tsd,periodms;
  float ***sv0gx,***sv0gi,***sv0vm,***sv1gx,***sv1gi,***sv1vm;
  struct pop_layer *lay;
  nlay = mpt->nlay;
  mylog(mylogf,"  MOD_POP_GET_LGN_SPIKES\n");
  corr_d = onode_getpar_int_dflt(m->o,"lgn_corr_distance",1);
  xn = mpt->xn;
  yn = mpt->yn;
  tn = mpt->tn;
  samp = 1.0/mpt->tscale;
  if (corr_d <= 1){
    flag0 = get_zero_2d_iarray(xn,yn); 
    flag1 = get_zero_2d_iarray(xn,yn); 
    for(i=0;i<nlay;i++){ 
      lay = mpt->lay[i];
      if ((lay->runflag == 1) && (lay->lgn_in_flag == 1)){
	pop_util_set_flag_grid_input_layer(mylogf,flag0,xn,yn,0,lay,lgn_name);
	pop_util_set_flag_grid_input_layer(mylogf,flag1,xn,yn,1,lay,lgn_name);
      }
    }
  }else{ 
    flag0 = get_const_2d_iarray(xn,yn,1); 
    flag1 = get_const_2d_iarray(xn,yn,1); 
  }
  sav0 = mod_pop_lgn_set_sav_flag(mpt->lay,mpt->nlay,r,flag0,xn,yn,lgn_name,0);
  sav1 = mod_pop_lgn_set_sav_flag(mpt->lay,mpt->nlay,r,flag1,xn,yn,lgn_name,1);
  mod_dog_get_spikes_2d_flag(m,r,rdog,lgn_name,1,xn,yn,tn,flag1,samp,seed2,
			     &s1,&cnt1,sav1,&sv1gx,&sv1gi,&sv1vm,&sv_vn);
  mod_dog_get_spikes_2d_flag(m,r,rdog,lgn_name,0,xn,yn,tn,flag0,samp,seed2,
			     &s0,&cnt0,sav0,&sv0gx,&sv0gi,&sv0vm,&sv_vn);
  free_2d_iarray(flag0,xn);
  free_2d_iarray(flag1,xn);
  if (corr_d > 1){
    periodms = (float)tn*mpt->tscale * 1000.0; 
    corr_p = onode_getpar_flt_dflt(m->o,"lgn_corr_ppick",1.0);
    corr_tsd = onode_getpar_flt_dflt(m->o,"lgn_corr_tsd",0.0);
    mylog(mylogf,"    Adjusting LGN spikes for correlation.\n");
    pop_util_corr_nearby_spikes(xn,yn,s0,cnt0,periodms,corr_d,corr_p,
				corr_tsd,seed0);
    pop_util_corr_nearby_spikes(xn,yn,s1,cnt1,periodms,corr_d,corr_p,
				corr_tsd,seed1);
  }
  if (sav0 > 0){
    mod_pop_store_lgn_sav(mpt,r,lgn_name,0,sv0gx,sv0gi,sv0vm,sv_vn,tn);
    free_3d_farray_null(sv0gx,xn,yn,0);
    free_3d_farray_null(sv0gi,xn,yn,0);
    free_3d_farray_null(sv0vm,xn,yn,0);
  }
  if (sav1 > 0){
    mod_pop_store_lgn_sav(mpt,r,lgn_name,1,sv1gx,sv1gi,sv1vm,sv_vn,tn);
    free_3d_farray_null(sv1gx,xn,yn,0);
    free_3d_farray_null(sv1gi,xn,yn,0);
    free_3d_farray_null(sv1vm,xn,yn,0);
  }
  *rs0 = s0;  *rcnt0 = cnt0;
  *rs1 = s1;  *rcnt1 = cnt1;
}
void mod_pop_cell_dump_lgn_input(popname,xi,yi,zi,outfile,rdog)
     char *popname;
     int xi,yi,zi;
     char *outfile;
     float ***rdog;   
{
  int i;
  int xn,yn,tn,n1,n0,tx,ty;
  float *tr;
  char tname[SLEN];
  struct pop_layer *lay;
  struct pop_cell *c;
  printf("  MOD_POP_CELL_DUMP_LGN_INPUT\n");
  xn = mpt->xn;
  yn = mpt->yn;
  tn = mpt->tn;
  lay = pop_cell_get_layer_pointer(mpt->lay,mpt->nlay,popname);
  c = &(lay->c[xi][yi][zi]);
  n1 = c->lgnin[0]->cn1;
  n0 = c->lgnin[0]->cn0;
  printf("    %d ON  %d OFF inputs\n",n1,n0);
  for(i=0;i<n1;i++){
    tx = c->lgnin[0]->cx1[i];
    ty = c->lgnin[0]->cy1[i];
    tr = &(rdog[tx+1][ty+1][1]);
    sprintf(tname,"%s_%d_%d_%d_ON_in_%d_%d",popname,xi,yi,zi,tx,ty);
    append_farray_plot(outfile,tname,tr,tn,1);
  }
}
void mod_pop_get_resp_lgn_lay(m,s,r,lay)
     struct model_struct *m;        
     struct stim_struct *s;         
     struct response_struct *r;     
     struct pop_layer *lay;         
{
  int i;
  int xn,yn,zn,seed0,seed1,seed0r,seed1r,seed2,seed2r;
  int **cnt0,**cnt1,**cnt0r,**cnt1r,**sflag,nrsp_flagged;
  float ***s0,***s1,***s0r,***s1r;
  float ***rdog,***rdogr;
  float ***rdogg,***rdoggr;
  float ****s3;
  int ***cnt3;
  mylog(mylogf,"  MOD_POP_GET_RESP_LGN_LAY\n");
  sprintf(ggstr,"    %s\n",lay->name);
  mylog(mylogf,ggstr);
  xn = lay->xn;  
  yn = lay->yn;
  zn = lay->zn;
  seed0  = lay->layseed[2]; 
  seed1  = lay->layseed[3];
  seed0r = lay->layseed[4];
  seed1r = lay->layseed[5];
  seed2  = lay->layseed[6];  
  seed2r = lay->layseed[7];  
  cnt0 = cnt1 = cnt0r = cnt1r = NULL;
  s0 = s1 = s0r = s1r = NULL;
  rdog = rdogr = NULL;
  s3 = NULL;
  cnt3 = NULL;
  if (mpt->binoc == 0){
    if ((mpt->retflag >= 100) && (mpt->retflag <= 101)){   
      mod_mesh_compute_diff_all(m,s,r); 
      if (mpt->retflag == 101){
	int nflagged;
	int ***cmap;
	nflagged = popc_flag_get_lay_to_all(lay,mpt->lay,mpt->nlay,&cmap,-1.0);
	nrsp_flagged = popu_flag_resp(r,lay,cmap);
	mod_mesh_get_spikes_3d_flag(m,s,r,"rgc",cmap,xn,yn,zn,&s3,&cnt3);
	free_3d_iarray(cmap,xn,yn,zn);
      }else{
	sflag = mod_pop_get_lgn_flag(m,r,"rgc",0); 
	mod_mesh_get_spikes_2d_flag(m,s,r,"rgc",0,xn,yn,sflag,&s0,&cnt0);
	free_2d_iarray(sflag,xn);
	sflag = mod_pop_get_lgn_flag(m,r,"rgc",1); 
	mod_mesh_get_spikes_2d_flag(m,s,r,"rgc",1,xn,yn,sflag,&s1,&cnt1);
	free_2d_iarray(sflag,xn);
      }
    }else{ 
      rdog = mod_dog_get_response(s,s->repno,lay->name,lay->lgn_gain);
      if (lay->lgn_gain == 1){
	rdogg = mod_dog_gain_ptr(lay->name,0);  
	rdoggr = NULL;
      }else{
	rdogg = rdoggr = NULL;
      }
      mod_pop_get_lgn_spikes(m,r,rdog,lay->name,seed0,seed1,seed2,&s0,&s1,
			     &cnt0,&cnt1);
    }
  }else{
    if (s->d_r == NULL){
      mylog(mylogf,"  *** MOD_POP_GET_RESPONSE - Using 0 spikes for R-eye.\n");
      s0r = get_3d_farray(xn,yn,0);
      s1r = get_3d_farray(xn,yn,0);
      cnt0r = get_zero_2d_iarray(xn,yn);
      cnt1r = get_zero_2d_iarray(xn,yn);
      rdog = mod_dog_get_response(s,s->repno,lay->name,lay->lgn_gain);
      mod_pop_get_lgn_spikes(m,r,rdog,lay->name,seed0,seed1,seed2,&s0,&s1,
			     &cnt0,&cnt1);
    }else{
      rdog = mod_dog_get_response_binoc(s,s->repno,lay->name,&rdogr);
      mod_pop_get_lgn_spikes(m,r,rdog,lay->name,seed0,seed1,seed2,&s0,&s1,
			     &cnt0,&cnt1);
      mod_pop_get_lgn_spikes(m,r,rdogr,lay->name,seed0r,seed1r,seed2r,
			     &s0r,&s1r,&cnt0r,&cnt1r);
    }
  }
  if (s3 != NULL){   
    if (lay->s != NULL){
      {
	int i1,i2,i3;
	int ***dcnt;
	float ****dd;
	dd = lay->s;
	dcnt = lay->cnt;
	for(i1=0;i1<zn;i1++){
	  for(i2=0;i2<xn;i2++){
	    for(i3=0;i3<yn;i3++)
	      if (lay->cnt[i1][i2][i3] > 0)
		myfree(dd[i1][i2][i3]);
	    myfree(dd[i1][i2]);
	  }
	  myfree(dd[i1]);
	}
	myfree(dd);
      }
    }
    if (lay->cnt != NULL){
      free_3d_iarray(lay->cnt,zn,xn,yn);
    }
    lay->cnt = cnt3;
    lay->s = s3;
  }else{
    for(i=0;i<zn;i++){
      if (lay->cnt[i] != NULL)
	free_2d_farray(lay->cnt[i],xn);
      if (lay->s[i] != NULL){
	free_3d_farray(lay->s[i],xn,yn,0);  
      }
    }
    lay->cnt[0] = cnt0;  lay->s[0] = s0;    
    lay->cnt[1] = cnt1;  lay->s[1] = s1;    
    lay->rl = rdog;
    lay->rgl = rdogg;
    if (lay->zn >= 4){
      lay->cnt[2] = cnt0r;  lay->s[2] = s0r;    
      lay->cnt[3] = cnt1r;  lay->s[3] = s1r;    
      lay->rr = rdogr;
      lay->rgr = rdoggr;
    }
  }
}
void mod_pop_get_response(m,s,r)
     struct model_struct *m;        
     struct stim_struct *s;         
     struct response_struct *r;     
{
  int i;
  int xn,yn,tn,nlay,**sflag;
  int seed,seed1,seed2,*seedlist;
  float ***rdog,***rdogr,***s0,***s1,tnsec;
  char outfile[SLEN],tname[SLEN],lname[SLEN];
  struct pop_layer **laylist;
  struct pop_layer *lay;
  mylog(mylogf,"  MOD_POP_GET_RESPONSE\n");
  laylist = mpt->lay;
  nlay = mpt->nlay;
  xn = mpt->xn;
  yn = mpt->yn;
  tn = mpt->tn;
  seed = m->mseed[r->tsi];
  sprintf(ggstr,"    trial seed = %d\n",seed);
  mylog(mylogf,ggstr);
  if (seed > 0)
    seed = -seed;
  seed1 = (int)( 57913.0 * myrand_util_ran2(&seed)); 
  seed2 = (int)( 73119.0 * myrand_util_ran2(&seed)); 
  seedlist = get_seeds(seed1,99999,nlay);
  for(i=0;i<nlay;i++){
    lay = mpt->lay[i];
    lay->layseed_n = 10; 
    if (lay->layseed != NULL)
      myfree(lay->layseed);
    lay->layseed = get_seeds(seedlist[i],77777,lay->layseed_n);
  }
  myfree(seedlist);
  if (mpt->retflag != -1){  
    mylog(mylogf,"    Running subcortical module\n");
    for(i=0;i<nlay;i++){
      lay = mpt->lay[i];
      if (pop_util_lclass_is_lgn0(lay) ||
	  (strcmp(lay->laytype,"retina0_gc0")==0)){
	mod_pop_get_resp_lgn_lay(m,s,r,lay);
      }
    }
  }
  mod_pop_store_lgn(laylist,nlay,r);
  if (mpt->retflag != -1){ 
    mylog(mylogf,"    Pre-computing subcortical g's and bg spikes.\n");
    tnsec = (float)tn * mpt->tscale;
    sprintf(tname,"%s.thalamocort.pl",mpt->out_prefix);
    tname[0] = '\0'; 
    for(i=0;i<nlay;i++){  
      lay = mpt->lay[i];
      if (lay->runflag == 1){
	if (lay->lgn_in_flag == 1){
	  pop_util_set_gt_on_off_input_layer(mylogf,lay,m,xn,yn,tnsec,
					     lay->layseed[0],tname,mpt->binoc);
	}else{
	  pop_util_set_gad_layer(lay,xn,yn,tnsec);
	}
	pop_util_init_bg_spikes(lay,&(lay->layseed[1]),mpt->tn,mpt->tscale);
      }
    }
  }
  if ((myid == -1) && (mpt->monflag || (m->run_mode == 1))){
    mod_mon_reset(m,r,s);
    mod_mon_monitor_check(mpt->lay,mpt->nlay);
  }
  pop_util_run_layer_list(myid,mylogf,m,r,s,laylist,nlay,mpt->x1,mpt->x2,
			  seed2,mpt->retflag);
  if (myid == -1){
    sprintf(outfile,"%s.sav.pl",mpt->out_prefix);
    pop_cell_write_sav_layer_list(laylist,nlay,outfile,s->name,s->repno);
  }
  if (m->run_mode == 0){  
    pop_util_response_handover(m,r,laylist,nlay,mylogf);
  }
  if (m->run_mode == 1)
    mod_gui_pop_main(laylist,nlay,mpt,m,s,r);
}
void mod_pop_run(m,s,r,action)
     struct model_struct *m;     
     struct stim_struct *s;      
     struct response_struct *r;  
     int action;                 
{
  myid = mylog_set_id_log(m->process_id,&mylogf);  
  mylog(mylogf,"  MOD_POP_RUN\n");
  if (action == 0){
    if (onode_child_get_unique(m->o,"retina0") != NULL){
      model_mesh_rgc_01_prep(m,s,r);
    }else{
      model_dog_01_prep_o(m,r);
      model_dog_01_prep_ft();
    }
    mod_pop_prep(m,r,s); 
  }else if (action == 1){
    mod_pop_get_response(m,s,r);
  }else if (action == -1){
    if (onode_child_get_unique(m->o,"retina0") != NULL){
      model_mesh_rgc_01_done();
    }else{
      model_dog_01_done(m->ppl); 
    }
  }else if (action == 10){
    if (onode_child_get_unique(m->o,"retina0") != NULL){
      model_mesh_rgc_01_prep(m,s,r);
    }
    mod_pop_prep(m,r,s); 
  }else if (action == 20){
    mod_pop_prep(m,r,s); 
  }
}
