/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "my_util.h"
#include "nr_util.h"
#include "carray_util.h"
#include "iarray_util.h"
#include "farray_util.h"
#define	LENG	17
static long ary[LENG] = {
  0x4B14EA50L,
  0x53C4A8E0L,
  0x67B1FA98L,
  0x9CFB3BB5L,
  0x18761AF1L,
  0x7970CD66L,
  0xDBAFE136L,
  0x3C31FC3EL,
  0x697B37DEL,
  0x07BC568BL,
  0xCAFD3967L,
  0xA8F48722L,
  0x4AB26824L,
  0xA479EE47L,
  0x5C7246E2L,
  0x954BF297L,
  0x20A713ADL,
};
static int i1 = 0;
static int i2 = 12;
int yarlrand()
{
  if(++i1 >= LENG)
    i1 = 0;
  if(++i2 >= LENG)
    i2 = 0;
  ary[i1] += ary[i2];
  return ((ary[i1] >> 15) & 0x7FFF);
}
void yarlsrand(seed)
     int seed;
{
  register int i;
  for(i=0;i<LENG;i++){
    ary[i] = 0x55555555L;
    if(seed & 1)
      ary[i] = 0xCCCCCCCCL;
    if(seed & 0x8000)
      ary[i] ^= 0xF0F0F0F0L;
    seed >>= 1;
  }
  i1 = 0;
  i2 = 12;
  for(i = 1; i < 32; i += i)
    do {
      (void)yarlrand();
      ary[i1] += ary[i1] >> i;
      ary[i2] += ary[i2] << i;
    } while (i1);
}
#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)
/*C  (C) Copr. 1986-92 Numerical Recipes Software #.3.                       */
float myrand_util_ran2(idum)
     int *idum; 
{
  int j;
  int k; 
  static int idum2=123456789; 
  static int iy=0; 
  static int iv[NTAB]; 
  float temp;
  if (*idum <= 0) {
    if (-(*idum) < 1) *idum=1;
    else *idum = -(*idum);
    idum2=(*idum);
    for (j=NTAB+7;j>=0;j--) {
      k=(*idum)/IQ1;
      *idum=IA1*(*idum-k*IQ1)-k*IR1;
      if (*idum < 0) *idum += IM1;
      if (j < NTAB) iv[j] = *idum;
    }
    iy=iv[0];
  }
  k=(*idum)/IQ1;
  *idum=IA1*(*idum-k*IQ1)-k*IR1;
  if (*idum < 0) *idum += IM1;
  k=idum2/IQ2;
  idum2=IA2*(idum2-k*IQ2)-k*IR2;
  if (idum2 < 0) idum2 += IM2;
  j=iy/NDIV;
  iy=iv[j]-idum2;
  iv[j] = *idum;
  if (iy < 1) iy += IMM1;
  if ((temp=AM*iy) > RNMX) return RNMX;
  else return temp;
}
#undef IM1
#undef IM2
#undef AM
#undef IMM1
#undef IA1
#undef IA2
#undef IQ1
#undef IQ2
#undef IR1
#undef IR2
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX
int myrand_get_poisson_count(mu,rseed)
     float mu;
     int *rseed;
{  
  int n;
  float ftime;
  if (mu == 0.0){
    return 0;
  }else if (mu < 0.0)
    exit_error("MYRAND_GET_POISSON_COUNT","Mu is negative");
  ftime = 0.0;
  n = 0;
  while (ftime < mu){
    ftime += -log(myrand_util_ran2(rseed)); 
    n += 1;
  }
  return n-1;
}
float myrand_get_poisson_count_mean(float mu,      
				    int n,         
				    int *rseed){   
  int i;
  float rv,ftime;
  rv = 0.0;
  if ((n > 0) && (mu > 0)){
    for(i=0;i<n;i++){
      rv += (float)myrand_get_poisson_count(mu,rseed);
    }
    rv /= (float)n;
  }
  return rv;
}
int myrand_poisson_count(mu,seed)
     float mu;
     int seed;
{  
  int n;
  float ftime;
  if (seed > 0)
    seed = -seed;
  if (mu == 0.0){
    return 0;
  }else if (mu < 0.0)
    exit_error("MYRAND_POISSON_COUNT","Mu is negative");
  ftime = 0.0;
  n = 0;
  while (ftime < mu){
    ftime += -log(myrand_util_ran2(&seed)); 
    n += 1;
  }
  return n-1;
}
/*C  (C) Copr. 1986-92 Numerical Recipes Software #.3.                       */
void hpsort(n,ra,rx)
     unsigned long n;
     float ra[];
     int *rx; 
{
  unsigned long i,ir,j,l;
  float rra;
  int rrx; 
  if (n < 2) return;
  l=(n >> 1)+1;
  ir=n;
  for (;;) {
    if (l > 1) {
      rra=ra[--l];
      rrx=rx[l]; 
    } else {
      rra=ra[ir];
      rrx=rx[ir]; 
      ra[ir]=ra[1];
      rx[ir]=rx[1]; 
      if (--ir == 1) {
	ra[1]=rra;
	rx[1]=rrx; 
	break;
      }
    }
    i=l;
    j=l+l;
    while (j <= ir) {
      if (j < ir && ra[j] < ra[j+1]) j++;
      if (rra < ra[j]) {
	ra[i]=ra[j];
	rx[i]=rx[j]; 
	i=j;
	j <<= 1;
      } else j=ir+1;
    }
    ra[i]=rra;
    rx[i]=rrx; 
  }
}
void shuffle_iarray(data,n,nshuffle,seed)
     int *data,n,nshuffle;
     int seed;
{
  int i,j;
  float *r; 
  if (seed > 0) 
    seed = -seed;
  r = (float *)myalloc(n*sizeof(float));
  for (i=0;i<nshuffle;i++){
    for (j=0;j<n;j++)
      r[j] = myrand_util_ran2(&seed);
    hpsort((unsigned long)n,r-1,data-1);
  }
  myfree(r);
}
void shuffle_iarray_nyu(data,n,nshuffle,seed)
     int *data,n,nshuffle;
     int seed;
{
  int i,j;
  float *r; 
  if (seed > 0) 
    seed = -seed;
  myrand_util_ran2(&seed); 
  r = (float *)myalloc(n*sizeof(float));
  for (i=0;i<nshuffle;i++){
    for (j=0;j<n;j++)
      r[j] = myrand_util_ran2(&seed);
    hpsort((unsigned long)n,r-1,data-1);
  }
  myfree(r);
}
void shuffle_iarray_return_seed(data,n,nshuffle,seed)
     int *data,n,nshuffle;
     int *seed;
{
  int i,j;
  float *r; 
  if (*seed > 0) 
    *seed = -*seed;
  r = (float *)myalloc(n*sizeof(float));
  for (i=0;i<nshuffle;i++){
    for (j=0;j<n;j++)
      r[j] = myrand_util_ran2(seed);
    hpsort((unsigned long)n,r-1,data-1);
  }
  myfree(r);
}
int *get_shuffle_index(n,nshuffle,seed)
     int n,nshuffle;
     int seed;
{
  int i;
  int *s;
  s = (int *)myalloc(n*sizeof(int));
  for (i=0;i<n;i++)
    s[i] = i;
  shuffle_iarray(s,n,nshuffle,seed);
  return s;
}
int *get_shuffle_index_return_seed(n,nshuffle,rseed)
     int n,nshuffle;
     int *rseed;
{
  int i;
  int *s;
  s = (int *)myalloc(n*sizeof(int));
  for (i=0;i<n;i++)
    s[i] = i;
  shuffle_iarray_return_seed(s,n,nshuffle,rseed);
  return s;
}
int *get_seeds(seed,scale,n)
     int seed,scale,n;
{
  int i;
  int *s;
  if (seed == 0)
    exit_error("(myrand_util) GET_SEEDS","seed = 0");
  if (seed > 0) 
    seed = -seed;
  s = (int *)myalloc(n*sizeof(int));
  for (i=0;i<n;i++){
    s[i] = (int)(myrand_util_ran2(&seed) * (double)scale);  
    while(s[i] == 0)
      s[i] = (int)(myrand_util_ran2(&seed) * (double)scale);
  }
  return s;
}
int **get_seeds_2d(seed,scale,xn,yn)
     int seed,scale;
     int xn,yn;
{
  int i,j;
  int **s;
  if (seed == 0)
    exit_error("(myrand_util) GET_SEEDS_2D","seed = 0");
  if (seed > 0) 
    seed = -seed;
  s = get_2d_iarray(xn,yn);
  for (i=0;i<xn;i++){
    for (j=0;j<yn;j++){
      s[i][j] = (int)(myrand_util_ran2(&seed) * (double)scale);
      while(s[i][j] == 0)
	s[i][j] = (int)(myrand_util_ran2(&seed) * (double)scale);
    }
  }
  return s;
}
float *get_random_floats(seed,n)
     int seed,n;
{
  int i;
  float *d;
  if (seed > 0) 
    seed = -seed;
  d = (float *)myalloc(n*sizeof(float));
  for (i=0;i<n;i++)
    d[i] = myrand_util_ran2(&seed);
  return d;
}
int *myrand_get_std_unif_int_seq(n,seed,m)
     int n,seed;
     int m;  
{
  int i;
  int *s;
  float fm;
  fm = (float)m;
  if (seed == 0)
    exit_error("MYRAND_GET_STD_UNIF_INT_SEQ","seed = 0");
  else if (seed > 0)
    seed *= -1;
  s = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++)
    s[i] = (int)(fm * myrand_util_ran2(&seed));
  return s;
}
float *myrand_get_std_unif_float_seq(n,seed,m)
     int n,seed;
     float m;  
{
  int i;
  float *s;
  if (seed == 0)
    exit_error("MYRAND_GET_STD_UNIF_FLOAT_SEQ","seed = 0");
  else if (seed > 0)
    seed *= -1;
  s = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    s[i] = m * myrand_util_ran2(&seed);
  return s;
}
int *myrand_get_std_bin_seq(n,seed)
     int n,seed;
{
  int i;
  int *s;
  if (seed == 0)
    exit_error("MYRAND_GET_STD_BIN_SEQ","seed = 0");
  else if (seed > 0)
    seed *= -1;
  s = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++)
    if (myrand_util_ran2(&seed) > 0.5)
      s[i] = 1;
    else
      s[i] = 0;
  return s;
}
int *myrand_get_std_tern_seq(n,seed)
     int n,seed;
{
  int i;
  int *s;
  float rv;
  if (seed == 0)
    exit_error("MYRAND_GET_STD_TERN_SEQ","seed = 0");
  else if (seed > 0)
    seed *= -1;
  s = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++){
    rv = myrand_util_ran2(&seed);
    if (rv < 0.33333)
      s[i] = -1;
    else if (rv < 0.66667)
      s[i] = 0;
    else
      s[i] = 1;
  }
  return s;
}
int *myrand_get_std_quad_seq(n,seed)
     int n,seed;
{
  int i;
  int *s;
  float rv;
  if (seed == 0)
    exit_error("MYRAND_GET_STD_QUAD_SEQ","seed = 0");
  else if (seed > 0)
    seed *= -1;
  s = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++){
    rv = myrand_util_ran2(&seed);
    if (rv < 0.25)
      s[i] = 0;
    else if (rv < 0.5)
      s[i] = 1;
    else if (rv < 0.75)
      s[i] = 2;
    else
      s[i] = 3;
  }
  return s;
}
int *myrand_get_std_mseq(n,tap_seed,ord)
     int n;
     int tap_seed;  
     int ord;       
{
  int i;
  int *s,nm;
  int ms_r,ms_b,ms_q,ms_t;
  nm = 1;
  for(i=0;i<ord;i++)
    nm *= 2;
  if (n > nm){
    printf("  *** m-seq order %d had %d values\n",ord,nm);
    printf("  *** %d values were requested\n",n);
    exit_error("MYRAND_GET_STD_MSEQ","requested too many values for order");
  }
  ms_r = 1;
  ms_q = 1 << (ord-1); 
  ms_t = tap_seed;
  s = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++){
    ms_b = ms_r & 1;
    if ((ms_r & ms_q) != ms_q) 
      ms_r = ms_r << 1;
    else
      ms_r = (ms_r << 1) ^ ms_t;
    if (ms_b == 0)
      s[i] = 0;
    else
      s[i] = 1;
  }
  return s;
}
float *myrand_get_std_gauss_seq(n,seed)
     int n,seed;
{
  int i;
  float *s;
  if (seed == 0)
    exit_error("MYRAND_GET_STD_GAUSS_SEQ","seed = 0");
  else if (seed > 0)
    seed *= -1;
  s = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    s[i] = nr_util_gasdev(&seed);   
  return s;
}
float **myrand_get_std_bin_float_2d(xn,yn,seed)
     int xn,yn,seed;
{
  int i,j,k;
  int *seqi;
  float **f2d;
  seqi = myrand_get_std_bin_seq(xn*yn,seed);
  f2d = get_2d_farray(xn,yn);
  k = 0;
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      if (seqi[k] == 0)
	f2d[i][j] = -1.0;
      else
	f2d[i][j] = 1.0;
      k += 1;
    }
  }
  myfree(seqi);
  return f2d;
}
float **myrand_get_std_quad_float_2d(xn,yn,seed)
     int xn,yn,seed;
{
  int i,j,k;
  int *seqi;
  float **f2d;
  seqi = myrand_get_std_quad_seq(xn*yn,seed);
  f2d = get_2d_farray(xn,yn);
  k = 0;
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      if (seqi[k] == 0)
	f2d[i][j] = -1.0;
      else if (seqi[k] == 1)
	f2d[i][j] = -1.0/3.0;
      else if (seqi[k] == 2)
	f2d[i][j] = 1.0/3.0;
      else
	f2d[i][j] = 1.0;
      k += 1;
    }
  }
  myfree(seqi);
  return f2d;
}
char **get_deviate_list_from_description(desc,n)
     char *desc;
     int n;
{
  int i,k;
  int ns,seed;
  float min,range;
  char **slist,**ranlist,tstr[SLEN];
  ranlist = (char **)myalloc(n*sizeof(char *));
  get_items_from_string(desc,&slist,&ns);
  if (strcmp(slist[0],"uniform_int")==0){
    seed = atoi(slist[1]);
    min = atof(slist[2]);
    range = atof(slist[3]) - min;
    if (seed > 0)
      seed *= -1;
    for(i=0;i<n;i++){
      sprintf(tstr,"%d",my_rint(min + range*myrand_util_ran2(&seed)));
      ranlist[i] = strdup(tstr);
    }
  }else if (strcmp(slist[0],"uniform_float")==0){
    seed = atoi(slist[1]);
    min = atof(slist[2]);
    range = atof(slist[3]) - min;
    if (seed > 0)
      seed *= -1;
    for(i=0;i<n;i++){
      sprintf(tstr,"%.6f",min + range*myrand_util_ran2(&seed));
      ranlist[i] = strdup(tstr);
    }
  }else if (strcmp(slist[0],"uniform_list")==0){
    seed = atoi(slist[1]);
    min = 0.0;
    range = (float)(ns-2); 
    if (seed > 0)
      seed *= -1;
    for(i=0;i<n;i++){
      k = my_rint(min + range*myrand_util_ran2(&seed));
      ranlist[i] = strdup(slist[2+k]);
    }
  }else
    exit_error("GET_DEVIATE_LIST_FROM_DESCRIPTION","Unknown distrib. type");
  free_2d_carray(slist,ns);
  return ranlist;
}
int StimRandShort(int *seed)
{
  int temp1,temp2,temp3,result;
  temp1 = (*seed & 0xFFFF) * 0x41A7;
  temp2 = (*seed >> 16) * 0x41A7 + (temp1 >> 16);
  temp3 = temp2 * 2 >> 16;
  temp1 = (temp1 & 0xFFFF) - 0x7FFFFFFF;
  temp2 &= 0x7FFF;
  temp1 += (temp2 << 16) | (temp2 >> 16) + temp3;
  if (temp1 < 0)
    temp1 += 0x7FFFFFFF;
  *seed = temp1;
  result = temp1 & 0xFFFF;  
  if (result == 0x8000)     
    result = 0;
  return result;
}
int ejrand_getnums(int *op,    
		   int iter,   
		   int seed)   
{
  int i;
  for (i=0;i<iter;i++)
    op[i] = StimRandShort(&seed);
  return seed;
}
float **myrand_get_std_quad_ejrand_2d(xn,yn,seed)
     int xn,yn,seed;
{
  int i,j,jj,k;
  int *seqi,tseed;
  float **f2d;
  seqi = (int *)myalloc(xn*yn*sizeof(int));  
  tseed = ejrand_getnums(seqi,xn*yn,seed);  
  f2d = get_2d_farray(xn,yn);
  k = 0;
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      jj = yn-1 - j;   
      if (seqi[k] % 4 == 0)
	f2d[i][jj] = -1.0;
      else if (seqi[k] % 4 == 1)
	f2d[i][jj] = -1.0/3.0;
      else if (seqi[k] % 4 == 2)
	f2d[i][jj] = 1.0/3.0;
      else
	f2d[i][jj] = 1.0;
      k += 1;
    }
  }
  myfree(seqi);
  return f2d;
}
void myrand_ejrand_test(n,seed)
     int n;      
     int seed;   
{
  int i,k;
  int maxval,mod4[4];
  float *hist;
  printf("  MYRAND_EJRAND_TEST\n");
  printf("    Histogram plot of %d random values for seed %d\n",n,seed);
  maxval = 65536;
  hist = get_zero_farray(maxval);
  mod4[0] = mod4[1] = mod4[2] = mod4[3] = 0;
  for(i=0;i<n;i++){
    k = StimRandShort(&seed);
    if ((k < 0) || (k >= maxval)){
      exit_error("MYRAND_EJRAND_TEST","Value out of bounds");
    }
    hist[k] += 1.0;
    mod4[k%4] += 1;
  }
  printf("   0:  %d\n",mod4[0]);
  printf("   1:  %d\n",mod4[1]);
  printf("   2:  %d\n",mod4[2]);
  printf("   3:  %d\n",mod4[3]);
  append_farray_plot("zzz.ejrand.hist.pl","hist",hist,maxval,1);
  printf("   Done.  Exiting.\n\n");
  exit(0);
}
int *ejrand_get_seeds(int seed,  
		      int vpf,   
		      int n)     
{
  int i;
  int *slist,*t;
  slist = (int *)myalloc(n*sizeof(int));    
  t = (int *)myalloc(vpf*sizeof(int));      
  for (i=0;i<n;i++){
    slist[i] = seed;
    seed = ejrand_getnums(t,vpf,slist[i]);
  }
  myfree(t);
  return slist;
}
