/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "my_util.h"
#include "farray_util.h"
#include "fft_util.h"
#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)
/*C  (C) Copr. 1986-92 Numerical Recipes Software #.3.                       */
float sig_util_ran2(idum)
     int *idum;
{
  int j;
  int k;
  static int idum2=123456789;
  static int iy=0;
  static int iv[NTAB];
  float temp;
  if (*idum <= 0) {
    if (-(*idum) < 1) *idum=1;
    else *idum = -(*idum);
    idum2=(*idum);
    for (j=NTAB+7;j>=0;j--) {
      k=(*idum)/IQ1;
      *idum=IA1*(*idum-k*IQ1)-k*IR1;
      if (*idum < 0) *idum += IM1;
      if (j < NTAB) iv[j] = *idum;
    }
    iy=iv[0];
  }
  k=(*idum)/IQ1;
  *idum=IA1*(*idum-k*IQ1)-k*IR1;
  if (*idum < 0) *idum += IM1;
  k=idum2/IQ2;
  idum2=IA2*(idum2-k*IQ2)-k*IR2;
  if (idum2 < 0) idum2 += IM2;
  j=iy/NDIV;
  iy=iv[j]-idum2;
  iv[j] = *idum;
  if (iy < 1) iy += IMM1;
  if ((temp=AM*iy) > RNMX) return RNMX;
  else return temp;
}
#undef IM1
#undef IM2
#undef AM
#undef IMM1
#undef IA1
#undef IA2
#undef IQ1
#undef IQ2
#undef IR1
#undef IR2
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX
/*C  (C) Copr. 1986-92 Numerical Recipes Software #.3.                       */
float sig_gasdev(idum)
     int *idum;
{
  static int iset=0;
  static float gset;
  float fac,rsq,v1,v2;
  if ((iset == 0) || (*idum <= 0)){  
    do {
      v1=2.0*sig_util_ran2(idum)-1.0;
      v2=2.0*sig_util_ran2(idum)-1.0;
      rsq=v1*v1+v2*v2;
    } while (rsq >= 1.0 || rsq == 0.0);
    fac=sqrt(-2.0*log(rsq)/rsq);
    gset=v1*fac;
    iset=1;
    return v2*fac;
  }else{
    iset=0;
    return gset;
  }
}
float sig_expdev(lambda,idum)
     float lambda;
     int *idum;
{
  if (lambda == 0.0)
    return 0.0;
  else
    return(-log(sig_util_ran2(idum))/lambda);
}
float sig_gaussian_float(mu,sigma,idum)
     float mu,sigma;
     int *idum;
{
  return(sig_gasdev(idum)*sigma + mu);
}
void mean_autocorr_noise(data,m,n,acmean,acsdev)
     float **data;
     int m,n;  
     float **acmean,**acsdev;
{
  int i,j;
  float **ac,*mean,*sdev,*noise;
  printf("  MEAN_AUTOCORR_NOISE\n");
  mean_sdev_2d_farray(data,m,n,&mean,&sdev); 
  ac = (float **)myalloc(m*sizeof(float *));
  noise = (float *)myalloc(n*sizeof(float));
  for(i=0;i<m;i++){  
    printf("%d ",i); fflush(stdout);
    for(j=0;j<n;j++)   
      noise[j] = data[i][j] - mean[j];
    z_score_farray(noise,n);
    ac[i] = autocorr_farray(noise,n);
  }
  mean_sdev_2d_farray(ac,m,n,acmean,acsdev); 
  for(i=0;i<m;i++)
    myfree(ac[i]);
  myfree(ac);
  myfree(mean); myfree(sdev); myfree(noise);
}
void crosscorr_mean(data1,data2,m,n,ccorr)
     float **data1,**data2;
     int m,n;  
     float **ccorr;
{
  int ncc;
  float *cc,*mean1,*mean2,*sdev1,*sdev2;
  printf("  CROSSCORR_MEAN\n");
  mean_sdev_2d_farray(data1,m,n,&mean1,&sdev1); 
  mean_sdev_2d_farray(data2,m,n,&mean2,&sdev2);
  z_score_farray(mean1,n);
  z_score_farray(mean2,n);
  cc = correlate(mean1,mean2,n,n,&ncc); 
  *ccorr = cc;
  myfree(mean1); myfree(mean2); myfree(sdev1); myfree(sdev2);
}
void mean_crosscorr_noise(data1,data2,m,n,ccmean,ccsdev)
     float **data1,**data2;
     int m,n;  
     float **ccmean,**ccsdev;
{
  int i,j,si;
  int ncc,shift_flag;
  float **cc,*mean1,*mean2,*sdev1,*sdev2,*noise1,*noise2;
  float mm1,mm2,ss;
  printf("  MEAN_CROSSCORR_NOISE\n");
  shift_flag = 1;
  mean_sdev_2d_farray(data1,m,n,&mean1,&sdev1); 
  mean_sdev_2d_farray(data2,m,n,&mean2,&sdev2);
  mean_sdev_farray(mean1,n,&mm1,&ss);
  mean_sdev_farray(mean2,n,&mm2,&ss);
  cc = (float **)myalloc(m*sizeof(float *));
  noise1 = (float *)myalloc(n*sizeof(float));
  noise2 = (float *)myalloc(n*sizeof(float));
  for(i=0;i<m;i++){  
    si = (i+m/3) % m; 
    if (shift_flag)
      printf("%d(%d) ",i,si);
    else
      printf("%d ",i);
    fflush(stdout);
    for(j=0;j<n;j++){   
      noise1[j] = data1[i][j] - mean1[j];
      if (shift_flag)
	noise2[j] = data2[si][j] - mean2[j];
      else
	noise2[j] = data2[i][j] - mean2[j];
    }
    cc[i] = correlate(noise1,noise2,n,n,&ncc); 
  }
  printf("\n");
  mean_sdev_2d_farray(cc,m,ncc,ccmean,ccsdev); 
  for(i=0;i<m;i++)
    myfree(cc[i]);
  myfree(cc);
  myfree(mean1); myfree(mean2); myfree(sdev1); myfree(sdev2);
  myfree(noise1); myfree(noise2);
}
float *gaussian_white_noise(mu,sigma,n,idum)
     float mu,sigma;
     int n;
     int idum;
{
  int i;
  float *noise;
  if (idum > 0)
    idum = -idum;
  noise = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    noise[i] = sig_gaussian_float(mu,sigma,&idum);
  return noise;
}
float *gaussian_corr_noise(corr_sigma,mu,sigma,n,seed)
     float corr_sigma,mu,sigma;
     int n;
     int seed;
{
  int i;
  float *gwn,*data,mean,sdev,cf,x;
  if (sigma > 0.0){
    gwn = gaussian_white_noise(0.0,1.0,n,seed);
    data = smooth_with_gaussian(gwn,n,corr_sigma,0.01);
    myfree(gwn);
    mean_sdev_farray(data,n,&mean,&sdev);
    cf = sigma/sdev; 
    for(i=0;i<n;i++){
      x = data[i] - mean;
      data[i] = mu + x*cf;
    }
  }else
    data = get_const_farray(n,mu);
  return data;
}
float *leaky_integrate(data,n,tau,tolerance,nm)
     float *data;
     int n;
     float tau,tolerance;
     int *nm; 
{
  int i;
  int nmask,nfull;
  float *mask,*result,mval;
  nmask = 0;
  mval = exp(-((float)nmask/tau));
  while(mval > tolerance){
    nmask += 1;
    mval = exp(-((float)nmask/tau));
  }
  nmask += 2;
  nfull = 2*nmask - 1;
  mask = get_zero_farray(nfull);
  for(i=0;i<nmask;i++)
    mask[nmask-1+i] = exp(-((float)i/tau));
  result = convolve_with_mask(data,n,mask,nfull);
  *nm = nmask;
  return result;
}
