/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h> 
#include <unistd.h>
#include "my_util.h"
#include "nr_util.h"
#include "carray_util.h"
#include "iarray_util.h"
#include "farray_util.h"
#include "func_util.h"
#include "myrand_util.h"
#include "spike_util.h"
#include "ndata_util.h"
#include "paramfile_util.h"
#include "fft_util.h"
#include "stim_util.h"
#include "stm_util.h"
#include "mod_util.h"
#include "pop_util.h"
#include "ifc_util.h"
#include "mod_dog_util.h"
#include "mod_lin_util.h"
#include "mod_me_util.h"
#include "mod_wav_util.h"
#include "mod_mesh_util.h"
#include "mod_pop_util.h"
#include "mod_vhf_util.h"
#include "mod_dcn_util.h"
#include "mod_test_util.h"
#include "mod_x_util.h"
#include "mod_srf_util.h"
#include "mod.h"
#include "ndata.h"
#include "paramfile.h"
void wm_moo_var_prep(m,s,mylogf)
     struct model_struct *m; 
     struct stim_struct *s;  
     char *mylogf;
{
  struct onode *to,*vpo;
  char *infile;
  mylog(mylogf,"    WM_MOO_VAR_PREP\n");
  vpo = onode_child_get_unique(m->o,"var_param");
  if (vpo == NULL){
    if (onode_item(m->o,"var_param_file")==1){
      mylog(mylogf,"      Reading moo var params from file\n");
      infile = onode_getpar_chr_exit(m->o,"var_param_file");
      to = paramfile_onode_file_read(infile,1);   
      myfree(infile);
      vpo = onode_child_get_unique(to,"var_param");
      if (vpo == NULL)
	exit_error("WM_MOO_VAR_PREP","<var_param> not found in file.");
      onode_insert_child_at_end(m->o,vpo);
    }else{
      return; 
    }
  }else{
    mylog(mylogf,"      Found 'var_param' child\n");
  }
  m->vps = mod_util_var_par(mylogf,m,vpo);
  m->nrun = m->vps->m_n;  
}
void wm_prep_var_const_gen(m,s,mylogf,arglist,narg,parflag)
     struct model_struct *m; 
     struct stim_struct *s;  
     char *mylogf;
     char **arglist;  
     int narg;
     int *parflag;    
{
  int i,k;
  int n,seed,nmoo,flag_moo_tn,flag_moo_tscale,flag_moo_sscale;
  char ggstr[LONG_SLEN];
  char **moo_name,**moo_val;
  if (m != NULL)  
    if (m->ppl == NULL)  
      wm_moo_var_prep(m,s,mylogf);
  prep_var_params(mylogf,s->ppl,1,&(s->vval),&(s->vname),&(s->vtype),
		  &(s->nvar),&(s->ntr),&(s->val),&(s->vcnt),&(s->nvl),
		  &(s->nvp));
  if (m == NULL)
    nmoo = 0;
  else{
    flag_moo_tn = 0;      
    flag_moo_tscale = 0;  
    flag_moo_sscale = 0;  
    nmoo = 3;             
    if ((narg > 0) && (parflag != NULL))
      nmoo += sum_iarray(parflag,narg,0,narg);
    moo_name = (char **)myalloc(nmoo*sizeof(char *));
    moo_val  = (char **)myalloc(nmoo*sizeof(char *));
    k = 0;
    for(i=0;i<narg;i++){  
      if (parflag[i] == 1){
	moo_name[k] = strdup(arglist[i]);
	moo_val[k]  = strdup(arglist[i+1]);
	if (strcmp(moo_name[k],"tn")==0)
	  flag_moo_tn = 1;
	else if (strcmp(moo_name[k],"tscale")==0)
	  flag_moo_tscale = 1;
	else if (strcmp(moo_name[k],"sscale")==0)
	  flag_moo_sscale = 1;
	k += 1;
      }
    }
    if (flag_moo_tn == 0){
      moo_name[k] = strdup("tn");
      if (m->ppl != NULL)
	moo_val[k] = param_getc_exit(m->ppl,"tn");
      else{
	moo_val[k] = onode_getpar_chr_exit(m->o,"tn");
      }
      k += 1;
    }
    if (flag_moo_tscale == 0){
      moo_name[k] = strdup("tscale");
      if (m->ppl != NULL)
	moo_val[k] = param_getc_exit(m->ppl,"tscale");
      else{
	moo_val[k] = onode_getpar_chr_exit(m->o,"tscale");
      }
      k += 1;
    }
    if (flag_moo_sscale == 0){
      moo_name[k] = strdup("sscale");
      if (m->ppl != NULL)
	moo_val[k] = param_getc_exit(m->ppl,"sscale");
      else{
	moo_val[k] = onode_getpar_chr_exit(m->o,"sscale");
      }
      k += 1;
    }
    nmoo = k;  
  }
  prep_const_params(mylogf,s,moo_name,moo_val,nmoo,&(s->cval),&(s->cname),
		    &(s->ctype),&(s->ncon));
  mylog(mylogf,"    Const params\n");
  for(i=0;i<s->ncon;i++){
    sprintf(ggstr,"    %2d %12s %c %s\n",i,s->cname[i],s->ctype[i],s->cval[i]);
    mylog(mylogf,ggstr);
  }
  if (s->nvar > 0){
    mylog(mylogf,"    Var params\n");
    for(i=0;i<s->nvar;i++){
      sprintf(ggstr,"    %2d %12s %c\n",i,s->vname[i],s->vtype[i]);
      mylog(mylogf,ggstr);
    }
  }
  n = s->ntr;
  if (m == NULL){
    mylog(mylogf,"  WM_PREP_VAR_CONST_GEN - No model, model seeds not set.\n");
  }else{
    if (m->ppl != NULL)
      seed = paramfile_get_int_param_default(m->ppl,"model_noise_seed",1777);
    else{
      seed = onode_getpar_int_dflt(m->o,"model_noise_seed",1777);
    }
    m->nseed = n;
    m->mseed = get_seeds(seed,100000,n);
  }
}
void wm_response_count(m,r,mylogf)
     struct model_struct *m;
     struct response_struct *r;
     char *mylogf;
{
  int nx,ny,nz;
  char *pop_name;
  struct param_pair *pp;
  pp = paramfile_get_first_prefix_pointer(r->ppl,"resp_save_");
  if (pp != NULL)
    mylog_exit(mylogf,"WM_RESPONSE_COUNT CHANGE resp_save_ to save_as_  !!\n");
  r->n = 0;
  r->n += paramfile_count_prefix_param(r->ppl,"save_as");
  r->n += paramfile_count_prefix_param(r->ppl,"save_pop_unit_as");
  r->n += paramfile_count_prefix_param(r->ppl,"save_pop_syn_as");
  pp = paramfile_get_first_prefix_pointer(r->ppl,"save_pop_grid_as_");
  while(pp != NULL){
    nx = paramfile_get_nth_int_param_ptr_or_exit(pp,6);
    ny = paramfile_get_nth_int_param_ptr_or_exit(pp,7);
    nz = paramfile_get_nth_int_param_ptr_or_exit(pp,8);
    r->n += (nx * ny * nz);
    if ((nx == 0) || (ny == 0) || (nz == 0)){
      printf("  *** In 'save_pop_grid_as_... %d %d %d ...'",nx,ny,nz);
      printf("   none of these lengths should be < 1\n");
      exit_error("WM_RESPONSE_COUNT","Instruction to record 0 data");
    }
    pp = paramfile_get_next_prefix_pointer(pp,"save_pop_grid_as_");
  }
  pp = paramfile_get_first_prefix_pointer(r->ppl,"save_pop_layer_as_");
  while(pp != NULL){
    pop_name = paramfile_get_nth_char_param_pp_or_exit(pp,2);
    pop_util_onode_get_pop_xn_yn_zn(m->o,pop_name,&nx,&ny,&nz);
    if (nx == -1){ 
      mylog(mylogf,"    Using model xn,yn for layer dimensions");
      nx = onode_getpar_int_exit(m->o,"xn");
      ny = onode_getpar_int_exit(m->o,"yn");
      nz = 1;
    }
    r->n += (nx * ny * nz);
    pp = paramfile_get_next_prefix_pointer(pp,"save_pop_layer_as_");
  }
}
void wm_response_config(m,r,mylogf)
     struct model_struct *m;
     struct response_struct *r;
     char *mylogf;
{
  int i,j,k,l;
  int nval,gx0,gy0,gz0,gxn,gyn,gzn;
  float gsamp;
  char *savtype,*gname,*gnd_name,tstr[SLEN],*gdid,*tdup;
  struct param_pair *pp;
  r->rformat =   (int *)myalloc(r->n*sizeof(int));
  r->datid   = (char **)myalloc(r->n*sizeof(char *));
  r->plname  = (char **)myalloc(r->n*sizeof(char *));
  r->xi      =   (int *)myalloc(r->n*sizeof(int));
  r->yi      =   (int *)myalloc(r->n*sizeof(int));
  r->zi      =   (int *)myalloc(r->n*sizeof(int));
  r->plname1 = (char **)myalloc(r->n*sizeof(char *));
  r->xi1     =   (int *)myalloc(r->n*sizeof(int));
  r->yi1     =   (int *)myalloc(r->n*sizeof(int));
  r->zi1     =   (int *)myalloc(r->n*sizeof(int));
  r->nd_name = (char **)myalloc(r->n*sizeof(char *));
  r->rtype   = (char **)myalloc(r->n*sizeof(char *));
  r->samp    = (float *)myalloc(r->n*sizeof(float));
  r->ri      =   (int *)myalloc(r->n*sizeof(int));
  k = 0;
  r->ns = r->nf = 0;
  pp = paramfile_get_first_prefix_pointer(r->ppl,"save_as_");
  while(pp != NULL){
    nval = paramfile_count_values_pointer(pp);
    if (nval != 3)
      exit_error("WM_RESPONSE_CONFIG","save_as_ bad value list");
    r->rformat[k] = 0; 
    r->nd_name[k] = strdup(pp->name+8);
    savtype = paramfile_get_nth_char_param_pp_or_exit(pp,0);
    r->samp[k] = paramfile_get_nth_float_param_ptr_or_exit(pp,1);
    r->plname[k] = strdup("NULL");
    r->datid[k] = paramfile_get_nth_char_param_pp_or_exit(pp,2);
    r->xi[k] = r->yi[k] = r->zi[k] = 0; 
    if (strcmp(savtype,"s")==0){
      r->rtype[k] = strdup("s");
      r->ri[k] = r->ns;
      r->ns += 1;
    }else if (strcmp(savtype,"f")==0){
      r->rtype[k] = strdup("f");
      r->ri[k] = r->nf;
      r->nf += 1;
    }else{
      printf("*** Error in .rsp file:\n");
      printf("  *** %s  %s\n",pp->name,pp->value);
      printf("  *** 'savtype' = %s\n",savtype);
      exit_error("WM_RESPONSE_CONFIG","save_as_ bad type");
    }
    myfree(savtype);
    k += 1;
    pp = paramfile_get_next_prefix_pointer(pp,"save_as_");
  }
  pp = paramfile_get_first_prefix_pointer(r->ppl,"save_pop_unit_as_");
  while(pp != NULL){
    nval = paramfile_count_values_pointer(pp);
    if (nval != 7)
      exit_error("WM_RESPONSE_CONFIG","save_pop_unit_as bad value list");
    r->rformat[k] = 1; 
    r->nd_name[k] = strdup(pp->name+17);
    savtype = paramfile_get_nth_char_param_pp_or_exit(pp,0);
    r->samp[k] = paramfile_get_nth_float_param_ptr_or_exit(pp,1);
    r->plname[k] = paramfile_get_nth_char_param_pp_or_exit(pp,2);
    r->xi[k]   = paramfile_get_nth_int_param_ptr_or_exit(pp,3);
    r->yi[k]   = paramfile_get_nth_int_param_ptr_or_exit(pp,4);
    r->zi[k]   = paramfile_get_nth_int_param_ptr_or_exit(pp,5);
    r->datid[k] = paramfile_get_nth_char_param_pp_or_exit(pp,6);
    if (strcmp(savtype,"s")==0){
      r->rtype[k] = strdup("s");
      r->ri[k] = r->ns;
      r->ns += 1;
    }else if (strcmp(savtype,"f")==0){
      r->rtype[k] = strdup("f");
      r->ri[k] = r->nf;
      r->nf += 1;
    }else{
      printf("*** Error in .rsp file:\n");
      printf("  *** %s  %s\n",pp->name,pp->value);
      printf("  *** 'savtype' = %s\n",savtype);
      exit_error("WM_RESPONSE_CONFIG","save_pop_unit_as_ bad type");
    }
    myfree(savtype);
    k += 1;
    pp = paramfile_get_next_prefix_pointer(pp,"save_pop_unit_as_");
  }
  pp = paramfile_get_first_prefix_pointer(r->ppl,"save_pop_syn_as_");
  while(pp != NULL){
    nval = paramfile_count_values_pointer(pp);
    if (nval != 11)
      exit_error("WM_RESPONSE_CONFIG","save_pop_syn_as bad value list");
    r->rformat[k] = 10; 
    r->nd_name[k] = strdup(pp->name+16);
    savtype      = paramfile_get_nth_char_param_pp_or_exit(pp,0);
    r->samp[k]   = paramfile_get_nth_float_param_ptr_or_exit(pp,1);
    r->plname[k] = paramfile_get_nth_char_param_pp_or_exit(pp,2);
    r->xi[k]     = paramfile_get_nth_int_param_ptr_or_exit(pp,3);
    r->yi[k]     = paramfile_get_nth_int_param_ptr_or_exit(pp,4);
    r->zi[k]     = paramfile_get_nth_int_param_ptr_or_exit(pp,5);
    r->plname1[k] = paramfile_get_nth_char_param_pp_or_exit(pp,6);
    r->xi1[k]     = paramfile_get_nth_int_param_ptr_or_exit(pp,7);
    r->yi1[k]     = paramfile_get_nth_int_param_ptr_or_exit(pp,8);
    r->zi1[k]     = paramfile_get_nth_int_param_ptr_or_exit(pp,9);
    r->datid[k]  = paramfile_get_nth_char_param_pp_or_exit(pp,10);
    if (strcmp(savtype,"s")==0){
      r->rtype[k] = strdup("s");
      r->ri[k] = r->ns;
      r->ns += 1;
    }else if (strcmp(savtype,"f")==0){
      r->rtype[k] = strdup("f");
      r->ri[k] = r->nf;
      r->nf += 1;
    }else{
      printf("*** Error in .rsp file:\n");
      printf("  *** %s  %s\n",pp->name,pp->value);
      printf("  *** 'savtype' = %s\n",savtype);
      exit_error("WM_RESPONSE_CONFIG","save_pop_syn_as_ bad type");
    }
    myfree(savtype);
    k += 1;
    pp = paramfile_get_next_prefix_pointer(pp,"save_pop_syn_as_");
  }
  pp = paramfile_get_first_prefix_pointer(r->ppl,"save_pop_grid_as_");
  while(pp != NULL){
    nval = paramfile_count_values_pointer(pp);
    if (nval != 10)
      exit_error("WM_RESPONSE_CONFIG","save_pop_grid_as_ bad value list");
    gnd_name = strdup(pp->name+17);
    savtype = paramfile_get_nth_char_param_pp_or_exit(pp,0);
    gsamp = paramfile_get_nth_float_param_ptr_or_exit(pp,1);
    gname = paramfile_get_nth_char_param_pp_or_exit(pp,2);
    gx0   = paramfile_get_nth_int_param_ptr_or_exit(pp,3);
    gy0   = paramfile_get_nth_int_param_ptr_or_exit(pp,4);
    gz0   = paramfile_get_nth_int_param_ptr_or_exit(pp,5);
    gxn   = paramfile_get_nth_int_param_ptr_or_exit(pp,6);
    gyn   = paramfile_get_nth_int_param_ptr_or_exit(pp,7);
    gzn   = paramfile_get_nth_int_param_ptr_or_exit(pp,8);
    gdid  = paramfile_get_nth_char_param_pp_or_exit(pp,9);
    for(i=0;i<gxn;i++){
      for(j=0;j<gyn;j++){
	for(l=0;l<gzn;l++){
	  r->rformat[k] = 2; 
	  r->samp[k] = gsamp;
	  if (gzn == 1)
	    sprintf(tstr,"%s_%d_%d",gnd_name,gx0+i,gy0+j);
	  else
	    sprintf(tstr,"%s_%d_%d_%d",gnd_name,gx0+i,gy0+j,gz0+l);
	  r->nd_name[k] = strdup(tstr);
	  r->plname[k] = strdup(gname);
	  r->xi[k]   = gx0+i;
	  r->yi[k]   = gy0+j;
	  r->zi[k]   = gz0+l;
	  r->datid[k]  = strdup(gdid);
	  if (strcmp(savtype,"s")==0){
	    r->rtype[k] = strdup("s");
	    r->ri[k] = r->ns;
	    r->ns += 1;
	  }else if (strcmp(savtype,"f")==0){
	    r->rtype[k] = strdup("f");
	    r->ri[k] = r->nf;
	    r->nf += 1;
	  }else{
	    printf("*** Error in .rsp file:\n");
	    printf("  *** %s\n",tstr);
	    printf("  *** 'savtype' = %s\n",savtype);
	    exit_error("WM_RESPONSE_CONFIG","save_pop_grid_as_ bad type");
	  }
	  k += 1;
	}
      }
    }
    myfree(savtype); myfree(gname); myfree(gnd_name); myfree(gdid);
    pp = paramfile_get_next_prefix_pointer(pp,"save_pop_grid_as_");
  }
  pp = paramfile_get_first_prefix_pointer(r->ppl,"save_pop_layer_as_");
  while(pp != NULL){
    nval = paramfile_count_values_pointer(pp);
    if (nval != 4)
      exit_error("WM_RESPONSE_CONFIG","save_pop_layer_as_ bad value list");
    gnd_name = strdup(pp->name+18);
    savtype = paramfile_get_nth_char_param_pp_or_exit(pp,0);
    gsamp = paramfile_get_nth_float_param_ptr_or_exit(pp,1);
    gname = paramfile_get_nth_char_param_pp_or_exit(pp,2);
    gdid  = paramfile_get_nth_char_param_pp_or_exit(pp,3);
    pop_util_onode_get_pop_xn_yn_zn(m->o,gname,&gxn,&gyn,&gzn);
    if (gxn == -1){ 
      gxn = onode_getpar_int_exit(m->o,"xn");
      gyn = onode_getpar_int_exit(m->o,"yn");
      gzn = 1;
    }
    for(i=0;i<gxn;i++){
      for(j=0;j<gyn;j++){
	for(l=0;l<gzn;l++){
	  r->rformat[k] = 3; 
	  r->samp[k] = gsamp;
	  if (gzn == 1)
	    sprintf(tstr,"%s_%d_%d",gnd_name,i,j);
	  else
	    sprintf(tstr,"%s_%d_%d_%d",gnd_name,i,j,l);
	  r->nd_name[k] = strdup(tstr);
	  r->plname[k] = strdup(gname);
	  r->xi[k]   = i;
	  r->yi[k]   = j;
	  r->zi[k]   = l;
	  r->datid[k]  = strdup(gdid);
	  if (strcmp(savtype,"s")==0){
	    r->rtype[k] = strdup("s");
	    r->ri[k] = r->ns;
	    r->ns += 1;
	  }else if (strcmp(savtype,"f")==0){
	    r->rtype[k] = strdup("f");
	    r->ri[k] = r->nf;
	    r->nf += 1;
	  }else{
	    printf("*** Error in .rsp file:\n");
	    printf("  *** %s\n",tstr);
	    printf("  *** 'savtype' = %s\n",savtype);
	    exit_error("WM_RESPONSE_CONFIG","save_pop_layer_as_ bad type");
	  }
	  k += 1;
	}
      }
    }
    myfree(savtype); myfree(gname); myfree(gnd_name); myfree(gdid);
    pp = paramfile_get_next_prefix_pointer(pp,"save_pop_layer_as_");
  }
  tdup = carray_get_duplicate_ptr(r->nd_name,r->n);
  if (tdup != NULL){
    sprintf(tstr,"Duplicate name for saved data: %s",tdup);
    mylogx(mylogf,"WM_RESPONSE_CONFIG",tstr);
  }
}
void wm_response_check_trial_full(r,ti,mylogf,proc_name)
     struct response_struct *r;
     int ti;                     
     char *mylogf;
     char *proc_name;            
{
  int i;
  int di,n;
  char tstr[SLEN];
  for(i=0;i<r->n;i++){  
    di = r->ri[i];  
    if (strcmp(r->rtype[i],"s")==0){
      n = r->cnt[di][ti];
    }else if (strcmp(r->rtype[i],"f")==0){
      n = r->fcnt[di][ti];
    }else{
      exit_error("WM_RESPONSE_CHECK_TRIAL_FULL","Unknown response type");
    }
    if (n < 0){
      if (proc_name != NULL)
	sprintf(tstr,"  DataID %s not filled on trial %d for processor %s",
		r->datid[i],ti,proc_name);
      else
	sprintf(tstr,"  DataID %s not filled on trial %d",r->datid[i],ti);
      printf("  *** WM_RESPONSE_CHECK_TRIAL_FULL  %s\n  *** Exiting.\n",tstr);
      mylogx(mylogf,"WM_RESPONSE_CHECK_TRIAL_FULL",tstr);
    }
  }
}
void wm_get_sform_xn_yn_tn(m,s,rsform,rxn,ryn,rtn)
     struct model_struct *m;     
     struct stim_struct *s;      
     char **rsform;
     int *rxn;
     int *ryn;
     int *rtn;
{
  int xn,yn,tn;
  char *sform;
  sform = paramfile_get_char_param_default(s->ppl,"stim_form","3d");
  if (m->ppl != NULL)
    tn = paramfile_get_int_param_or_exit(m->ppl,"tn");
  else
    tn = onode_getpar_int_exit(m->o,"tn");
  if ((strcmp(sform,"3d")==0) || (strcmp(sform,"3d_b")==0) ||
      (strcmp(sform,"3c")==0) || (strcmp(sform,"3c_b")==0)){
    if (m->ppl != NULL){
      xn = paramfile_get_int_param_or_exit(m->ppl,"xn");
      yn = paramfile_get_int_param_or_exit(m->ppl,"yn");
    }else{
      xn = onode_getpar_int_exit(m->o,"xn");
      yn = onode_getpar_int_exit(m->o,"yn");
    }
  }else{
    xn = yn = 0;
  }
  *rsform = sform;
  *rxn = xn;
  *ryn = yn;
  *rtn = tn;
}
void wm_response_init(m,s,r,mylogf)
     struct model_struct *m;
     struct stim_struct *s;
     struct response_struct *r;
     char *mylogf;
{
  int i,j;
  int n,ntr,tn;
  float tscale;
  char ggstr[LONG_SLEN];
  if (m->ppl != NULL){
    r->ns = paramfile_count_values_param(m->ppl,"save_spikes") / 2;
    r->nf = paramfile_count_values_param(m->ppl,"save_data") / 2;
    if ((r->ns > 0) || (r->nf > 0)){
      mylog_exit(mylogf,"WM_RESPONSE_INIT - OLD FORMAT\n");
    }
  }
  wm_response_count(m,r,mylogf);   
  wm_response_config(m,r,mylogf);  
  mylog(mylogf,"  Response Configuration\n");
  for(i=0;i<r->n;i++){
    if ((r->n < 40) || (i < 5) || (i >= (r->n - 6))){
      if  (r->rformat[i] == 10){
	sprintf(ggstr,
	      "%4d %2s %8.1f %2d %8s %2d %2d %2d %8s %2d %2d %2d %14s %14s\n",
	      r->rformat[i],r->rtype[i],r->samp[i],r->ri[i],r->plname[i],
	      r->xi[i],r->yi[i],r->zi[i],r->plname1[i],
	      r->xi1[i],r->yi1[i],r->zi1[i],r->datid[i],r->nd_name[i]);
      }else{
	sprintf(ggstr,"%4d %2s %8.1f %2d %8s %2d %2d %2d %17s %14s %14s\n",
	      r->rformat[i],r->rtype[i],r->samp[i],r->ri[i],r->plname[i],
	      r->xi[i],r->yi[i],r->zi[i],"  ",r->datid[i],r->nd_name[i]);
      }
      mylog(mylogf,ggstr);
    }else if (i == 6){
      strcpy(ggstr,"   ...     ...\n");
      mylog(mylogf,ggstr);
    }
  }
  r->ts0 = 0.0;
  r->ttref = 0;
  if (m->ppl != NULL){
    tscale = paramfile_get_float_param_or_exit(m->ppl,"tscale");
    tn = paramfile_get_int_param_or_exit(m->ppl,"tn");
  }else{
    tscale = onode_getpar_flt_exit(m->o,"tscale");
    tn     = onode_getpar_int_exit(m->o,"tn");
    tn     = onode_getpar_int_dflt(m->o,"tn_response",tn);
  }
  r->tsn = (float)tn*tscale;   
  r->tsi  = 0;  
  r->gtsi = 0;  
  r->tri  = 0;  
  ntr = s->ntr * m->nrun;  
  r->dflag = get_zero_iarray(ntr);      
  s->tribyx = get_const_iarray(ntr,-1); 
  n = r->ns;
  if (n > 0){
    r->s = (float ***)myalloc(n*sizeof(float **));
    r->cnt = (int **)myalloc(n*sizeof(int *));
    for(i=0;i<n;i++){
      r->s[i] = (float **)myalloc(ntr*sizeof(float *));
      r->cnt[i] = (int *)myalloc(ntr*sizeof(int));
      for(j=0;j<ntr;j++){
	r->s[i][j] = NULL;   
	r->cnt[i][j] = -1;   
      }
    }
  }
  n = r->nf;
  if (n > 0){
    r->f = (float ***)myalloc(n*sizeof(float **));
    r->fcnt = (int **)myalloc(n*sizeof(int *));
    for(i=0;i<n;i++){
      r->f[i] = (float **)myalloc(ntr*sizeof(float *));
      r->fcnt[i] = (int *)myalloc(ntr*sizeof(int));
      for(j=0;j<ntr;j++){
	r->f[i][j] = NULL;  
	r->fcnt[i][j] = -1; 
      }
    }
  }
}
void wm_get_nd_t1_outfile(r,mylogf,myid,rpre,rform,rout,rout2)
     struct response_struct *r; 
     char *mylogf;
     int myid;
     char **rpre;      
     char **rform;     
     char **rout;      
     char **rout2;     
{
  char *outfile,*outfile2,*prefix,*format;
  prefix = paramfile_get_char_param_default(r->ppl,"outfile","zz");
  format = paramfile_get_char_param_default(r->ppl,"ndata_format","nd");
  if (strcmp(format,"nd")==0)
    outfile = paramfile_make_fname(r->ppl,"write_ndata",prefix,".nd");
  else if (strcmp(format,"nd,t1")==0){
    outfile = paramfile_make_fname(r->ppl,"write_ndata",prefix,".nd");
    outfile2 = paramfile_make_fname(r->ppl,"write_ndata",prefix,".t1");
  }else
    outfile = paramfile_make_fname(r->ppl,"write_ndata",prefix,".t1");
  *rpre  = prefix;
  *rform = format;
  *rout  = outfile;
  *rout2 = outfile2;
}
void wm_make_ndtrial(m,s,r,mylogf,myid,ti,nchan,nd,t)
     struct model_struct *m;
     struct stim_struct *s;
     struct response_struct *r;
     char mylogf[];
     int myid;
     int ti;
     int nchan;
     struct ndata_struct *nd;
     struct ndtrial_struct *t;  
{
  int j,k;
  int ri,vi,mi;
  char *name,tstr[SLEN];
  t->tcode = ti;
  t->tref = my_rint_double(r->ttref * 1000.0);  
  t->nparam = nd->nvar;
  t->pname = (char **)myalloc(t->nparam*sizeof(char *));
  t->pval  = (char **)myalloc(t->nparam*sizeof(char *));
  vi = s->tribyx[ti];
  for(j=0;j<s->nvar;j++){
    t->pname[j] = strdup(nd->vname[j]);
    t->pval[j]  = strdup(s->vval[vi][j]); 
  }
  if (m->vps != NULL){  
    mi = ti / s->ntr;  
    for(j=0;j<m->vps->m_nvar;j++){
      t->pname[s->nvar + j] = strdup(m->vps->m_name[j]);
      t->pval[s->nvar + j]  = strdup(m->vps->m_vval[mi][j]); 
    }
    t->pname[s->nvar + m->vps->m_nvar] = strdup("MOO_var_i");
    sprintf(tstr,"%d",mi);
    t->pval[ s->nvar + m->vps->m_nvar] = strdup(tstr);
  }
  t->nrec = nchan;
  t->r = (struct ndrec_struct *)myalloc(t->nrec*sizeof(struct ndrec_struct));
  k = 0;
  for(j=0;j<r->n;j++){
    t->r[k].name = strdup(r->nd_name[j]);
    t->r[k].rcode = 0;
    t->r[k].sampling = r->samp[j];
    t->r[k].t0 = (int)(r->ts0 * r->samp[j]);
    t->r[k].tn = (int)(r->tsn * r->samp[j]);
    ri = r->ri[j]; 
    if (strcmp(r->rtype[j],"s")==0){
      if (r->cnt[ri][ti] < 0){
	sprintf(tstr,"No response stored for %s",r->datid[j]);
	mylogx(mylogf,"WM_MAKE_NDTRIAL",tstr);
      }
      t->r[k].rtype = 0;
      t->r[k].n = r->cnt[ri][ti];
      t->r[k].p = f2iarray_round(r->s[ri][ti],r->cnt[ri][ti]);
    }else if (strcmp(r->rtype[j],"f")==0){
      if (r->fcnt[ri][ti] < 0){
	sprintf(tstr,"No response stored for %s",r->datid[j]);
	mylogx(mylogf,"WM_MAKE_NDTRIAL",tstr);
      }
      t->r[k].rtype = 1;
      t->r[k].tn = r->fcnt[ri][ti];
      t->r[k].n = r->fcnt[ri][ti];
      t->r[k].x = copy_farray(r->f[ri][ti],r->fcnt[ri][ti]);
    }
    k += 1;
  }
}
struct ndata_struct *wm_make_ndata(m,s,r,mylogf,myid)
     struct model_struct *m;
     struct stim_struct *s;
     struct response_struct *r;
     char mylogf[];
     int myid;
{
  int i,j,k,ti;
  int n,nchan,ri,vi,tot_ntr;
  float samp;
  char *name,tstr[SLEN];
  struct ndata_struct *nd;
  mylog(mylogf,"  WM_MAKE_NDATA\n");
  nd = get_ndata();
  nd->class = strdup("MODEL");
  nchan = r->ns + r->nf;
  if (nchan <= 0)
    mylog_exit(mylogf,
	       "WM_MAKE_NDATA  No response data to write ndata file\n");
  nd->nconst = s->ncon; 
  nd->cname = (char **)myalloc(nd->nconst*sizeof(char *));
  nd->ctype = (char *)myalloc(nd->nconst*sizeof(char));
  nd->cval = (char **)myalloc(nd->nconst*sizeof(char *));
  for(i=0;i<nd->nconst;i++){
    nd->ctype[i] = s->ctype[i];
    nd->cname[i] = strdup(s->cname[i]);
    nd->cval[i] = strdup(s->cval[i]);
  }
  nd->nvar = s->nvar; 
  if (m->vps != NULL){
    nd->nvar += m->vps->m_nvar; 
    nd->nvar += 1;  
  }
  nd->vname = (char **)myalloc(nd->nvar*sizeof(char *));
  nd->vtype = (char *)myalloc(nd->nvar*sizeof(char));
  for(i=0;i<s->nvar;i++){
    nd->vtype[i] = s->vtype[i];
    nd->vname[i] = strdup(s->vname[i]);
  }
  if (m->vps != NULL){
    for(i=0;i<m->vps->m_nvar;i++){
      nd->vtype[s->nvar + i] = 'c';  
      nd->vname[s->nvar + i] = strdup(m->vps->m_name[i]);
    }
    nd->vtype[s->nvar + m->vps->m_nvar] = 'i';
    nd->vname[s->nvar + m->vps->m_nvar] = strdup("MOO_var_i");
  }
  nd->ntable = 0;
  tot_ntr = s->ntr * m->nrun;
  if (myid == 0){ 
    n = 0;
    for(i=0;i<tot_ntr;i++){
      if (r->dflag[i] == 1)
	n += 1;
    }
  }else{
    n = tot_ntr;  
  }
  nd->ntrial = n;
  nd->t = (struct ndtrial_struct *)myalloc(n*sizeof(struct ndtrial_struct));
  ti = 0; 
  for(i=0;i<tot_ntr;i++){
    if (r->dflag[i] == 1){
      wm_make_ndtrial(m,s,r,mylogf,myid,i,nchan,nd,&(nd->t[ti]));
      ti += 1;
    }
  }
  return nd;
}
struct ndata_struct *wm_stat1_cut(mylogf,nd)
     char *mylogf;
     struct ndata_struct *nd;
{
  int i,j,k;
  int n,di,seqflag,seed,nblk,ndur,*ton,*toff,**ndx,nvar,ts,dt,ti,*t0,*tn;
  int *sdata,ns,dur,**s,*cnt,period,pflag;
  float tscale,tsep,sampling;
  char **vname,*vtype,***vval,tstr[SLEN],ggstr[LONG_SLEN];
  struct ndata_struct *ndcut;
  mylog(mylogf,"  WM_STAT1_CUT\n");
  pflag = 0;
  if (nd->ntrial != 1)
    exit_error("WM_STAT1_CUT","File must have only 1 trial to cut.");
  if (nd->t[0].nrec != 1)
    exit_error("WM_STAT1_CUT","Trial must have only 1 record to cut.");
  dur =    nd->t[0].r[0].tn;
  sdata =  nd->t[0].r[0].p;
  ns =     nd->t[0].r[0].n;
  sampling = nd->t[0].r[0].sampling;
  sprintf(ggstr,"    Single trial, duration %d, spikes %d\n",dur,ns);
  mylog(mylogf,ggstr);
  seed    = ndata_get_const_param_int(nd,"seed");
  seqflag = ndata_get_const_param_int(nd,"seq_type");
  nblk    = ndata_get_const_param_int(nd,"nblock");
  tsep    = ndata_get_const_param_float(nd,"tsep");
  tscale  = 0.001;  
  ts = my_rint(tsep/tscale);  
  stm_stat1d_get_seq(NULL,nblk,seqflag,seed,tscale,&ndur,&ton,&toff,&ndx);
  if (pflag){
    printf("  Stimuli:\n");
    for(i=0;i<ndur;i++){
      printf("    %2d %6d %6d\n",i,ton[i],toff[i]);
    }
  }
  n = nblk * ndur; 
  t0 = (int *)myalloc(n*sizeof(int));  
  tn = (int *)myalloc(n*sizeof(int));  
  s = (int **)myalloc(n*sizeof(int *));  
  cnt = (int *)myalloc(n*sizeof(int));
  nvar = 2;
  vname = (char **)myalloc(nvar*sizeof(char *));
  vtype = (char *)myalloc(nvar*sizeof(char));
  vname[0] = strdup("stn");   vtype[0] = 'f';
  vname[1] = strdup("stn2");  vtype[1] = 'f';
  vval = get_2d_pointer_carray(n,nvar);
  k = 0; 
  ti = 0;  
  period = 0;  
  for(i=0;i<nblk;i++){
    for(j=0;j<ndur;j++){
      di = ndx[i][j];
      sprintf(tstr,"%.2f",(float)(ton[di]) * tscale);
      vval[k][0] = strdup(tstr);
      sprintf(tstr,"%.2f",(float)(toff[di]) * tscale);
      vval[k][1] = strdup(tstr);
      t0[k] = ti;                    
      dt = ts + ton[di] + toff[di];  
      tn[k] = dt;                    
      ti += dt;                      
      if (ti > dur){ 
	mylog_exit(mylogf,"");
	sprintf(ggstr,"*** WM_STAT1_CUT  Stimuli truncated");
      }
      s[k] = extract_spikes(sdata,ns,t0[k],tn[k],&(cnt[k]),1);
      if (tn[k] > period)
	period = tn[k];
      k += 1;
    }
  }
  if (pflag){
    printf("  Var Params:\n");
    for(i=0;i<n;i++){
      printf("    trial %2d %8d %6d  ",i,t0[i],tn[i]);
      for(j=0;j<nvar;j++){
	printf("  %s",vval[i][j]);
      }
      printf("\n");
    }
  }
  ndcut = make_ndata_sarray(s,cnt,n,period,nd->nconst,nvar,nd->cname,nd->ctype,
			    nd->cval,vname,vtype,vval,nd->class,sampling);
  myfree(t0); myfree(tn);
  myfree(ton); myfree(toff); free_2d_iarray(ndx,nblk);
  free_sarray(s,cnt,n);
  myfree(vtype);
  free_2d_pointer_carray(vval,n,nvar);
  free_2d_carray(vname,nvar);
  return ndcut;
}
void wm_write_response(m,s,r,mylogf,myid)
     struct model_struct *m;    
     struct stim_struct *s;     
     struct response_struct *r; 
     char *mylogf;
     int myid;
{
  int cutflag;
  char *outfile,*prefix,*format,tstr[SLEN],ggstr[LONG_SLEN],*stim_type;
  char *outfile2;
  struct ndata_struct *nd,*ndcut;
  struct param_pair *pp;
  wm_get_nd_t1_outfile(r,mylogf,myid,&prefix,&format,&outfile,&outfile2);
  if (outfile != (char *)NULL){
    nd = wm_make_ndata(m,s,r,mylogf,myid);
    if (nd->ntrial < s->ntr){
      sprintf(tstr,"%s.partial",outfile);
      myfree(outfile);
      outfile = strdup(tstr);
      printf("    Partial results:  %d of %d trials\n",nd->ntrial,s->ntr);
    }
    cutflag = paramfile_get_int_param_default(r->ppl,"ndata_cut",0);
    if (cutflag == 1){
      if (!ndata_get_const_param(nd,"stim_type",&stim_type)){
	mylog_exit(mylogf,"WM_WRITE_RESPONSE  Cannot find stim_type");
      }else{
	ndcut = NULL;
	if (strcmp(stim_type,"stat1")==0){
	  ndcut = wm_stat1_cut(mylogf,nd);
	}else{
	  sprintf(ggstr,"  stim_type = %s\n",stim_type);
	  mylog(mylogf,ggstr);
	  mylog_exit(mylogf,"WM_WRITE_RESPONSE  Do not know how to cut");
	}
      }
      free_ndata(nd);  
      nd = ndcut;      
      myfree(stim_type);
    }
    pp = paramfile_get_param_pointer(r->ppl,"decision_gen");
    if (pp != NULL){
      int nval,t0,tn;
      char *ch_name,*ctype,*name_add,*name_sub;
      float *dval,*dvadd,*dvsub;
      mylog(mylogf,"  WM_WRITE_RESPONSE  Writing decision file\n");
      nval = paramfile_count_values_pointer(pp);
      if (nval != 6)
	exit_error("WM_WRITE_RESPONSE","decision_gen has bad value list");
      ch_name  = paramfile_get_nth_char_param_pp_or_exit(pp,0);
      ctype    = paramfile_get_nth_char_param_pp_or_exit(pp,1);
      name_add = paramfile_get_nth_char_param_pp_or_exit(pp,2);
      name_sub = paramfile_get_nth_char_param_pp_or_exit(pp,3);
      t0       = paramfile_get_nth_int_param_ptr_or_exit(pp,4);
      tn       = paramfile_get_nth_int_param_ptr_or_exit(pp,5);
      ndata_util_choice_01(nd,ctype,t0,tn,name_add,name_sub,ch_name,
			   &dval,&dvadd,&dvsub);
      myfree(dval); myfree(dvadd); myfree (dvsub);
    }
    if (strcmp(format,"nd")==0)
      write_ndata(outfile,nd);
    else if (strcmp(format,"t1")==0)
      write_ndata_t1(outfile,"iModel",nd,4,0); 
    else if (strcmp(format,"nd,t1")==0){
      write_ndata(outfile,nd);
      write_ndata_t1(outfile2,"iModel",nd,4,0); 
    }else
      mylog_exit(mylogf,"WM_WRITE_RESPONSE  Unknown ndata format\n");
    free_ndata(nd);
    myfree(outfile);
    myfree(format);
  }
  myfree(prefix);
}
void get_stimulus_data_tuning_curve(m,s,r,k,ktr,mylogf)
     struct model_struct *m;     
     struct stim_struct *s;      
     struct response_struct *r;  
     int k,ktr;                  
     char *mylogf;
{
  int i;
  int tsamp,xn,yn,tn,write_nd,ndstim_flag;
  float sscale,tscale,stim_samp;
  char *sform,tname[2048];
  s->stimno = k;
  tname[0] = '\0';
  if (s->nvar > 0){
    for(i=0;i<s->nvar;i++){
      strcat(tname,s->vname[i]);
      strcat(tname,"_");
      strcat(tname,s->vval[ktr][i]);
      if (i < (s->nvar-1))
	strcat(tname,"_");
    }
  }else
    strcpy(tname,"stim0");  
  if (s->name != NULL)
    myfree(s->name);
  s->name = strdup(tname);
  if (m->ppl != NULL)
    tscale = paramfile_get_float_param_or_exit(m->ppl,"tscale");
  else
    tscale = onode_getpar_flt_exit(m->o,"tscale");
  wm_get_sform_xn_yn_tn(m,s,&sform,&xn,&yn,&tn);
  if ((strcmp(sform,"3d")==0) || (strcmp(sform,"3d_b")==0) ||
      (strcmp(sform,"3c")==0) || (strcmp(sform,"3c_b")==0)){
    if (m->ppl != NULL)
      sscale = paramfile_get_float_param_or_exit(m->ppl,"sscale");
    else
      sscale = onode_getpar_flt_exit(m->o,"sscale");
  }else
    sscale = 0.0;
  stim_samp = paramfile_get_float_param_or_exit(s->ppl,"stim_samp");
  tsamp = stm_get_tsamp(tscale,stim_samp);
  if (strcmp(sform,"3d")==0){
    mod_util_get_3d_stim(s,ktr,xn,yn,tn,sscale,tscale,tsamp,mylogf);
  }else if (strcmp(sform,"3c")==0){
    mod_util_get_3c_stim(s,ktr,xn,yn,tn,sscale,tscale,tsamp,mylogf);
  }else if (strcmp(sform,"3d_b")==0){
    mod_util_get_3d_stim_binoc(s,ktr,xn,yn,tn,sscale,tscale,tsamp,mylogf);
  }else if (strcmp(sform,"1d")==0){
    s->d1 = mod_util_get_1d_stim(s,ktr,tn,tscale,tsamp,mylogf);
  }else
    mylog_exit(mylogf,"GET_STIMULUS_DATA_TUNING_CURVE stim_form invalid\n");
  write_nd = paramfile_get_int_param_default(r->ppl,"write_nd",0);
  if (write_nd)
    ndstim_flag = paramfile_get_int_param_default(r->ppl,"write_nd_stim",0);
  if (write_nd && ndstim_flag){
    if (0){ 
      float *tseq;
      int ntseq;
      mylog(mylogf,"***** WYETH - writing 10 zeros instead of random sequ.\n");
      ntseq = 10;
      tseq = get_zero_farray(ntseq);
      myfree(tseq);
    }else
      mylog_exit(mylogf,
		 "GET_STIMULUS_DATA_TUNING_CURVE  Expecting tseq data\n");
  }
  myfree(sform);
}
void wm_moo_var_par_set(mylogf,m,k)
     char *mylogf;
     struct model_struct *m;    
     int k;                     
{
  int i;
  int n,flag;
  char ggstr[SLEN],*pval,*pname;
  n = m->vps->m_nvar;  
  for(i=0;i<n;i++){
    pval  = m->vps->m_vval[k][i];  
    pname = m->vps->m_name[i];     
    flag = onode_update_value(m->o,pname,pval);
    if (flag == 0){
      sprintf(ggstr,"*** WM_MOO_VAR_PAR_SET Cannot update: %s",pname);
      mylog_exit(mylogf,ggstr);
    }
  }
}
void wm_slave_action(m,s,r,actflag,mylogf)
     struct model_struct *m;    
     struct stim_struct *s;     
     struct response_struct *r; 
     int actflag;               
     char *mylogf;
{
  int k;
  char *modtype;
  m->action = actflag;  
  if (actflag == 1)
    k = s->repno;  
  else
    k = -1;
  if (m->ppl != NULL)
    modtype = paramfile_get_char_param_or_exit(m->ppl,"mod_type");
  else
    modtype = onode_getpar_chr_exit(m->o,"mod_type");
  if (strcmp(modtype,"lin_01")==0)
    mod_lin_run_lin_01(m,s,r,k,actflag);
  else if (strcmp(modtype,"dog_01")==0)
    mod_dog_run_dog_01(m,s,r,k,actflag);
  else if (strcmp(modtype,"simp_01")==0)
    mod_dog_run_simp_01(m,s,r,k,actflag);
  else if (strcmp(modtype,"comp_01")==0)
    mod_dog_run_comp_01(m,s,r,k,actflag);
  else if (strcmp(modtype,"ds_01")==0)
    mod_dog_run_ds_01(m,s,r,k,actflag);
  else if (strcmp(modtype,"ds_02")==0)
    mod_dog_run_ds_02(m,s,r,k,actflag);
  else if (strcmp(modtype,"ifc_01")==0)
    ifc_util_run_ifc_01(m,s,r,k,actflag);
  else if (strcmp(modtype,"ifc_02")==0)
    ifc_util_run_ifc_02(m,s,r,k,actflag);
  else if (strcmp(modtype,"ifc_02_02")==0)
    ifc_util_run_ifc_02_02(m,s,r,k,actflag);
  else if (strcmp(modtype,"ifc_03")==0)
    ifc_util_run_ifc_03(m,s,r,k,actflag);
  else if (strcmp(modtype,"me_gabor_01")==0)
    mod_me_run_me_gabor_01(m,s,r,k,actflag);
  else if (strcmp(modtype,"me_ab_01")==0)
    mod_me_run_me_gabor_01(m,s,r,k,actflag);
  else if (strcmp(modtype,"me_v5")==0)
    mod_me_run_me_v5(m,s,r,k,actflag);
  else if (strcmp(modtype,"random_filter")==0)
    mod_me_run_me_gabor_01(m,s,r,k,actflag);
  else if (strcmp(modtype,"binoc_filter")==0)
    mod_me_run_me_gabor_01(m,s,r,k,actflag);
  else if (strcmp(modtype,"filter_xyt")==0)
    mod_me_run_filter_xyt(m,s,r,k,actflag);
  else if (strcmp(modtype,"gabor_comp")==0)
    mod_me_run_gabor_comp(m,s,r,k,actflag);
  else if (strcmp(modtype,"rd_2gabor")==0)
    mod_me_run_gabor_comp(m,s,r,k,actflag);
  else if (strcmp(modtype,"ds_simp_01")==0)
    mod_me_run_ds_simp_01(m,s,r,k,actflag);
  else if (strcmp(modtype,"ds_liv3")==0)
    mod_me_run_ds_liv3(m,s,r,k,actflag);
  else if (strcmp(modtype,"mem")==0)
    mod_me_run_mem(m,s,r,k,actflag);
  else if (strcmp(modtype,"reichardt")==0)
    mod_me_run_reich(m,s,r,k,actflag);
  else if (strcmp(modtype,"mesh_rgc_01")==0)
    mod_mesh_run_mesh_rgc_01(m,s,r,k,actflag);
  else if (strcmp(modtype,"mesh_mag_01")==0)
    mod_mesh_run_mesh_mag_01(m,s,r,k,actflag);
  else if (strcmp(modtype,"pop")==0)
    mod_pop_run(m,s,r,actflag);
  else if (strcmp(modtype,"vhf")==0)
    mod_vhf_run_01(m,s,r,actflag);
  else if (strcmp(modtype,"caffe_net")==0)
    mod_dcn_run_01(m,s,r,actflag);
  else if (strcmp(modtype,"spop_01")==0)
    mylog_exit(mylogf,"WM_SLAVE_ACTION  'spop_01' deactivated\n");
  else if (strcmp(modtype,"spop_02")==0)
    mylog_exit(mylogf,"WM_SLAVE_ACTION  'spop_02' deactivated\n");
  else if (strcmp(modtype,"xytcorr")==0)
    mod_test_run_xytcorr01(m,s,r,actflag);
  else if (strcmp(modtype,"x")==0)
    mod_x_run_01(m,s,r,actflag);
  else if (strcmp(modtype,"srf")==0)
    mod_srf_run_01(m,s,r,actflag);
  else if (strcmp(modtype,"wave_fbars97")==0)
    mod_wav_run_fbars(m,s,r,actflag);
  else
    mylog_exit(mylogf,"WM_SLAVE_ACTION  Unknown model type\n");
  myfree(modtype);
}
