/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <GL/glx.h>
#include <GL/gl.h>
#include <math.h>
#include <unistd.h>
#include <dirent.h>  
#include "my_util.h"
#include "carray_util.h"
#include "farray_util.h"
#include "glplot_util.h"
#include "wygit.h"
Bool wygit_true_test()
{
  return 1;
}
void wygit_free(ws)
     struct wygit_struct *ws;
{
  int i;
  int n;
  if (strcmp(ws->wtype,"select")==0){
    n = ws->select_n;
    if (n > 0){
      myfree(ws->select_r);
      myfree(ws->select_g);
      myfree(ws->select_b);
      free_2d_carray(ws->select_list,n);
    }
  }else if (strcmp(ws->wtype,"buttonlist")==0){
    if (ws->button_n > 0){
      for(i=0;i<ws->button_n;i++){
	if (ws->button[i]->name != NULL)
	  myfree(ws->button[i]->name);
	if (ws->button[i]->shape != NULL)
	  myfree(ws->button[i]->shape);
	myfree(ws->button[i]);
      }
      myfree(ws->button);
    }
  }else if (strcmp(ws->wtype,"input_text")==0){
    myfree(ws->text_txt);
    if (ws->text_label != NULL)
      myfree(ws->text_label);
  }else{
    printf("  wtype:  %s\n",ws->wtype);
    exit_error("WYGIT_FREE","Unknown type");
  }
  myfree(ws);
}
void wygit_get_gl_coords(ws,xi,yi,rx,ry)
     struct wygit_struct *ws;
     int xi,yi;  
     float *rx,*ry;
{
  *rx = ws->glx + (float)xi/(float)ws->winw * ws->glw;
  *ry = ws->gly + (1.0 - (float)yi/(float)ws->winh) * ws->glh;
}
void wygit_update_coords(ws,winw,winh)
     struct wygit_struct *ws;
     int winw,winh;
{
  int i;
  float dx,dy;
  dx = (float)winw - ws->winw;
  dy = (float)winh - ws->winh;
  ws->winw = (float)winw;
  ws->winh = (float)winh;
  ws->glw = (float)winw;
  ws->glh = (float)winh;
  if (strcmp(ws->wtype,"buttonlist")==0){
    for(i=0;i<ws->button_n;i++){
      ws->button[i]->y += dy;
    }
  }else{
    exit_error("WYGIT_UPDATE_COORDS","Not implemented yet");
  }
}
struct wygit_button *wygit_button_make(name,x,y,w,h,r,g,b,tr,tg,tb,xm,ym)
     char *name;       
     float x,y,w,h;    
     float r,g,b;      
     float tr,tg,tb;   
     float xm,ym;      
{
  struct wygit_button *wb;
  wb = (struct wygit_button *)myalloc(sizeof(struct wygit_button));
  wb->name = strdup(name);
  wb->x = x;
  wb->y = y;
  wb->w = w;
  wb->h = h;
  wb->r = r;
  wb->g = g;
  wb->b = b;
  wb->tr0 = tr;
  wb->tg0 = tg;
  wb->tb0 = tb;
  wb->tr = 0.0;
  wb->tg = 0.0;
  wb->tb = 0.0;
  wb->xm = xm;
  wb->ym = ym;
  wb->state = 0;
  wb->avail = 1;
  wb->shape = strdup("square");
  wb->xi = -1;
  wb->yi = -1;
  return wb;
}
void wygit_button_label_set(wb,name,tr,tg,tb,tr0,tg0,tb0,xm,ym)
     struct wygit_button *wb;  
     char *name;               
     float tr,tg,tb;           
     float tr0,tg0,tb0;        
     float xm,ym;              
{
  if (wb->name != NULL)
    myfree(wb->name);
  wb->name = strdup(name);
  wb->tr = tr;
  wb->tg = tg;
  wb->tb = tb;
  wb->tr0 = tr0;
  wb->tg0 = tg0;
  wb->tb0 = tb0;
  wb->xm = xm;
  wb->ym = ym;
}
void wygit_buttonlist_deactivate_all(ws)
     struct wygit_struct *ws;
{
  int i;
  for (i=0;i<ws->button_n;i++)
    ws->button[i]->state = 0;
}
int wygit_buttonlist_get_active(ws)
     struct wygit_struct *ws;
{
  int i;
  int bi,flag;
  flag = 0;  
  bi = -1;
  for (i=0;i<ws->button_n;i++){
    if (ws->button[i]->state == 1){
      if (flag == 0){
	flag = 1;
	bi = i;  
      }else{
	bi = -1;   
      }
    }
  }
  return bi;
}
int wygit_buttonlist_bi_for_name(ws,name)
     struct wygit_struct *ws;
     char *name;
{
  int i;
  int done,bi;
  bi = -1;
  done = 0;
  i = 0;
  while(!done){
    if (i >= ws->button_n)
      done = 1;
    else if (ws->button[i]->name != NULL){
      if (strcmp(ws->button[i]->name,name) == 0){
	done = 1;
	bi = i;
      }
    }
    i += 1;
  }
  return bi;
}
struct wygit_struct *wygit_buttonlist_make(winw,winh,glx,gly,glw,glh,nbut)
     int winw,winh;         
     float glx,gly,glw,glh; 
     int nbut;              
{
  int i;
  struct wygit_struct *ws;
  ws = (struct wygit_struct *)myalloc(sizeof(struct wygit_struct));
  strcpy(ws->wtype,"buttonlist");
  ws->winw = winw;
  ws->winh = winh;
  ws->glx = glx;
  ws->gly = gly;
  ws->glw = glw;
  ws->glh = glh;
  ws->font = NULL;
  ws->state = 0;
  ws->x = 0.0;  
  ws->y = 0.0;
  ws->button_n = nbut;
  ws->button = (struct wygit_button **)
    myalloc(nbut*sizeof(struct wygit_button *));
  for(i=0;i<nbut;i++)
    ws->button[i] = NULL;
  ws->button_tflag = 1;  
  return ws;
}
struct wygit_struct *wygit_buttonlist_make_x_h(winw,winh,glx,gly,glw,glh,
					       nbut,bstr,font)
     int winw,winh;         
     float glx,gly,glw,glh; 
     int nbut;              
     char **bstr;           
     XFontStruct *font;     
{
  int i;
  float x,y,w,h,r,g,b,tr,tg,tb,xm,ym;
  float cf,bmarg,tlen;
  struct wygit_struct *ws;
  ws = wygit_buttonlist_make(winw,winh,glx,gly,glw,glh,nbut);
  r = g = b = tr = tg = tb = 1.0;
  cf = glw / (float)winw;
  bmarg = cf * 4.0;  
  h = cf * 20.0;
  xm = ym = bmarg;
  x = 0.0;
  y = 0.0;
  for(i=0;i<nbut;i++){
    tlen = cf * glplot_text_len_pix(font,bstr[i]);
    w = tlen + 2.0*bmarg;
    ws->button[i] = wygit_button_make(bstr[i],x,y,w,h,r,g,b,tr,tg,tb,xm,ym);
    x += w + bmarg;
  }
  return ws;
}
struct wygit_struct *wygit_buttongrid_make(winw,winh,glx,gly,glw,glh,nrow,ncol,
					   x,y,w,h,r,g,b,tr,tg,tb,xm,ym,mm)
     int winw,winh;         
     float glx,gly,glw,glh; 
     int nrow;              
     int ncol;              
     float x,y,w,h;    
     float r,g,b;      
     float tr,tg,tb;   
     float xm,ym;      
     float mm;         
{
  int i,j,k;
  int nbut;
  char name[SLEN];
  float xx,yy;
  struct wygit_struct *ws;
  nbut = nrow * ncol;
  ws = wygit_buttonlist_make(winw,winh,glx,gly,glw,glh,nbut);
  k = 0;
  for(i=0;i<nrow;i++){
    for(j=0;j<ncol;j++){
      sprintf(name,"%d,%d",i,j);
      xx = x + j*(w+mm);
      yy = y - (i+1)*(h+mm);
      ws->button[k] = wygit_button_make(name,xx,yy,w,h,r,g,b,tr,tg,tb,xm,ym);
      ws->button[k]->xi = j;
      ws->button[k]->yi = i;
      k += 1;
    }
  }
  return ws;
}
int wygit_buttonlist(ws,xi,yi)
     struct wygit_struct *ws;
     int xi,yi;  
{
  int i;
  int n,bi;
  float x,y,bx,by;
  struct wygit_button *b;
  wygit_get_gl_coords(ws,xi,yi,&x,&y);
  n = ws->button_n;
  bi = -1;
  i = 0;
  while( (i<n) && (bi == -1)){
    b = ws->button[i];
    bx = b->x + ws->x; 
    by = b->y + ws->y;
    if ((bx < x) && ((bx + b->w) > x) &&
	(by < y) && ((by + b->h) > y)){
      bi = i;
    }else{
      i += 1;
    }
  }
  return bi;
}
void wygit_buttonlist_draw(dpy,win,ws)
     Display *dpy;
     Window win;
     struct wygit_struct *ws;
{
  int i;
  float x0,y0,x,y,w,h,r,g,b,tr,tg,tb,tr0,tg0,tb0,xm,ym;
  struct wygit_button *but;
  x0 = ws->x;
  y0 = ws->y;
  glDisable(GL_DEPTH_TEST); 
  for(i=0;i<ws->button_n;i++){
    but = ws->button[i];
    x = ws->button[i]->x + x0;
    y = ws->button[i]->y + y0;
    w = ws->button[i]->w;
    h = ws->button[i]->h;
    r = ws->button[i]->r;
    g = ws->button[i]->g;
    b = ws->button[i]->b;
    tr = ws->button[i]->tr;
    tg = ws->button[i]->tg;
    tb = ws->button[i]->tb;
    tr0 = ws->button[i]->tr0;
    tg0 = ws->button[i]->tg0;
    tb0 = ws->button[i]->tb0;
    xm = ws->button[i]->xm;
    ym = ws->button[i]->ym;
    if (strcmp(but->shape,"square")==0){
      glBegin(GL_POLYGON);
      if (but->state == 1)
	glColor3f(r,g,b);
      else
	glColor3f(0.0,0.0,0.0);
      glVertex2f(x,y);
      glVertex2f(x+w,y);
      glVertex2f(x+w,y+h);
      glVertex2f(x,y+h);
      glEnd();
      glBegin(GL_LINE_LOOP);
      glColor3f(r,g,b);
      glVertex2f(x,y);
      glVertex2f(x+w,y);
      glVertex2f(x+w,y+h);
      glVertex2f(x,y+h);
      glEnd();
    }
    if (ws->button_tflag){
      if (but->state == 1){
	glplot_draw_txt(ws->button[i]->name,x+xm,y+ym,tr,tg,tb);
      }else{
	glplot_draw_txt(ws->button[i]->name,x+xm,y+ym,tr0,tg0,tb0);
      }
    }
  }
}
void wygit_buttonlist_set_reset_exclusive(ws,bi)
     struct wygit_struct *ws;
     int bi;  
{
  int i;
  struct wygit_button *but;
  for(i=0;i<ws->button_n;i++){
    but = ws->button[i];
    if (i == bi){
      if (but->state == 0)
	but->state = 1;
      else
	but->state = 0;
    }else
      but->state = 0;
  }
}
int wygit_buttonlist_get_excl(ws)
     struct wygit_struct *ws;
{
  int i;
  int n,bi;
  struct wygit_button *but;
  n = ws->button_n;
  bi = -1;
  for(i=0;i<n;i++){
    but = ws->button[i];
    if (but->state == 1){
      if (bi == -1)
	bi = i;
      else
	printf("WYGIT_BUTTONLIST_GET_EXCL:  two buttons active\n");
    }
  }
  return bi;
}
int wygit_buttonlist_next_prev_excl(ws,npflag)
     struct wygit_struct *ws;
     int npflag;  
{
  int i;
  int n,bi;
  struct wygit_button *but;
  n = ws->button_n;
  bi = -1;
  for(i=0;i<n;i++){
    but = ws->button[i];
    if (but->state == 1){
      if (bi == -1)
	bi = i;
      else
	printf("WYGIT_BUTTONLIST_NEXT_PREV_EXCL:  two buttons active\n");
    }
  }
  if (bi != -1){
    if (npflag == 0){ 
      bi += 1;
      if (bi >= n)
	bi = 0;
    }else{
      bi -= 1;
      if (bi < 0)
	bi = n-1;
    }
  }
  return bi;
}
struct wygit_struct *wygit_select_make(winw,winh,glx,gly,glw,glh,
				       slist,ns,x,y,w,dy,hn)
     int winw,winh;         
     float glx,gly,glw,glh; 
     char **slist;          
     int ns;                
     float x,y;             
     float w;               
     float dy;              
     int hn;                
{
  int i;
  struct wygit_struct *ws;
  ws = (struct wygit_struct *)myalloc(sizeof(struct wygit_struct));
  strcpy(ws->wtype,"select");
  ws->winw = winw;
  ws->winh = winh;
  ws->glx = glx;
  ws->gly = gly;
  ws->glw = glw;
  ws->glh = glh;
  ws->font = NULL;
  ws->state = 0;
  ws->x = x;
  ws->y = y;
  ws->w = w;
  ws->h = (float)hn * dy;
  ws->hn = hn;
  ws->r = ws->g = ws->b = 1.0;
  if (ns > 0){
    ws->select_r = (float *)myalloc(ns*sizeof(float));
    ws->select_g = (float *)myalloc(ns*sizeof(float));
    ws->select_b = (float *)myalloc(ns*sizeof(float));
    for(i=0;i<ns;i++){
      ws->select_r[i] = ws->select_g[i] = ws->select_b[i] = 1.0;
    }
    ws->select_list = copy_2d_carray(slist,ns);
    ws->select_flag = get_zero_iarray(ns); 
  }else{
    ws->select_r = NULL;
    ws->select_g = NULL;
    ws->select_b = NULL;
    ws->select_list = NULL;
    ws->select_flag = NULL;
  }
  ws->select_n    = ns;
  ws->select_dy   = dy;
  ws->select_ym  = 0.0;  
  ws->select_xm  = 0.0;  
  ws->select_cancel = 1;
  ws->select_pos = 0;  
  ws->select_1 = 1;    
  return ws;
}
int wygit_select_get_item_for_coord(ws,xi,yi)
     struct wygit_struct *ws;
     int xi,yi;  
{
  float x,y,dy,bot;
  wygit_get_gl_coords(ws,xi,yi,&x,&y);
  if (ws->select_n > ws->hn)
    bot = (float)(ws->hn);
  else
    bot = (float)(ws->select_n);
  dy = ((ws->y + ws->h) - y) / ws->select_dy;  
  if ((dy > 0.0) && (dy < bot)){
    yi = (int)dy;
    yi += ws->select_pos;
  }else
    yi = -1;
  if ((x < ws->x) || (x > (ws->x + ws->w)))
    yi = -1;
  return yi;
}
void wygit_select_draw(dpy,win,ws)
     Display *dpy;
     Window win;
     struct wygit_struct *ws;
{
  int i;
  int i1,i2,ns,hn;
  char **slist;
  float x,y,w,h,dy,r,g,b;
  float xx,yy;
  x = ws->x;
  y = ws->y;
  w = ws->w;
  h = ws->h;
  dy = ws->select_dy;
  hn = ws->hn;
  ns = ws->select_n;
  slist = ws->select_list;
  i1 = ws->select_pos;
  if (ns < hn)
    i2 = ns-1;
  else
    i2 = i1+hn - 1;
  glClear(GL_DEPTH_BUFFER_BIT);
  yy = y + (float)(hn-1)*dy + ws->select_ym + 0.20*dy;
  xx = x + ws->select_xm;
  for(i=i1;i<=i2;i++){
    r = ws->select_r[i];
    g = ws->select_g[i];
    b = ws->select_b[i];
    if (ws->select_1 != 1){
      if (ws->select_flag[i] == 0){  
	r *= 0.5;
	g *= 0.5;
	b *= 0.5;
      }
    }
    glplot_draw_txt(slist[i],xx,yy,r,g,b);
    yy -= dy;
  }
  r = ws->r;
  g = ws->g;
  b = ws->b;
  glBegin(GL_POLYGON);
  glColor3f(0.2*r,0.2*g,0.2*b);
  glVertex2f(x,y);
  glVertex2f(x+w,y);
  glVertex2f(x+w,y+h);
  glVertex2f(x,y+h);
  glEnd();
  glXSwapBuffers(dpy,win);
}
void wygit_select_process_event(dpy,win,ws,pevent)
     Display *dpy;
     Window win;
     struct wygit_struct *ws;
     XEvent *pevent;
{
  int i;
  int redrawflag,n,xi,yi,scroll_max;
  XKeyEvent *kevent;
  char kbuf[10];
  ws->select_i = -1; 
  scroll_max = ws->select_n - ws->hn;
  if (scroll_max < 0)
    scroll_max = 0;
  redrawflag = 0;
  switch(pevent->type){
  case ButtonPress:
    xi = pevent->xbutton.x;
    yi = pevent->xbutton.y;
    switch(pevent->xbutton.button){
    case 1:
      i = wygit_select_get_item_for_coord(ws,xi,yi);
      if (i != -1){
	ws->select_i = i;
	return;
      }
      break;
    case 2:
      break;
    case 3:
      break;
    case 4:
      if (ws->select_pos > 0){
	ws->select_pos -= 1;
	redrawflag = 1;
      }
      break;
    case 5:
      if (ws->select_pos < scroll_max){
	ws->select_pos += 1;
	redrawflag = 1;
      }
      break;
    }
    break;
  case MotionNotify:
    break;
  case KeyPress:
    kevent = (XKeyEvent *)pevent;
    XLookupString(kevent,kbuf,sizeof(kbuf),NULL,NULL);
    if ((kbuf[0]=='h')||(kbuf[0]=='?')){
      ;
    }else if ((int)kbuf[0]== 27){  
      ws->select_i = -2;
      return;
    }
    break;
  case ConfigureNotify:
    break;
  case Expose:
    break;
  }
  if (redrawflag){
    wygit_select_draw(dpy,win,ws);
  }
}
void wygit_select_loop(dpy,win,ws)
     Display *dpy;
     Window win;
     struct wygit_struct *ws;
{
  int i;
  XEvent event;
  while(1){
    if (XCheckIfEvent(dpy,&event,wygit_true_test,(char *)NULL)){
      wygit_select_process_event(dpy,win,ws,&event);
      i = ws->select_i;
      if (i >= 0)
	ws->select_flag[i] = 1 - ws->select_flag[i];
      if (i != -1)
	return;
    }
  }
}
void wygit_select(dpy,win,ws)
     Display *dpy;
     Window win;
     struct wygit_struct *ws;
{
  wygit_select_draw(dpy,win,ws);
  wygit_select_loop(dpy,win,ws);
  if (ws->select_1 == 1)
    return;
  else{
    while(ws->select_i != -2){
      wygit_select_draw(dpy,win,ws);
      wygit_select_loop(dpy,win,ws);
    }
  }
}
void wygit_select_update_with_flist(ws,flist,rn,depth,newpath)
     struct wygit_struct *ws;
     char **flist;              
     int *rn;                   
     int depth;                 
     char *newpath;             
{
  int i;
  int n;
  char fullname[LONG_SLEN];
  n = *rn;
  if (strcmp(flist[0],".")==0){
    myfree(flist[0]);
    for(i=0;i<(n-1);i++)
      flist[i] = flist[i+1];
    n -= 1;
  }
  if (depth == 0){
    if (strcmp(flist[0],"..")==0){
      myfree(flist[0]);
      for(i=0;i<(n-1);i++)
	flist[i] = flist[i+1];
      n -= 1;
    }
  }
  if (ws->select_n > 0){
    myfree(ws->select_r);
    myfree(ws->select_g);
    myfree(ws->select_b);
    free_2d_carray(ws->select_list,ws->select_n);
    myfree(ws->select_flag);
  }
  ws->select_r = (float *)myalloc(n*sizeof(float));
  ws->select_g = (float *)myalloc(n*sizeof(float));
  ws->select_b = (float *)myalloc(n*sizeof(float));
  ws->select_flag = get_zero_iarray(n);
  ws->select_list = flist;
  ws->select_n    = n;
  ws->select_pos  = 0;
  ws->select_i    = -1;
  for(i=0;i<n;i++){
    sprintf(fullname,"%s/%s",newpath,flist[i]);
    if (is_dir(fullname)){
      ws->select_r[i] = 1.0;
      ws->select_g[i] = 0.4;
      ws->select_b[i] = 0.4;
    }else if (is_suffix_string(flist[i],".stm")){
      ws->select_r[i] = 1.0;
      ws->select_g[i] = 1.0;
      ws->select_b[i] = 0.0;
    }else{
      ws->select_r[i] = 0.5;
      ws->select_g[i] = 0.5;
      ws->select_b[i] = 1.0;
    }
  }
  *rn = n;
}
int wygit_select_file_return(dpy,win,ws,pathname,pevent,rpath)
     Display *dpy;
     Window win;
     struct wygit_struct *ws;
     char *pathname;
     XEvent *pevent;
     char **rpath;              
{
  int i,k;
  int n,done;
  char **flist,tpath[LONG_SLEN];
  static int flag = 0;
  static int depth;
  static char newpath[LONG_SLEN];
  done = 0;
  if (flag == 0){ 
    opendir(pathname);
    sprintf(newpath,"%s",pathname);
    depth = 0;
    flag = 1;
    flist = filelist(newpath,&n);
    if (n <= 0){
      done = 1;  
    }else{
      wygit_select_update_with_flist(ws,flist,&n,depth,newpath);
      wygit_select_draw(dpy,win,ws);
    }
  }else{
    wygit_select_process_event(dpy,win,ws,pevent);
    i = ws->select_i;
    strcpy(tpath,newpath);
    if (i >= 0){
      if (strcmp(ws->select_list[i],"..")==0){
	k = strlen(newpath) - 1;
	while(newpath[k] != '/')
	  k -= 1;
	newpath[k] = '\0';
	depth -= 1;
      }else{
	sprintf(newpath,"%s/%s",tpath,ws->select_list[i]);
	depth += 1;
      }
      flist = filelist(newpath,&n);
      if (n <= 0){
	done = 1;  
      }else{
	wygit_select_update_with_flist(ws,flist,&n,depth,newpath);
	wygit_select_draw(dpy,win,ws);
      }
    }else if (i == -2){  
      newpath[0] = '\0';
      done = 1;
    }
  }
  *rpath = (char *)NULL;
  if (done == 1){
    flag = 0;
    if (newpath[0] != '\0')
      *rpath = strdup(newpath);
  }
  return done;
}
char *wygit_select_file(dpy,win,ws,pathname)
     Display *dpy;
     Window win;
     struct wygit_struct *ws;
     char *pathname;
{
  int i,k;
  int n,done,depth;
  char **flist;
  char newpath[LONG_SLEN],tpath[LONG_SLEN];
  opendir(pathname);
  sprintf(newpath,"%s",pathname);
  depth = 0;
  done = 0;
  while(!done){
    flist = filelist(newpath,&n);
    if (n <= 0){
      done = 1;
    }else{
      wygit_select_update_with_flist(ws,flist,&n,depth,newpath);
      wygit_select(dpy,win,ws);
      i = ws->select_i;
      strcpy(tpath,newpath);
      if (i >= 0){
	if (strcmp(ws->select_list[i],"..")==0){
	  k = strlen(newpath) - 1;
	  while(newpath[k] != '/')
	    k -= 1;
	  newpath[k] = '\0';
	  depth -= 1;
	}else{
	  sprintf(newpath,"%s/%s",tpath,ws->select_list[i]);
	  depth += 1;
	}
      }else{  
	newpath[0] = '\0';
	done = 1;
      }
    }
  }
  if (newpath[0] == '\0')
    return (char *)NULL;
  else
    return strdup(newpath);
}
struct wygit_struct *wygit_input_text_make(winw,winh,glx,gly,glw,glh,
					   x,y,w,h)
     int winw,winh;         
     float glx,gly,glw,glh; 
     float x,y,w,h;
{
  int i;
  struct wygit_struct *ws;
  ws = (struct wygit_struct *)myalloc(sizeof(struct wygit_struct));
  strcpy(ws->wtype,"input_text");
  ws->winw = winw;
  ws->winh = winh;
  ws->glx = glx;
  ws->gly = gly;
  ws->glw = glw;
  ws->glh = glh;
  ws->font = NULL;
  ws->state = 0;
  ws->x = x;
  ws->y = y;
  ws->w = w;
  ws->h = h;
  ws->r = ws->g = ws->b = 0.2;
  ws->text_maxchar = LONG_SLEN;
  strcpy(ws->text_type,"all");
  ws->text_xm = 0.0;
  ws->text_ym = 0.0;
  ws->text_r = 1.0;
  ws->text_g = 1.0;
  ws->text_b = 1.0;
  ws->text_txt = (char *)myalloc(ws->text_maxchar * sizeof(char));
  ws->text_txt[0] = '\0';
  ws->text_n   = 0;
  ws->text_label = NULL;
  ws->text_label_flag = 0;
  return ws;
}
int wygit_input_text_click(ws,xi,yi)
     struct wygit_struct *ws;
     int xi,yi;  
{
  int i;
  int n,flag;
  float x,y,bx,by;
  wygit_get_gl_coords(ws,xi,yi,&x,&y);
  flag = 0;
  bx = ws->x;
  by = ws->y;
  if ((bx < x) && ((bx + ws->w) > x) &&
      (by < y) && ((by + ws->h) > y)){
    flag = 1;
  }
  return flag;
}
void wygit_input_text_draw(dpy,win,ws)
     Display *dpy;
     Window win;
     struct wygit_struct *ws;
{
  float x,y,w,h,tr,tg,tb,xm,ym,tlen;
  char *tstr;
  x    = ws->x;
  y    = ws->y;
  w    = ws->w;
  h    = ws->h;
  xm   = ws->text_xm;
  ym   = ws->text_ym;
  tr   = ws->text_r;
  tg   = ws->text_g;
  tb   = ws->text_b;
  tstr = ws->text_txt;
  glDisable(GL_DEPTH_TEST); 
  glBegin(GL_POLYGON);
  glColor3f(ws->r,ws->g,ws->b);
  glVertex2f(x,y);
  glVertex2f(x+w,y);
  glVertex2f(x+w,y+h);
  glVertex2f(x,y+h);
  glEnd();
  glBegin(GL_LINE_LOOP);
  if (ws->state == 1)
    glColor3f(1.0,1.0,1.0);
  else
    glColor3f(0.0,0.0,0.0);
  glVertex2f(x,y);
  glVertex2f(x+w,y);
  glVertex2f(x+w,y+h);
  glVertex2f(x,y+h);
  glEnd();
  glplot_draw_txt(tstr,x+xm,y+ym,tr,tg,tb);
  if (ws->text_label_flag == 1){
    tlen = glplot_text_len_pix(ws->font,ws->text_label);
    glplot_draw_txt(ws->text_label,x-tlen-2.0*xm,y+ym,tr,tg,tb);
  }else if (ws->text_label_flag == 2)
    glplot_draw_txt(ws->text_label,x+w+2.0*xm,y+ym,tr,tg,tb);
  else if (ws->text_label_flag == 3)
    glplot_draw_txt(ws->text_label,x+xm,y+h+2.0*ym,tr,tg,tb);
  else if (ws->text_label_flag == 4)
    glplot_draw_txt(ws->text_label,x+xm,y-h,tr,tg,tb);
}
void wygit_input_text_process_char(dpy,win,ws,k,fflag)
     Display *dpy;
     Window win;
     struct wygit_struct *ws;
     int k;      
     int fflag;  
{
  int n;
  float x,y,w,h,tr,tg,tb,xm,ym,pixmax;
  char *tstr;
  x    = ws->x;
  y    = ws->y;
  w    = ws->w;
  h    = ws->h;
  xm   = ws->text_xm;
  ym   = ws->text_ym;
  tr   = ws->text_r;
  tg   = ws->text_g;
  tb   = ws->text_b;
  tstr = ws->text_txt;
  n    = ws->text_n;
  pixmax = w - 2.0*xm;
  if ((k >= 32) && (k <= 126)){
    tstr[n] = (char)k;
    n += 1;
    tstr[n] = '\0';
    if (glplot_text_len_pix(ws->font,tstr) > pixmax){
      n -= 1;
      tstr[n] = '\0';
    }
  }else if (k == 8){  
    if (n > 0){
      n -= 1;
      tstr[n] = '\0';
    }else{
      tstr[0] = '\0';  
    }
  }else if (k == 13){  
    ; 
  }
  ws->text_n = n; 
  wygit_input_text_draw(dpy,win,ws);
  if (fflag == 1)
    glFlush();
  else if (fflag == 2)
    glXSwapBuffers(dpy,win);
}
char *wygit_input_text_loop(dpy,win,ws,dbflag)
     Display *dpy;
     Window win;
     struct wygit_struct *ws;
     int dbflag;
{
  int i,k;
  int n,xi,yi,fflag;
  XEvent event;
  XKeyEvent *kevent;
  char kbuf[10];
  fflag = dbflag + 1;
  wygit_input_text_process_char(dpy,win,ws,0,fflag);
  while(1){
    if (XCheckIfEvent(dpy,&event,wygit_true_test,(char *)NULL)){
      switch(event.type){
      case ButtonPress:
	xi = event.xbutton.x;
	yi = event.xbutton.y;
	switch(event.xbutton.button){
	case 1:
	  if (wygit_input_text_click(ws,xi,yi) == 0) 
	    return strdup(ws->text_txt);
	  break;
	case 2:
	  break;
	case 3:
	  break;
	case 4:
	  break;
	case 5:
	  break;
	}
	break;
      case MotionNotify:
	break;
      case KeyPress:
	kevent = (XKeyEvent *)&event;
	XLookupString(kevent,kbuf,sizeof(kbuf),NULL,NULL);
	k = (int)(kbuf[0]);
	if (k == 13){  
	  return strdup(ws->text_txt);
	}else{
	  wygit_input_text_process_char(dpy,win,ws,(int)(kbuf[0]),fflag);
	}
	break;
      case ConfigureNotify:
	break;
      case Expose:
	break;
      }
    }
    usleep(1000); 
  }
}
char *wygit_input_text(dpy,win,ws,dbflag)
     Display *dpy;
     Window win;
     struct wygit_struct *ws;
     int dbflag;
{
  int done;
  char *txt;
  txt = NULL;
  done = 0;
  while(!done){
    txt = wygit_input_text_loop(dpy,win,ws,dbflag);
    if (txt != NULL)
      done = 1;
  }
  return txt;
}
struct wygit_group *wygit_group_get_init(n,winw,winh,x,y,w,h)
     int n;
     int winw,winh;  
     float x,y,w,h;  
{
  int i;
  struct wygit_group *wyg;
  wyg = (struct wygit_group *)myalloc(sizeof(struct wygit_group));
  wyg->nmax = n;
  wyg->n = 0;
  wyg->w = (struct wygit_struct **)myalloc(n*sizeof(struct wygit_struct *));
  wyg->flag = (int *)myalloc(n*sizeof(int));
  wyg->name = (char **)myalloc(n*sizeof(char *));
  for(i=0;i<n;i++){
    wyg->w[i] = NULL;
    wyg->flag[i] = 0;
    wyg->name[i] = NULL;
  }
  wyg->winw = winw;           
  wyg->winh = winh;
  wyg->glx = x;   
  wyg->gly = y;
  wyg->glw = w;
  wyg->glh = h;
  wyg->x = 0.0;     
  wyg->y = 0.0;
  wyg->redrawflag = 1;  
  wyg->dbflag = 0;      
  return wyg;
}
void wygit_group_draw(dpy,win,wg)
     Display *dpy;
     Window win;
     struct wygit_group *wg;
{
  int i;
  int n;
  struct wygit_struct *w;
  n = wg->nmax;
  for(i=0;i<n;i++){
    if (wg->flag[i] == 1){
      w = wg->w[i];
      if (strcmp(w->wtype,"buttonlist")==0)
	wygit_buttonlist_draw(dpy,win,w);
      else if (strcmp(w->wtype,"input_text")==0)
	wygit_input_text_draw(dpy,win,w);
      else
	printf("  *** Don't know how to draw %s\n",w->wtype);
    }
  }
}
void wygit_group_event(dpy,win,wg)
     Display *dpy;
     Window win;
     struct wygit_group *wg;
{
  int i,k;
  int n,xi,yi,bi,done,flag;
  XEvent event;
  XKeyEvent *kevent;
  char kbuf[10],*txt;
  struct wygit_struct *w;
  n = wg->n;
  if (XCheckIfEvent(dpy,&event,wygit_true_test,(char *)NULL)){
    switch(event.type){
    case ButtonPress:
      xi = event.xbutton.x;
      yi = event.xbutton.y;
      switch(event.xbutton.button){
      case 1:
	bi = -1;
	i = 0;
	done = 0;
	while(!done){
	  w = wg->w[i];
	  if (strcmp(w->wtype,"buttonlist")==0){
	    bi = wygit_buttonlist(w,xi,yi); 
	    if (bi >= 0){
	      wygit_buttonlist_set_reset_exclusive(w,bi);
	      done = 1;
	      wg->redrawflag = 1;
	    }
	  }else if (strcmp(w->wtype,"input_text")==0){
	    flag = wygit_input_text_click(w,xi,yi);
	    if (flag == 1){
	      w->state = 1;
	      txt = wygit_input_text(dpy,win,w,wg->dbflag);
	      myfree(txt);
	      w->state = 0;
	      wg->redrawflag = 1;
	      done = 1;
	    }
	  }
	  if (!done){
	    i += 1;
	    if (i >= n)
	      done = 1;
	  }
	}
	break;
      case 2:
	break;
      case 3:
	break;
      case 4:
	break;
      case 5:
	break;
      }
      break;
    case MotionNotify:
      break;
    case KeyPress:
      kevent = (XKeyEvent *)&event;
      XLookupString(kevent,kbuf,sizeof(kbuf),NULL,NULL);
      k = (int)(kbuf[0]);
      if (k == 13){  
	printf("Enter\n");
      }else if ((int)kbuf[0]== 27){  
	wg->state = 0;
      }else{
	;
      }
      break;
    case ConfigureNotify:
      break;
    case Expose:
      break;
    }
  }
}
void wygit_group_loop(dpy,win,wg)
     Display *dpy;
     Window win;
     struct wygit_group *wg;
{
  int i;
  wg->redrawflag = 1;
  wg->state = 1;  
  i = 0;
  while(wg->state > 0){
    if (wg->redrawflag == 1){
      wygit_group_draw(dpy,win,wg);
      if (wg->dbflag == 1)
	glXSwapBuffers(dpy,win);
      else
	glFlush();
      wg->redrawflag = 0;
    }
    usleep(1000); 
    wygit_group_event(dpy,win,wg);
    i += 1;
  }
}
void wygit_group_get_data(wg,rdval)
     struct wygit_group *wg;
     char ***rdval;
{
  int i;
  int n,bi;
  char **dval;
  struct wygit_struct *w;
  n = wg->n;
  dval  = (char **)myalloc(n*sizeof(char *));
  for(i=0;i<wg->n;i++){
    w = wg->w[i];
    if (strcmp(w->wtype,"buttonlist")==0){
      bi = wygit_buttonlist_get_active(w);
      if (bi >= 0){
	dval[i] = strdup(w->button[bi]->name);
      }else
	dval[i] = NULL;
    }else if (strcmp(w->wtype,"input_text")==0)
      dval[i] = strdup(w->text_txt);
    else
      dval[i] = NULL;
  }
  *rdval = dval;
}
