/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <unistd.h>
#include "my_util.h"
#include "nr_util.h"
#include "carray_util.h"
#include "iarray_util.h"
#include "farray_util.h"
#include "paramfile_util.h"
#include "ndata_util.h"
#include "mctrl_util.h"
#include "mod_util.h"
#include "wm_util.h"
#include "paramfile.h"
#include "ndata.h"
#include "mod.h"
#include "mm_util.h"
#include "mm.h"
#define BUFFSIZE  100000
#define MAX_BUF_RESULT 1000000  
static int myid = -1;        
static char *mylogf = NULL;  
static char *tlogf = NULL;   
char ggstr[LONG_SLEN];       
int numprocs;        
char processor_name[MPI_MAX_PROCESSOR_NAME];
int mm_gui_flag = 1;         
char *mm_make_var_name(s,ti)
     struct stim_struct *s;      
     int ti;                     
{
  int i;
  char tstr[LONG_SLEN];
  if (s->nvar == 0){
    sprintf(tstr,"No parameters vary.");
  }else{
    tstr[0] = '\0';
    for(i=0;i<s->nvar;i++){
      strcat(tstr,s->vname[i]);
      strcat(tstr,"=");
      strcat(tstr,s->vval[ti][i]);
      strcat(tstr,"  ");
    }
  }
  return strdup(tstr);
}
void mm_slave_wait_to_send(r,gti)
     struct response_struct *r; 
     int gti;                   
{
  int i;
  int done,tsig;
  MPI_Status status;
  MPI_Send(&gti,1,MPI_INT,0,tag_req2send,MPI_COMM_WORLD); 
  mylog(mylogf," ==> Want to send results, waiting for OK to send ...\n");
  MPI_Recv(&tsig,1,MPI_INT,0,tag_ok2send,MPI_COMM_WORLD,&status);
  mylog(mylogf,"     Received OK to send.\n");
}
void mm_slave_send_results(r,gti)
     struct response_struct *r; 
     int gti;                   
{
  int i;
  int n,ns,nf;
  float *pf;
  ns = r->ns;
  nf = r->nf;
  wm_response_check_trial_full(r,gti,mylogf,processor_name);
  mm_slave_wait_to_send(r,gti);
  sprintf(ggstr," *** Sending %d spike trains (%d float)\n",ns,nf);
  mylog(mylogf,ggstr);
  MPI_Send(&gti,1,MPI_INT,0,tag_result,MPI_COMM_WORLD);  
  MPI_Send(&ns,1,MPI_INT,0,0,MPI_COMM_WORLD);  
  for(i=0;i<ns;i++){
    pf = r->s[i][gti];
    n  = r->cnt[i][gti];
    if (n > MAX_BUF_RESULT){
      sprintf(ggstr,"gti %d  i %d  n %d  MAX %d\n",gti,i,n,MAX_BUF_RESULT);
      mylog(mylogf,ggstr);
      mylog_exit(mylogf,"MM_SLAVE_SEND_RESULTS  cnt > MAX_BUF_RESULT\n");
    }else if (n == -1){
      sprintf(ggstr,"gti %d  i %d  n %d  %s\n",gti,i,n,r->datid[gti]);
      mylog(mylogf,ggstr);
      mylog_exit(mylogf,"MM_SLAVE_SEND_RESULTS  cnt > MAX_BUF_RESULT\n");
    }
    MPI_Send(pf,n,MPI_FLOAT,0,0,MPI_COMM_WORLD);
  }
  MPI_Send(&nf,1,MPI_INT,0,0,MPI_COMM_WORLD);  
  for(i=0;i<nf;i++){
    pf = r->f[i][gti];
    n  = r->fcnt[i][gti];
    if (n > MAX_BUF_RESULT)
      mylog_exit(mylogf,"MM_SLAVE_SEND_RESULTS  fcnt > MAX_BUF_RESULT\n");
    MPI_Send(pf,n,MPI_FLOAT,0,0,MPI_COMM_WORLD);
  }
  sprintf(ggstr,"    xxxxxx DONE slave_send_results.\n");
  mylog(mylogf,ggstr);
}
void mm_master_receive_results(r,ti,source)
     struct response_struct *r;  
     int ti;      
     int source;  
{
  int i;
  int n,ns,nf,state,tsig;
  float fbuf[MAX_BUF_RESULT];
  MPI_Status status;
  state = mm_get_proc_state(source); 
  tsig = 1;  
  MPI_Send(&tsig,1,MPI_INT,source,tag_ok2send,MPI_COMM_WORLD); 
  MPI_Recv(&ns,1,MPI_INT,source,0,MPI_COMM_WORLD,&status);
  if (ns != r->ns)
    mylog_exit(mylogf,"MM_MASTER_RECEIVE_RESULTS  ns != r->ns\n");
  for(i=0;i<ns;i++){
    MPI_Recv(fbuf,MAX_BUF_RESULT,MPI_FLOAT,source,0,MPI_COMM_WORLD,&status);
    MPI_Get_count(&status,MPI_FLOAT,&n);
    if (state == 1){
      r->cnt[i][ti] = n;
      r->s[i][ti] = copy_farray(fbuf,n);
    }
  }
  MPI_Recv(&nf,1,MPI_INT,source,0,MPI_COMM_WORLD,&status);
  if (nf != r->nf)
    mylog_exit(mylogf,"MM_MASTER_RECEIVE_RESULTS  nf != r->nf\n");
  for(i=0;i<nf;i++){
    MPI_Recv(fbuf,MAX_BUF_RESULT,MPI_FLOAT,source,0,MPI_COMM_WORLD,&status);
    MPI_Get_count(&status,MPI_FLOAT,&n);
    if (state == 1){
      r->fcnt[i][ti] = n;
      r->f[i][ti] = copy_farray(fbuf,n);
    }
  }
  if (state == 1)
    r->dflag[ti] = 1; 
  if (state == 2) 
    mm_proc_allow_reactivate(source); 
}
void mm_slave_run(m,s,r)
     struct model_struct *m;     
     struct stim_struct *s;      
     struct response_struct *r;  
{
  int si,tnum,snum,oldsnum,oldmi,done;
  int xn,yn,tn,nrpt,mi;
  char *sform,*tname;
  MPI_Status status;
  mylog(mylogf,"  MM_SLAVE_RUN\n");
  wm_get_sform_xn_yn_tn(m,s,&sform,&xn,&yn,&tn);
  nrpt = paramfile_get_int_param_or_exit(s->ppl,"stim_nrpt");
  oldsnum = -1;   
  oldmi   = -1;   
  done = 0;
  while (!done){
    mylog(mylogf,"  Waiting to receive a job ...");
    MPI_Recv(&tnum,1,MPI_INT,0,tag_runtrial,MPI_COMM_WORLD,&status);
    if (tnum == -1){
      done = 1;
      mylog(mylogf,"  Instructed to terminate.\n");
    }else{
      r->gtsi = tnum;  
      mi = (int)(tnum / s->ntr);   
      si = tnum - mi*s->ntr;       
      snum = (int)(si / nrpt);
      s->repno = si - snum*nrpt;
      m->m_i = mi;  
      sprintf(ggstr,"  Got job:  Mod %d  Stim %d  Rpt %d\n",mi,snum,s->repno);
      mylog(mylogf,ggstr);
      if (mi != oldmi){  
	if (oldmi > -1)  
	  wm_slave_action(m,s,r,-1,mylogf); 
	if (m->nrun > 1)
	  wm_moo_var_par_set(mylogf,m,mi); 
	r->tri = si;
	r->tsi = si;  
	wm_slave_action(m,s,r,0,mylogf);  
	oldmi = mi;
      }else{
	r->tri = si;  
	r->tsi = si;
      }
      if (snum != oldsnum){ 
	if (strcmp(sform,"3d")==0){
	  if (s->d != NULL)
	    free_f3tensor(s->d,1,xn,1,yn,1,tn);
	}else if (strcmp(sform,"3d_b")==0){
	  if (s->d != NULL)
	    free_f3tensor(s->d,1,xn,1,yn,1,tn);
	  if (s->d_r != NULL)
	    free_f3tensor(s->d_r,1,xn,1,yn,1,tn);
	}else if (strcmp(sform,"3c")==0){
	  if (s->dc != NULL)
	    free_3d_iarray(s->dc,xn,yn,tn);  
	}else if (strcmp(sform,"1d")==0){
	  if (s->d1 != NULL)
	    myfree(s->d1);
	}else if (strcmp(sform,"fseq")==0){
	  mylog_exit(mylogf,"MM_SLAVE_RUN  WYETH - must clean up fseq\n");
	}
	get_stimulus_data_tuning_curve(m,s,r,snum,si,mylogf);
	oldsnum = snum;
      }
      tname = mm_make_var_name(s,si);
      sprintf(ggstr,"    Stimulus:  %s\n",tname);
      mylog(mylogf,ggstr);
      wm_slave_action(m,s,r,1,mylogf);  
      mm_slave_send_results(r,tnum);
      myfree(tname);
    }
  }
  myfree(sform);
}
void mm_master_run(m,s,r,arglist,narg,parflag)
     struct model_struct *m;     
     struct stim_struct *s;      
     struct response_struct *r;  
     char **arglist;             
     int narg;                   
     int *parflag;               
{
  int i,k;
  int n,done,tnum,rpti,stimi;
  int nmod,nstim,nrpt,jnum,pnum;
  int flag,aflag;
  int tmin_curr,tmin_last_save;
  MPI_Request *req;
  MPI_Request req1;
  MPI_Status status;
  mylog(mylogf,"  MM_MASTER_RUN\n");
  wm_response_init(m,s,r,mylogf);
  nrpt = paramfile_get_int_param_or_exit(s->ppl,"stim_nrpt");
  nstim = s->ntr / nrpt;
  nmod = m->nrun;
  mm_init_job_control(numprocs,nmod,nstim,nrpt,mylogf,processor_name,NULL);
  mm_ja_log_proc_names(tlogf);  
  if (mm_gui_flag == 1)
    mm_gui_init(r);
  n = s->ntr * nmod;   
  req = NULL; 
  tmin_curr = tmin_last_save = 0; 
  aflag = mm_suggest_next_assignment(&pnum,&jnum);
  while(aflag == 1){
    mm_assign_job(jnum,pnum,mylogf,mm_gui_flag,tlogf);
    aflag = mm_suggest_next_assignment(&pnum,&jnum);
  }
  while (mm_get_ndone() < n){  
    MPI_Irecv(&tnum,1,MPI_INT,MPI_ANY_SOURCE,tag_req2send,MPI_COMM_WORLD,
	      &req1);
    done = 0;
    while(!done){
      MPI_Test(&req1,&flag,&status);  
      if (flag){
	done = 1;
	k = status.MPI_SOURCE; 
	mm_master_receive_results(r,tnum,k);
	stimi = mm_job_get_stimi(tnum);  
	s->tribyx[tnum] = stimi;
	if (mm_get_proc_state(k) == 1){ 
	  mm_job_done(r,tnum,k,mm_gui_flag,tlogf); 
	  if (mm_gui_flag == 0){
	    sprintf(ggstr,"    Job %d done by processor %d\n",tnum,k);
	    mylog(mylogf,ggstr);
	  }
	  aflag = mm_suggest_next_assignment(&pnum,&jnum);
	  while(aflag == 1){
	    mm_assign_job(jnum,pnum,mylogf,mm_gui_flag,tlogf);
	    aflag = mm_suggest_next_assignment(&pnum,&jnum);
	  }
	  tmin_curr = mm_time_elapsed_min();
	  if (tmin_curr >= (tmin_last_save + 5)){ 
	    wm_write_response(m,s,r,mylogf,myid);
	    tmin_last_save = tmin_curr;
	  }
	}
      }else{
	aflag = mm_suggest_next_assignment(&pnum,&jnum);
	while(aflag == 1){
	  mm_assign_job(jnum,pnum,mylogf,mm_gui_flag,tlogf);
	  aflag = mm_suggest_next_assignment(&pnum,&jnum);
	}
	if (mm_gui_flag == 1){
	  mm_gui_update();
	  mm_gui_event();
	  mm_mon_event(r); 
	}
	usleep(10000); 
      }
    }
  }
  if (mm_gui_flag == 1)
    mm_gui_hold(r,-1);
  mm_print_stat();
}
void mm_fit(m,s,r,arglist,narg,parflag)
     struct model_struct *m;     
     struct stim_struct *s;      
     struct response_struct *r;  
     char **arglist;             
     int narg;                   
     int *parflag;               
{
  int i,k;
  int n,done,tnum,rpti,stimi;
  int nmod,nstim,nrpt,jnum,pnum;
  int flag,aflag;
  int tmin_curr,tmin_last_save;
  MPI_Request *req;
  MPI_Request req1;
  MPI_Status status;
  mylog(mylogf,"  MM_FIT\n");
  if (myid == 0){
    wm_slave_action(m,s,r,2,mylogf);  
    mylog_exit(mylogf," *** MM_FIT --- Master - Under development\n");
  }else{
    wm_slave_action(m,s,r,2,mylogf);  
    mylog_exit(mylogf," *** MM_FIT --- Slave - Under development\n");
  }
}
void mm_make_msr(paramfile,stimfile,respfile,mo,sppl,rppl,rm,rs,rr)
     char paramfile[],stimfile[],respfile[];
     struct onode *mo;
     struct param_pair_list *sppl;
     struct param_pair_list *rppl;
     struct model_struct **rm;    
     struct stim_struct **rs;     
     struct response_struct **rr; 
{
  struct model_struct *m;    
  struct stim_struct *s;     
  struct response_struct *r; 
  s = (struct stim_struct *)myalloc(sizeof(struct stim_struct));
  s->ppl = sppl;
  strcpy(s->paramfile,stimfile);
  s->name = (char *)NULL;
  s->d = (float ***)NULL;
  s->d_r = (float ***)NULL;
  s->d1 = (float *)NULL;
  s->dc = (int ***)NULL;
  s->dc_r = (int ***)NULL;
  m = (struct model_struct *)myalloc(sizeof(struct model_struct));
  m->ppl = NULL;
  m->o = mo;
  strcpy(m->paramfile,paramfile);
  m->process_id = myid;
  m->process_n = numprocs;
  m->run_mode = 0;    
  m->marfile = NULL;
  m->action = -999;   
  m->nrun = 1;        
  m->vps = NULL;
  r = (struct response_struct *)myalloc(sizeof(struct response_struct));
  r->ppl = rppl;
  strcpy(r->paramfile,respfile);
  r->ns = -1; 
  r->nf = -1; 
  *rm = m; *rs = s; *rr = r;
}
void mm_slave(mmode,paramfile,stimfile,respfile)
     char mmode[];
     char paramfile[],stimfile[],respfile[];
{
  int cn;
  char *c,*ct;
  struct onode *mo;
  struct param_pair_list *sppl,*rppl;
  struct model_struct *m;    
  struct stim_struct *s;     
  struct response_struct *r; 
  MPI_Status status;
  mm_mpi_receive_carray(0,tag_mppl,100000,mylogf,&c,&cn);
  mylog(mylogf,"  MM_SLAVE received carray for mppl\n");
  ct = c;  
  mo = paramfile_get_onode_from_carray(&ct);
  myfree(c);
  mm_mpi_receive_carray(0,tag_sppl,100000,mylogf,&c,&cn);
  mylog(mylogf,"  MM_SLAVE received carray for sppl\n");
  sppl = paramfile_carray_to_ppl(c,cn);
  myfree(c);
  if (strcmp(mmode,"fit")==0){
    rppl = NULL;
  }else{
    mm_mpi_receive_carray(0,tag_rppl,100000,mylogf,&c,&cn);
    mylog(mylogf,"  MM_SLAVE received carray for rppl\n");
    rppl = paramfile_carray_to_ppl(c,cn);
    myfree(c);
  }
  MPI_Send(processor_name,strlen(processor_name)+1,MPI_CHAR,0,tag_procname,
	   MPI_COMM_WORLD);
  mm_make_msr(paramfile,stimfile,respfile,mo,sppl,rppl,&m,&s,&r);
  wm_prep_var_const_gen(m,s,mylogf,NULL,0,NULL);
  if (strcmp(mmode,"fit")==0){
    mm_fit(m,s,r,NULL,0,NULL);  
  }else{
    wm_response_init(m,s,r,mylogf);
    mm_slave_run(m,s,r);
  }
  wm_slave_action(m,s,r,-1,mylogf); 
}
void mm_master(mmode,paramfile,stimfile,respfile,arglist,narg)
     char mmode[];
     char paramfile[],stimfile[],respfile[];
     char **arglist;
     int narg;
{
  int i;
  int cn;
  int *parflag;              
  char *c;
  struct param_pair_list *sppl,*rppl;
  struct model_struct *m;    
  struct stim_struct *s;     
  struct response_struct *r; 
  struct onode *mo;
  if (is_suffix_string(paramfile,".mod"))
    mylog_exit(mylogf,"MM_MASTER  'mm' no longer accepts .mod files.\n");
  mo = paramfile_onode_file_read(paramfile,1);
  paramfile_onode_resolve(mo,mylogf); 
  parflag = onode_update_from_slist(mo,mylogf,arglist,narg);
  sppl = paramfile_get_initial_params_update_only(stimfile,arglist,narg,1);
  if (strcmp(mmode,"fit")==0){
    rppl = NULL;
    for(i=0;i<narg;i+=2){
      if (!(onode_test_ostr(mo,arglist[i])||
	    paramfile_test_param(sppl,arglist[i]))){
	sprintf(ggstr,"  *** Parameter name:  %s\n",arglist[i]);
	mylog(mylogf,ggstr);
	mylog_exit(mylogf,"MM_MASTER  Command line param not found\n");
      }
    }
  }else{
    rppl = paramfile_get_initial_params_update_only(respfile,arglist,narg,1);
    for(i=0;i<narg;i+=2){
      if (!(onode_test_ostr(mo,arglist[i])||
	    paramfile_test_param(sppl,arglist[i])||
	    paramfile_test_param(rppl,arglist[i]))){
	sprintf(ggstr,"  *** Parameter name:  %s\n",arglist[i]);
	mylog(mylogf,ggstr);
	mylog_exit(mylogf,"MM_MASTER  Command line param not found\n");
      }
    }
  }
  mm_make_msr(paramfile,stimfile,respfile,mo,sppl,rppl,&m,&s,&r);
  wm_prep_var_const_gen(m,s,mylogf,arglist,narg,parflag);
  c = paramfile_onode_carray(mo,&cn);
  for(i=1;i<numprocs;i++)
    MPI_Send(c,cn,MPI_CHAR,i,tag_mppl,MPI_COMM_WORLD);
  myfree(c);
  paramfile_ppl_to_carray(sppl,&c,&cn);
  for(i=1;i<numprocs;i++)
    MPI_Send(c,cn,MPI_CHAR,i,tag_sppl,MPI_COMM_WORLD);
  myfree(c);
  if (strcmp(mmode,"fit")!=0){
    paramfile_ppl_to_carray(rppl,&c,&cn);
    for(i=1;i<numprocs;i++)
      MPI_Send(c,cn,MPI_CHAR,i,tag_rppl,MPI_COMM_WORLD);
    myfree(c);
  }
  if (strcmp(mmode,"fit")==0){
    mm_fit(m,s,r,arglist,narg,parflag);  
    exit(0);
  }else
    mm_master_run(m,s,r,arglist,narg,parflag);
  for(i=1;i<numprocs;i++)
    mm_assign_job(-1,i,mylogf,mm_gui_flag,tlogf); 
  wm_write_response(m,s,r,mylogf,myid);
}
void mm_replay(jafile,ndfile)
     char jafile[];  
     char ndfile[];  
{
  int i;
  FILE *fopen(),*fin;
  int nproc,nstim,nrpt,pnum,jnum,tmsec,done,ntr,dt,pflag,delay_10th_sec,ns;
  int nmod;
  unsigned long tnow,t0ms;
  char **pname,tstr[LONG_SLEN];
  struct response_struct *r;
  struct ndata_struct *nd;
  struct ndtrial_struct *t0;
  printf("  MM_REPLAY\n");
  pflag = 0; 
  read_ndata(ndfile,&nd);
  if ((fin = fopen(jafile,"r")) == NULL){
    printf("  *** Cannot open file %s.\n",jafile);
    exit(0);
  }else
    printf("    Reading job log file:  %s\n",jafile);
  ns = fscanf(fin,"%*s %d",&nproc);
  printf("    NProc %d\n",nproc);
  pname = (char **)myalloc(nproc*sizeof(char *));
  for(i=0;i<nproc;i++){
    ns = fscanf(fin,"%*s %d %s",&pnum,tstr);
    pname[i] = strdup(tstr);
  }
  for(i=0;i<nproc;i++){
    printf("   %4d  %s\n",i,pname[i]);
  }
  nmod = 1; 
  ns = fscanf(fin,"%*s %d",&nstim);
  ns = fscanf(fin,"%*s %d",&nrpt);
  printf("    NStim %d\n",nstim);
  printf("    NRpt %d\n",nrpt);
  printf("    *** Assuming:  NMod %d\n",nmod);
  ntr = nmod*nstim*nrpt;  
  mm_init_job_control(nproc,nmod,nstim,nrpt,NULL,pname[0],pname);
  r = (struct response_struct *)myalloc(sizeof(struct response_struct));
  r->paramfile[0] = '\0';
  r->ppl = NULL;
  r->n = 1;  
  r->rtype = (char **)myalloc(r->n * sizeof(char *));
  r->rtype[0] = strdup("s");
  r->ns = 1;  
  r->nd_name = (char **)myalloc(r->n * sizeof(char *));
  r->nd_name[0] = strdup("spikes");
  t0 = &(nd->t[0]);
  r->tsn = (float)t0->r[0].tn / t0->r[0].sampling;
  r->s   = get_3d_farray(r->ns,ntr,0);
  r->cnt = get_2d_iarray(r->ns,ntr);
  for(i=0;i<ntr;i++){
    t0 = &(nd->t[i]);
    r->s[0][i] = i2farray(t0->r[0].p,t0->r[0].n);
    r->cnt[0][i] = t0->r[0].n;
  }
  if (ntr != nd->ntrial){
    exit_error("MM_REPLAY","ndata trials does not match nstim x nrpt.");
  }
  printf("    Trial duration %f s\n",r->tsn);
  mm_gui_init(r);
  printf("\n    *** Press 'q' to begin ***\n\n");    
  mm_gui_hold(r,1);  
  printf("    *** Press 'q' to exit ***\n\n");    
  done = 0;
  while(!done){
    ns = fscanf(fin,"%s",tstr);
    if (strcmp(tstr,"Assign")==0){
      ns = fscanf(fin,"%*s %d %*s %d",&jnum,&pnum);
      if (pflag) printf("  Assign job %d proc %d\n",jnum,pnum);
      mm_assign_job(jnum,pnum,NULL,mm_gui_flag,NULL);
      if (jnum == -1)
	done = 1;
    }else if (strcmp(tstr,"jnum")==0){
      ns = fscanf(fin,"%d %*s %d %*s %d",&jnum,&pnum,&tmsec);
      if (pflag) printf("  Done job %d proc %d in %d msec\n",jnum,pnum,tmsec);
      t0ms = mm_get_start_time_for_job(jnum);  
      tnow = mm_current_time_ms();
      dt = (int)(tnow - t0ms);
      if (tmsec > dt){
	if (pflag) printf("   Only %d of %d ms elapsed, Sleep it off\n",
			  dt,tmsec);
	delay_10th_sec = (tmsec-dt)/100; 
	for(i=0;i<delay_10th_sec;i++){
	  usleep(100000); 
	  mm_gui_update();
	  mm_mon_event(r);
	}
      }
      mm_job_done(r,jnum,pnum,mm_gui_flag,NULL); 
    }else{
      printf("  *** Found  %s\n",tstr);
      exit_error("MM_REPLAY","Unexpected string.");
    }
  }
  fclose(fin);
  mm_gui_hold(r,1);  
}
int main(argc,argv)
     int argc;
     char *argv[];
{
  int namelen,sysres;
  char option[SLEN],paramfile[SLEN],stimfile[SLEN],respfile[SLEN],tstr[SLEN];
  char jafile[SLEN],ndfile[SLEN];
  MPI_Init(&argc,&argv);
  MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
  MPI_Comm_rank(MPI_COMM_WORLD,&myid); 
  MPI_Get_processor_name(processor_name,&namelen);
  MPI_Buffer_attach(malloc(BUFFSIZE),BUFFSIZE); 
  if (myid == 0){
    mylogf = NULL;
    tlogf = mylog_get_global_ja_fname();
    remove_file(tlogf);
    sprintf(tstr,"NPROC %d\n",numprocs);  
    append_string_to_file(tlogf,tstr);
  }else{
    mylogf = mylog_get_global_fname(myid);
    remove_file(mylogf);
    mylog(mylogf,"Begin\n");
    sprintf(tstr,"chmod ugo+rwx %s",mylogf);
    sysres = system(tstr);
  }
  if (argc == 1){
    if (myid==0){
      mylog(mylogf,"\n  mm <option>\n");
      mylog(mylogf,"    where <option> is\n");
      mylog(mylogf,"      mod <paramfile> <stimfile> <respfile>");
      mylog(mylogf," [<pname> <pval> ...]\n");
      mylog(mylogf,"      modm <paramfile> <stimfile> <respfile>");
      mylog(mylogf," [<pname> <pval> ...]\n");
      mylog(mylogf,"      modq <paramfile> <stimfile> <respfile>");
      mylog(mylogf," [<pname> <pval> ...]\n");
      mylog(mylogf,"        The quiet option - no graphics.\n");
      mylog(mylogf,"      replay <ja_file> <ndata_file>\n");
      mylog(mylogf,"      fit <paramfile> <stimfile>\n");
      mylog(mylogf,"\n");
    }
    fflush(stdout);
    MPI_Finalize();
    mylog(mylogf,"  Version 1.0.1\n\n");
    return 0;
  }else{
    strcpy(option,argv[1]);
  }
  if ((strcmp(option,"mod")==0)||(strcmp(option,"modm")==0)||
      (strcmp(option,"modq")==0)||(strcmp(option,"fit")==0)){
    if (strcmp(option,"modm")==0){ 
      mctrl_util_determine_dir_structure();
      if (myid == 0){
	mylog(mylogf,"<br>\n<pre>\n");
	mctrl_util_report_pid_run();
      }
      mm_gui_flag = 0;
    }
    if (strcmp(option,"modq")==0){ 
      mm_gui_flag = 0;
    }
    strcpy(paramfile,argv[2]);
    strcpy(stimfile, argv[3]);
    if (strcmp(option,"fit")==0)
      respfile[0] = '\0';   
    else
      strcpy(respfile, argv[4]);
    if (myid == 0){
      mm_master(option,paramfile,stimfile,respfile,&(argv[5]),argc-5);
    }else{
      mm_slave(option,paramfile,stimfile,respfile);
    }
  }else if (strcmp(option,"replay")==0){
    strcpy(jafile,argv[2]);
    strcpy(ndfile,argv[3]);
    mm_replay(jafile,ndfile);
  }else
    mylog(mylogf,"  *** Unknown option.\n");
  mylog(mylogf,"\n");
  fflush(stdout);
  MPI_Finalize();
  return 0;
}
